
#include <linux/module.h>
#include <linux/init.h>
#include <linux/platform_device.h>
#include <linux/delay.h>
#include <linux/leds.h>
#include <linux/i2c.h>

#include <asm/gpio.h>
#include <mach/pwc.h>

#define DEFAULT_BACKLIGHT_BRIGHTNESS 255
#define HW_MAX_BRIGHTNESS 95
#define HW_MIN_BRIGHTNESS 16

/*******************************/
/* LCD BackLight */
static void mp200_brightness_set(struct led_classdev *led_cdev, enum led_brightness value)
{
	int hw_val = (value * HW_MAX_BRIGHTNESS) / 255;

	if (hw_val < HW_MIN_BRIGHTNESS)
		hw_val = HW_MIN_BRIGHTNESS;
	if (hw_val > HW_MAX_BRIGHTNESS)
		hw_val = HW_MAX_BRIGHTNESS;
	pwc_reg_write(DA9052_LED1CONT_REG, hw_val);
	pwc_reg_write(DA9052_LED2CONT_REG, hw_val);
}

static struct led_classdev mp200_backlight_led = {
	.name = "lcd-backlight",
	.brightness = DEFAULT_BACKLIGHT_BRIGHTNESS,
	.brightness_set = mp200_brightness_set,
};


/*******************************/
/* LED */

static struct gpio_led mp200_led_list[] = {
#if 1 // BEGIN eflow-changed
	{
		.name = "led1",
		.gpio = 73,
		.active_low = 0,
	},
	{
		.name = "led2",
		.gpio = 74,
		.active_low = 0,
	},
	{
		.name = "led3",
		.gpio = 75,
		.active_low = 0,
	},
#else
	{
		.name = "led1",
		.gpio = GPIO_LED1,
		.active_low = 1,
	},
	{
		.name = "led2",
		.gpio = GPIO_LED2,
		.active_low = 1,
	},
	{
		.name = "led3",
		.gpio = GPIO_LED3,
		.active_low = 1,
	},
	{
		.name = "led4",
		.gpio = GPIO_LED4,
		.active_low = 1,
	},
	{
		.name = "led5",
		.gpio = GPIO_PWC_LED1,
		.active_low = 1,
	},
	{
		.name = "led6",
		.gpio = GPIO_PWC_LED2,
		.active_low = 1,
	},
#endif // END eflow-changed
};
static struct gpio_led_platform_data mp200_leds_data = {
	.num_leds	= ARRAY_SIZE(mp200_led_list),
	.leds		= mp200_led_list,
};
static struct platform_device mp200_leds = {
	.name		= "leds-gpio",
	.id		= -1,
	.dev		= {
		.platform_data	= &mp200_leds_data,
	},
};

/*******************************/
static int mp200_light_probe(struct platform_device *pdev)
{
#if 0 // BEGIN eflow-removed
	/* Init LCD Backlight */
	pwc_reg_write(DA9052_LED1CONF_REG, 0xEA);  /* LCD panel max electric current 15mA, set 14.986mA */
	pwc_reg_write(DA9052_LED1CONT_REG, 0xDF);
	pwc_reg_write(DA9052_LED2CONF_REG, 0xEA);  /* LCD panel max electric current 15mA, set 14.986mA */
	pwc_reg_write(DA9052_LED2CONT_REG, 0xDF);
	led_classdev_register(&pdev->dev, &mp200_backlight_led);
#endif // END eflow-removed

	/* Init LED */
	platform_device_register(&mp200_leds);

	return 0;
}

static int mp200_light_remove(struct platform_device *pdev)
{
// eflow-removed
//	led_classdev_unregister(&mp200_backlight_led);

	platform_device_unregister(&mp200_leds);

	return 0;
}

static struct platform_driver mp200_light_driver = {
	.probe		= mp200_light_probe,
	.remove		= mp200_light_remove,
	.driver		= {
		.name		= "mp200-light",
		.owner		= THIS_MODULE,
	},
};

static int __init mp200_light_init(void)
{
	return platform_driver_register(&mp200_light_driver);
}

late_initcall(mp200_light_init);

