/*
 *  File Name       : linux/arch/arm/mach-mp200/mp200_board.c
 *  Function        : mp200_board
 *  Release Version : Ver 1.02
 *  Release Date    : 2009/06/19
 *
 *  Copyright (C) NEC Electronics Corporation 2005-2009
 *
 *
 *  This program is free software;you can redistribute it and/or modify it under the terms of
 *  the GNU General Public License as published by Free Softwere Foundation; either version 2
 *  of License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 *  without even the implied warrnty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *  See the GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License along with this program;
 *  If not, write to the Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 *  MA 02111-1307, USA.
 *
 */

#include <linux/device.h>
#include <linux/init.h>
#include <linux/interrupt.h>
#include <linux/types.h>
#include <linux/mtd/nand.h>
#include <linux/mtd/partitions.h>
#include <linux/platform_device.h>
#include <linux/mm.h>
#include <linux/i2c.h>
#include <linux/irq.h>

#include <mach/hardware.h>
#include <asm/io.h>
#include <asm/setup.h>
#include <asm/mach-types.h>
#include <asm/mach/arch.h>
#include <asm/mach/map.h>
#include <asm/mach/flash.h>
#include <asm/hardware/cache-l2x0.h>

#include <mach/smu.h>
#include <asm/gpio.h>

#include <linux/mtd/physmap.h>

#include "generic.h"
#include "time.h"

static int __initdata mp200_serial_ports[] = { 0, 1, 0, 0 };

#ifndef CONFIG_MP200_SDMOBILE
#if defined(CONFIG_SMSC911X)
#include <linux/smsc911x.h>
static struct resource smsc911x_resources[] = {
	[0] = {
		.start		= MP200_ETHER_BASE,
		.end		= MP200_ETHER_BASE + SZ_64K - 1,
		.flags		= IORESOURCE_MEM,
	},
	[1] = {
		.start		= INT_GPIO_ETHER_INT,
		.end		= INT_GPIO_ETHER_INT,
		.flags		= IORESOURCE_IRQ,
	},
};
static struct smsc911x_platform_config smsc911x_platdata = {
	.flags		= SMSC911X_USE_32BIT,
	.irq_type	= SMSC911X_IRQ_TYPE_PUSH_PULL,
	.irq_polarity = SMSC911X_IRQ_POLARITY_ACTIVE_HIGH,
};
static struct platform_device smc91x_device = {
	.name	= "smsc911x",
	.id		= 0,
	.dev  = {
		.platform_data = &smsc911x_platdata,
		},
	.num_resources	= ARRAY_SIZE(smsc911x_resources),
	.resource	= smsc911x_resources,
};
#else	/* defined(CONFIG_SMC91X) */
static struct resource smc91x_resources[] = {
	[0] = {
	       .start = MP200_ETHER_BASE+0x300,
	       .end = MP200_ETHER_BASE + 0x400,
	       .flags = IORESOURCE_MEM,
	       },
	[1] = {
	       .start = INT_GPIO_ETHER_INT,
	       .end = INT_GPIO_ETHER_INT,
	       .flags = IORESOURCE_IRQ,
	       },
};
static struct platform_device smc91x_device = {
	.name = "smc91x",
	.id = -1,
	.num_resources = ARRAY_SIZE(smc91x_resources),
	.resource = smc91x_resources,
};
#endif
#endif

static struct resource mp200_rtc_resources[] = {
	{
		.start = INT_PWC_RTCINT,
		.end   = INT_PWC_RTCINT,
		.flags = IORESOURCE_IRQ,
	},
};
static struct platform_device mp200_rtc_device = {
	.name = "mp200-rtc",
	.id = -1,
	.num_resources = ARRAY_SIZE(mp200_rtc_resources),
	.resource = mp200_rtc_resources,
};

static struct mtd_partition mp200_flash0_partitions[] = {
	{
	 .name   = "boot",
	 .offset = 0,
	 .size   = 0x00080000,
	 .mask_flags = MTD_WRITEABLE,	/* force read-only */
	},
	{
	 .name   = "kernel",
	 .offset = MTDPART_OFS_APPEND,
	 .size   = 0x00380000,
	 .mask_flags = MTD_WRITEABLE,	/* force read-only */
	},
	{
	 .name   = "root",
	 .offset = MTDPART_OFS_APPEND,
// eflow-changed
//	 .size   = 0x00400000,
	 .size   = 0x01100000,
	 .mask_flags = MTD_WRITEABLE,	/* force read-only */
	},
// BEGIN eflow-added
	{
	 .name   = "flash0",
	 .offset = MTDPART_OFS_APPEND,
	 .size   = 0x00300000,
	 .mask_flags = 0,
	},
	{
	 .name   = "flash1",
	 .offset = MTDPART_OFS_APPEND,
	 .size   = 0x00400000,
	 .mask_flags = 0,
	},
	{
	 .name   = "flash2",
	 .offset = MTDPART_OFS_APPEND,
	 .size   = 0x00200000,
	 .mask_flags = 0,
	},
	{
	 .name   = "flash3",
	 .offset = MTDPART_OFS_APPEND,
	 .size   = 0x001e0000,
	 .mask_flags = 0,
	},
	{
	 .name   = "flash4",
	 .offset = MTDPART_OFS_APPEND,
	 .size   = 0x00020000,
	 .mask_flags = 0,
	},
// END eflow-added
};
#if 1 // BEGIN eflow-changed
static struct physmap_flash_data mp200_flash0_data = {
	.width    = 2,
	.parts    = mp200_flash0_partitions,
	.nr_parts = ARRAY_SIZE(mp200_flash0_partitions),
};
static struct resource mp200_flash0_resource = {
	.start = MP200_FLASH_BASE,
	.end   = MP200_FLASH_BASE + MP200_FLASH_SIZE - 1,
	.flags = IORESOURCE_MEM,
};
static struct platform_device mp200_flash0_device = {
	.name = "physmap-flash",
//	.name = "mp200_flash",
	.id   = 0,
	.dev  = {
		.platform_data = &mp200_flash0_data,
		},
	.num_resources = 1,
	.resource = &mp200_flash0_resource,
};

#else
static struct flash_platform_data mp200_flash0_data = {
	.map_name = "cfi_probe",
	.width    = 2,
	.parts    = mp200_flash0_partitions,
	.nr_parts = ARRAY_SIZE(mp200_flash0_partitions),
};
static struct resource mp200_flash0_resource = {
	.start = MP200_FLASH_BASE,
	.end   = MP200_FLASH_BASE + MP200_FLASH_SIZE - 1,
	.flags = IORESOURCE_MEM,
};
static struct platform_device mp200_flash0_device = {
	.name = "mp200_flash",
	.id   = 0,
	.dev  = {
		.platform_data = &mp200_flash0_data,
		},
	.num_resources = 1,
	.resource = &mp200_flash0_resource,
};
#endif // END eflow-changed

// eflow-changed
#if 0
//#ifdef CONFIG_MP200_SDMOBILE
static struct mtd_partition mp200_nand_partition[] = {
	{
	 .name   = "nand device",
	 .offset = 0,
	 .size   = SZ_128M
	 },
};
static struct platform_nand_chip mp200_nand_data = {
	.nr_chips      = 1,
	.chip_delay    = 15,
	.options       = NAND_BUSWIDTH_16,
	.partitions    = mp200_nand_partition,
	.nr_partitions = ARRAY_SIZE(mp200_nand_partition),
};
static struct resource mp200_nand_resource[] = {
	{
		.start = MP200_NAND_DATA_BASE,
		.end   = MP200_NAND_DATA_BASE + 4 - 1,
		.flags = IORESOURCE_MEM,
	},
	{
		.start = MP200_NAND_COMMAND_BASE,
		.end   = MP200_NAND_COMMAND_BASE + 4 - 1 ,
		.flags = IORESOURCE_MEM,
	},
	{
		.start = MP200_NAND_ADDRESS_BASE,
		.end   = MP200_NAND_ADDRESS_BASE + 4 - 1,
		.flags = IORESOURCE_MEM,
	},
};
static struct platform_device mp200_nand_device = {
	.name = "mp200_nand",
	.id   = -1,
	.dev  = {
		.platform_data = &mp200_nand_data,
		},
	.num_resources = ARRAY_SIZE(mp200_nand_resource),
	.resource = mp200_nand_resource,
};
#endif

#ifndef CONFIG_MP200_SDMOBILE
#ifdef CONFIG_MP200_EM1_DKIT
static struct platform_device da9052_ts_device = {
	.name = "da9052-ts",
	.id = -1,
};

static struct platform_device mp200_light_device = {
	.name	= "mp200-light",
	.id		= -1,
};
#endif
// BEGIN eflow-added
#else
static struct platform_device mp200_light_device = {
	.name	= "mp200-light",
	.id		= -1,
};
// END eflow-added
#endif

// BEGIN eflow-added
static struct platform_device mp200_battery_device = {
	.name	= "mp200-battery",
	.id		= -1,
};
// END eflow-added

static struct platform_device *devs[] __initdata = {
#ifndef CONFIG_MP200_SDMOBILE
	&smc91x_device,
#endif
	&mp200_rtc_device,
	&mp200_flash0_device,
#ifdef CONFIG_MP200_SDMOBILE
/// eflow-removed
///	&mp200_nand_device,
// eflow-added
	&mp200_light_device,
#else
#ifdef CONFIG_MP200_EM1_DKIT
	&da9052_ts_device,
	&mp200_light_device,
#endif
#endif
// eflow-added
	&mp200_battery_device,
};

#ifndef CONFIG_MP200_SDMOBILE
#ifndef CONFIG_MP200_EM1_DKIT
#include <linux/i2c/tsc2007.h>
static int mp200_get_pendown_state(void)
{
	return gpio_get_value(GPIO_TP_INT) ? 0 : 1;
}

static int mp200_init_platform_hw(void)
{
	gpio_direction_input(GPIO_TP_INT);
	set_irq_type(GPIO_TP_INT + INT_GPIO_BASE, IRQ_TYPE_EDGE_FALLING);
	return 0;
}

static struct tsc2007_platform_data mp200_touch_data = {
	.model			= 2007,
	.x_plate_ohms	= 685,
	.get_pendown_state = mp200_get_pendown_state,
	.init_platform_hw = mp200_init_platform_hw,
};
#endif

static struct i2c_board_info mp200_i2c_devices[] = {
#ifdef CONFIG_MP200_EM1_DKIT
	{
		I2C_BOARD_INFO("ak4648", 0x13),
	},
	{
		I2C_BOARD_INFO("adv7179", 0x2B),
	},
	{
		I2C_BOARD_INFO("max7324_w", 0x5D),
	},
	{
		I2C_BOARD_INFO("max7324_r", 0x6D),
	},
#else
	{
		I2C_BOARD_INFO("ak4648", 0x12),
	},
	{
		I2C_BOARD_INFO("adv7179", 0x2B),
	},
	{
		I2C_BOARD_INFO("tsc2007", 0x48),
		.platform_data = &mp200_touch_data,
		.irq = GPIO_TP_INT + INT_GPIO_BASE,
	},
#endif
};
#endif

static void __init mp200_board_map_io(void)
{
	mp200_map_io();
}

static void __init mp200_board_init(void)
{

#ifdef CONFIG_MP200_L220
	unsigned int l2_pte, l2_page;
	unsigned long l2_addr = IO_ADDRESS(MP200_L220_BASE);
	pgd_t *pgd;
	pmd_t *pmd;
#endif
	void *em1_bdinfo;

	mp200_serial_init(mp200_serial_ports);

#ifdef CONFIG_MP200_QR
	outl(0x00020402, SMU_QR_CLKDIV);
	outl(0x00530103, SMU_QR_ENA);
#endif

#ifdef CONFIG_MP200_L220
	pgd = pgd_offset_k(l2_addr);
	pmd = pmd_offset(pgd, l2_addr);
	l2_pte = (unsigned int)pte_offset_kernel(pmd, l2_addr);
	l2_pte -= 2048;
	l2_page = inl(l2_pte);
	l2_page &= ~0x0c;
	l2_page |= 0x80;
	outl(l2_page, l2_pte);
	/* flush pte */
	__asm__ __volatile__("mcr p15, 0, %0, c7, c14, 1" : : "r"(l2_pte));
	__asm__ __volatile__("mcr p15, 0, %0, c8, c6, 1" : : "r"(l2_pte));	/* Inv D TLB */
	l2_pte = 0;
	__asm__ __volatile__("mcr p15, 0, %0, c7, c10, 4" : : "r"(l2_pte));	/* Sync */

#ifdef CONFIG_MP200_L220_WT
	/* Force L2 write through */
	outl(0x2, l2_addr + 0xf40);
#endif

	/* 1MB (128KB/way), 8-way associativity, evmon/parity/share enabled
	 * Bits:  .... ...0 0111 1001 0000 .... .... .... */
/*	l2x0_init((void __iomem *)l2_addr, 0x0003092d, 0xfe000000); */
	l2x0_init((void __iomem *)l2_addr, 0x00020000, 0xfe000000);
#endif

	platform_add_devices(devs, ARRAY_SIZE(devs));
#ifndef CONFIG_MP200_SDMOBILE
	i2c_register_board_info(0, mp200_i2c_devices, ARRAY_SIZE(mp200_i2c_devices));
#endif

	em1_bdinfo = ioremap(MP200_SSRAM_BASE + 0x1fffc, 4);
	system_rev = ioread32(em1_bdinfo);
	iounmap(em1_bdinfo);

}

MACHINE_START(MP200, "MP200")
	.phys_io      = MP200_BANK5_BASE,
	.boot_params  = PHYS_OFFSET+0x100,
	.map_io       = mp200_board_map_io,
	.init_irq     = mp200_init_irq,
	.init_machine = mp200_board_init,
	.timer        = &mp200_timer,
MACHINE_END
