#include <asm/regdef.h>
#include <asm/ftrace.h>
#include <asm/asm.h>

/*
 * The problem here is that _mcount() is not a normal function.  It does
 * not obey the calling conventions of a normal function.  It takes its
 * arguments in $at and $ra, and it is required to preserve $ra - and most
 * other registers, in fact, if I remember correctly.  The arguments have
 * not yet been saved.  You need to do this.  Another bizarreness is that
 * the caller needs to pop an extra 8 bytes for the O32.
 */

#if _MIPS_SIM == _ABIO32
#define POP_EXTRA 8
#define ARG_SAVE 16
#define N_ARG_REGS 4
#else
#define POP_EXTRA 0
#define ARG_SAVE 0
#define N_ARG_REGS 8
#endif

	.macro		save_mcount_frame
	PTR_SUBU	sp, sp, ARG_SAVE + (N_ARG_REGS + 4) * LONGSIZE
	LONG_S		AT, ARG_SAVE + 0 * LONGSIZE(sp)
	LONG_S		ra, ARG_SAVE + 1 * LONGSIZE(sp)
	LONG_S		a0, ARG_SAVE + 2 * LONGSIZE(sp)
	LONG_S		a1, ARG_SAVE + 3 * LONGSIZE(sp)
	LONG_S		a2, ARG_SAVE + 4 * LONGSIZE(sp)
	LONG_S		a3, ARG_SAVE + 5 * LONGSIZE(sp)
#if _MIPS_SIM == _ABI64
	LONG_S		a4, ARG_SAVE + 6 * LONGSIZE(sp)
	LONG_S		a5, ARG_SAVE + 7 * LONGSIZE(sp)
	LONG_S		a6, ARG_SAVE + 8 * LONGSIZE(sp)
	LONG_S		a7, ARG_SAVE + 9 * LONGSIZE(sp)
#endif
	LONG_S		v0, ARG_SAVE + (2 + N_ARG_REGS) * LONGSIZE(sp)
	.endm

	.macro		restore_mcount_frame
	LONG_L		AT, ARG_SAVE + 0 * LONGSIZE(sp)
	LONG_L		ra, ARG_SAVE + 1 * LONGSIZE(sp)
	LONG_L		a0, ARG_SAVE + 2 * LONGSIZE(sp)
	LONG_L		a1, ARG_SAVE + 3 * LONGSIZE(sp)
	LONG_L		a2, ARG_SAVE + 4 * LONGSIZE(sp)
	LONG_L		a3, ARG_SAVE + 5 * LONGSIZE(sp)
#if _MIPS_SIM == _ABI64
	LONG_L		a4, ARG_SAVE + 6 * LONGSIZE(sp)
	LONG_L		a5, ARG_SAVE + 7 * LONGSIZE(sp)
	LONG_L		a6, ARG_SAVE + 8 * LONGSIZE(sp)
	LONG_L		a7, ARG_SAVE + 9 * LONGSIZE(sp)
#endif
	LONG_L		v0, ARG_SAVE + (2 + N_ARG_REGS) * LONGSIZE(sp)
	PTR_ADDU	sp, sp, ARG_SAVE + (N_ARG_REGS + 4) * LONGSIZE + POP_EXTRA
	.endm

#ifdef CONFIG_DYNAMIC_FTRACE

LEAF(_mcount)
	.set		push
	.set		noreorder
	.set		noat
	jr		ra
	 move		ra, AT
	.set		pop
	END(_mcount)

LEAF(ftrace_caller)
	.set		push
	.set		noreorder
	.set		noat
	save_mcount_frame
	PTR_SUBU	a0, ra, 2 * MCOUNT_INSN_SIZE
	move		a1, AT

FEXPORT(ftrace_call)
	jal		ftrace_stub
	 nop

	restore_mcount_frame
	jr		ra
	 move		ra, AT
	.set		pop
	END(ftrace_caller)

#else

LEAF(_mcount)
	.set		push
	.set		noreorder
	.set		noat
	PTR_L		t0, ftrace_trace_function
	PTR_LA		t1, ftrace_stub
	bne		t0, t1, trace
	 nop

#if POP_EXTRA
	PTR_ADDIU	sp, sp, 8
#endif
	jr		ra
	 move		 ra, AT

trace:
	save_mcount_frame
	PTR_SUBU	a0, ra, 2 * MCOUNT_INSN_SIZE
	move		a1, AT
	jalr		t0
	 nop
	restore_mcount_frame
	jr		ra
	 move		ra, AT
	.set		pop
	END(_mcount)

#endif /* !CONFIG_DYNAMIC_FTRACE */

LEAF(ftrace_stub)
	jr		ra
	END(ftrace_stub)
