/*
 *  Registration of Malta RTC platform device.
 *
 *  Copyright (C) 2008  Tiejun Chen <tiejun.chen@windriver.com>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */

#include <linux/mc146818rtc.h>
#include <linux/platform_device.h>
#include <linux/errno.h>
#include <linux/init.h>
#include <linux/ioport.h>

static struct resource malta_platform_rtc_resource[] __initdata = {
	{
		.start	= RTC_PORT(0),
		.end	= RTC_PORT(7),
		.flags	= IORESOURCE_IO,
	},
	{
		.start	= RTC_IRQ,
		.end	= RTC_IRQ,
		.flags	= IORESOURCE_IRQ,
	},
};

static __init int malta_platform_rtc_add(void)
{
	struct platform_device *pdev;
	int ret;

	pdev = platform_device_alloc("rtc_cmos", -1);
	if (!pdev)
		return -ENOMEM;

	ret = platform_device_add_resources(pdev, malta_platform_rtc_resource,
	                                       ARRAY_SIZE(malta_platform_rtc_resource));
	if (ret)
		goto err;

	ret = platform_device_add(pdev);
	if (ret)
		goto err;

	/* Try setting rtc as BCD mode to support
	 * current alarm code if possible.
	 */
	if (!RTC_ALWAYS_BCD)
		CMOS_WRITE(CMOS_READ(RTC_CONTROL) & ~RTC_DM_BINARY, RTC_CONTROL);

	return 0;

err:
	platform_device_put(pdev);

	return ret;
}

device_initcall(malta_platform_rtc_add);
