/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 *
 * Copyright IBM Corp. 2008
 *
 * Authors:
 * 	Hollis Blanchard <hollisb@us.ibm.com>
 * 	Christian Ehrhardt <ehrhardt@de.ibm.com>
 */

#ifndef __POWERPC_KVM_PARA_H__
#define __POWERPC_KVM_PARA_H__

#ifdef __KERNEL__

#include <linux/of.h>

#define KVM_HYPERCALL_BIN 0x03ffffff

#define KVM_HCALL_RESERVE_MAGICPAGE	0

#define KVM_PVDATA_MAGICPAGE_SIZE	"data,pv-magicpage-size"

/* List of PV features supported, returned as a bitfield */
#define KVM_FEATURE_PPCPV_MAGICPAGE	0

static struct kvmppc_para_features {
	char *dtcell;
	int feature;
} para_features[] = {
	{ "feature,pv-magicpage", KVM_FEATURE_PPCPV_MAGICPAGE }
};

#define KVM_HYPERCALL_BIN 0x03ffffff

static inline int kvm_para_available(void)
{
	struct device_node *dn;

	dn = of_find_node_by_path("/hypervisor");

	return !!dn;
}

static inline unsigned int kvm_arch_para_features(void)
{
	struct device_node *dn;
	const int *dtval;
	unsigned int features = 0;
	int i;

	dn = of_find_node_by_path("/hypervisor");
	if (!dn)
		return 0;

	for (i = 0; i < ARRAY_SIZE(para_features)-1; i++) {
		dtval = of_get_property(dn, para_features[i].dtcell, NULL);
		if (dtval && *dtval == 1)
			features |= (1 << para_features[i].feature);
	}

	return features;
}

/* reads the specified data field out of the hypervisor node */
static inline int kvmppc_pv_read_data(char *dtcell)
{
	struct device_node *dn;
	const int *dtval;

	dn = of_find_node_by_path("/hypervisor");
	if (!dn)
		return -EINVAL;

	dtval = of_get_property(dn, dtcell, NULL);
	of_node_put(dn);
	if (dtval)
		return *dtval;
	else
		return -EINVAL;
}

void kvm_guest_init(void);

static inline long kvm_hypercall1(unsigned int nr, unsigned long p1)
{
	register unsigned long hcall asm ("r0") = nr;
	register unsigned long arg1 asm ("r3") = p1;
	register long ret asm ("r11");

	asm volatile(".long %1"
			: "=r"(ret)
			: "i"(KVM_HYPERCALL_BIN), "r"(hcall), "r"(arg1)
			: "r4", "r5", "r6", "r7", "r8",
			  "r9", "r10", "r12", "cc");
	return ret;
}

static inline long kvm_hypercall2(unsigned int nr,
					unsigned long p1, unsigned long p2)
{
	register unsigned long hcall asm ("r0") = nr;
	register unsigned long arg1 asm ("r3") = p1;
	register unsigned long arg2 asm ("r4") = p2;
	register long ret asm ("r11");

	asm volatile(".long %1"
			: "=r"(ret)
			: "i"(KVM_HYPERCALL_BIN), "r"(hcall),
				"r"(arg1), "r"(arg2)
			: "r5", "r6", "r7", "r8",
			  "r9", "r10", "r12", "cc");
	return ret;
}

#endif /* __KERNEL__ */

#endif /* __POWERPC_KVM_PARA_H__ */
