/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 *
 * Copyright IBM Corp. 2008
 *
 * Authors:
 * 	Hollis Blanchard <hollisb@us.ibm.com>
 * 	Christian Ehrhardt <ehrhardt@de.ibm.com>
 */

#include <linux/percpu.h>
#include <linux/mm.h>
#include <linux/kvm_para.h>
#include <linux/bootmem.h>
#include <asm/fixmap.h>

/*
 * this is guest memory granted to the hypervisor;
 * the hypervisor can place data in this area and rewrite
 * privileged instructions to read from this area without
 * trapping.
 * Only the Hypervisor needs to be aware of the structure layout
 * which makes the guest more felxible - the guest only guarantees
 * the size which is requested by the hypervisor and read from a
 * device tree entry.
 */
static void *kvm_magicpage;

static void __init kvmppc_register_magic_page(void)
{
	unsigned long gvaddr;
	unsigned long gpaddr;
	int size;
	long err;

	size = kvmppc_pv_read_data(KVM_PVDATA_MAGICPAGE_SIZE);
	if (size < 0) {
		printk(KERN_ERR "%s: couldn't read size for kvmppc style "
			"paravirtualization support (got %d)\n",
			__func__, size);
		return;
	}

	/* FIXME Guest SMP needs that percpu which */
	kvm_magicpage = alloc_bootmem(size);
	if (!kvm_magicpage) {
		printk(KERN_ERR "%s - failed to allocate %d bytes\n",
			 __func__, size);
		return;
	}
	gpaddr = (unsigned long)__pa(kvm_magicpage);
	gvaddr = fix_to_virt(FIX_KVM_PV);

	err = kvm_hypercall2(KVM_HCALL_RESERVE_MAGICPAGE, gvaddr, gpaddr);
	if (err)
		printk(KERN_ERR "%s: couldn't register pv mem\n", __func__);
	else
		printk(KERN_NOTICE "%s: registered %d bytes for pv mem support"
			" (gvaddr 0x%08lx gpaddr 0x%08lx)\n",
			 __func__, size, gvaddr, gpaddr);
}

void __init kvm_guest_init(void)
{
	if (!kvm_para_available())
		return;

	if (kvm_para_has_feature(KVM_FEATURE_PPCPV_MAGICPAGE))
		kvmppc_register_magic_page();
}
