

#ifndef MP200_DTV_H
#define MP200_DTV_H

#include <mach/pmu.h>
#include <mach/smu.h>
#include <mach/dtv_ioctl.h>



/* DTV_DMACONT */
#define DMACNT_BUS_SERIAL         0x00000004 /* serial bus */
#define DMACNT_BUS_PARALLEL       0x00000000 /* parallel bus */
#define DMACNT_PKT_WITH_PARITY    0x00000002 /* 204bytes */
#define DMACNT_PKT_WITHOUT_PARITY 0x00000000 /* 188bytes(default) */
#define DMACNT_ENDIAN_LITTLE      0x00000001 /* little endian(default) */
#define DMACNT_ENDIAN_BIG         0x00000000 /* big endian */
#define DMACNT_DEFAULT (DMACNT_BUS_SERIAL | \
						DMACNT_PKT_WITHOUT_PARITY | DMACNT_ENDIAN_LITTLE)
/* DTV_DMAREQ / DTV_DMASTOP */
#define DMAREQ_DTV               0x00000001
/* DTV_MODCONT */
#define MODULECONT_HW_RSTZ       0x00000001
/* DTV_STATUS / DTV_RAWSTATUS / DTV_ENSET / DTV_ENCLR / DTV_FFCLR */


#define DTV_INT_LONGPACKET       0x00000040
#define DTV_INT_SHORTPACKET      0x00000020
#define DTV_INT_SYNCBYTE         0x00000010
#define DTV_INT_OVERRUN          0x00000004
#define DTV_INT_DMA              0x00000002
#define DTV_INT_DMAERR           0x00000001
#define DTV_INT_ALL            (DTV_INT_OVERRUN|DTV_INT_DMA|DTV_INT_DMAERR| \
								DTV_INT_LONGPACKET|DTV_INT_SHORTPACKET|DTV_INT_SYNCBYTE)
/* DTV_ERRORADR */
#define ERRORADR_LOCK            0x00000001
#define ERRORADR_MASK            0xFFFFFFFC
/* .pktsiz */
#define PKT_WITH_PARITY          204
#define PKT_WITHOUT_PARITY       188
/* .active */
#define ACT_STOP        0x00 /* stop complete */
#define ACT_ACTIVE      0x01 /* running */
#define ACT_STOP_REQ    0x02 /* during a stop request */
#define ACT_STOP_ERR    0x03 /* error stop */
/* .errsts */
#define ERRSTS_DTVOR    0x01 /* DTV FIFO Overrun */
#define ERRSTS_DMAERR   0x02 /* DMA Error */
#define ERRSTS_BUFOR    0x04 /* DMA Buffer Overrun */
#define ERRSTS_TIMEOUT  0x08 /* DMA Stop Timeout */
#define ERRSTS_DTVSYNC  0x10
#define ERRSTS_DTVSP    0x20
#define ERRSTS_DTVLP    0x40

/* .iunit */
#define DEFAULT_IUNIT     8
#define DEFAULT_PKTSIZ    PKT_WITHOUT_PARITY /* 188byte */
/* .nb_clstrs */
#define NBCLSTRS_UNIT     4 /* Cluster Unit */
#define NBCLSTRS_MIN      4 /* Minimum Number of cluster */
/* usr flag */
#define SETFLG_STOP 0x01
#define SETFLG_RST  0x02

/* DTV Registers */
#define DTV_ADDR	IO_ADDRESS (MP200_DTV_BASE)
#define DTV_REG(OFF)	\
	(*(unsigned int volatile *)((unsigned char *)(DTV_ADDR) + OFF))

#define REG_DTV_STATUS			DTV_REG(0x0000)
#define REG_DTV_RAWSTATUS		DTV_REG(0x0004)
#define REG_DTV_ENSET			DTV_REG(0x0008)
#define REG_DTV_ENCLR			DTV_REG(0x000C)
#define REG_DTV_FFCLR			DTV_REG(0x0010)
#define REG_DTV_ERRORADR		DTV_REG(0x0014)
#define REG_DTV_DMACNT			DTV_REG(0x0020)
#define REG_DTV_DMAREQ			DTV_REG(0x0024)
#define REG_DTV_DMASTOP			DTV_REG(0x0028)
#define REG_DTV_START			DTV_REG(0x002C)
#define REG_DTV_BUFSIZE			DTV_REG(0x0030)
#define REG_DTV_BLANK			DTV_REG(0x0034)
#define REG_DTV_CURRENT			DTV_REG(0x0038)
#define REG_DTV_INTCONT			DTV_REG(0x003C)
#define REG_DTV_MODULECONT		DTV_REG(0x0040)
#define REG_DTV_SIGNALINVERT	DTV_REG(0x0044)
#define REG_DTV_MONITOR			DTV_REG(0x0048)
#define REG_DTV_SWITCH			DTV_REG(0x0200)


union dtv_dmacnt {
	u_int dmacnt;
	struct {
		u_char endian : 1; /* bit 0 */
		u_char mode : 1;   /* bit 1 */
		u_char sp : 1;     /* bit 2 */
		u_int: 29;       /* bit 31-4 */
	} reg;
};

union dtv_signalinvert {
	u_int signalinvert;
	struct {
		u_char vld : 1;   /* bit 0 */
		u_char psync : 1; /* bit 1 */
		u_int: 30;      /* bit 31-2 */
	} reg;
};


/* monitor */
struct dtv_monitor {
	u_int intcnt;      /* interrupt count */
	u_int dmaerr_addr; /* dma error address */
	u_int dmaerr_cnt;  /* dma error count */
	u_int dtvor_cnt;   /* dtv fifo overrun count */
	u_int bufor_cnt;   /* buffer overrun count */
	u_int dtvsync_cnt; /* invalid SyncByte count */
	u_int dtvsp_cnt;   /* short packet count */
	u_int dtvlp_cnt;   /* long packet count */
};



struct dtv_state {
	u_int usr_head;                 /* user index */
	u_int dma_head;                 /* dma index */
	u_int use_clstrs;               /* used cluster number */
	u_int errsts;                   /* error status */
	int active;                     /* dma status */

	int req;                        /* dma request flag */
	u_int set_flg;                  /* user flag */

	int iunit;                      /* input unit size */
	u_int pktsiz;                   /* packet size */
	u_int nb_clstrs;                /* cluster number */
	u_int blank;                    /* blank size */
	u_int dtv_switch;               /* module switch */

	spinlock_t lock;                /* lock */
	wait_queue_head_t wq;           /* wait queue */
	struct semaphore sem;           /* semaphore */

	union dtv_dmacnt dcnt;          /* cont register info */
	union dtv_signalinvert siginv;  /* signal polarity info */
	struct dtv_monitor mt;          /* monitor info */
};


#endif	/* MP200_DTV_H */
