/*
 *  linux/drivers/input/touchscreen/mp200_max7324.c
 *
 *  Copyright (C) NEC Electronics Corporation 2009
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 */

#include <linux/delay.h>
#include <linux/device.h>
#include <linux/input.h>
#include <linux/interrupt.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/i2c.h>

#include <asm/mach/irq.h>

#include "mp200_max7324.h"

/* #define BUTTON_DEBUG */

static const int irq_pwc_gpio_key[] = {
	INT_PWC_GPIO_KEY_DATA0, INT_PWC_GPIO_KEY_DATA1,
	INT_PWC_GPIO_KEY_DATA2, INT_PWC_GPIO_KEY_DATA3
};

/**** Global parameters ****/
static unsigned int scan_delay = BTN_CHAT_GUARDTIME;
static unsigned int scan_interval = BTN_SCAN_INTERVAL;

static struct input_dev *mp200_max7324_dev;

/* work queue */
static struct workqueue_struct *mp200_max7324_workqueue;
static struct delayed_work mp200_max7324_work;

static int mp200_max7324_initialized = 0;

static unsigned long last_bcode = 0;	/* store last button matrix state */
static unsigned long black_bcode = 0;	/* the button should be abandoned */

/* Translation table from the got data to keycode */
static struct mp200_btn_table mp200_max7324_code_tbl[] = {
	/* KEY_OUT0 */
	{.scancode = KEY_F4, .btn_id_bit = (1 << 0), },	/* SW4 */
	{.scancode = KEY_F8, .btn_id_bit = (1 << 1), },	/* SW8 */
	{.scancode = KEY_F11, .btn_id_bit = (1 << 2), },	/* SW11 */
	{.scancode = KEY_F14, .btn_id_bit = (1 << 4), },	/* SW14 */

	/* KEY_OUT1 */
	{.scancode = KEY_F3, .btn_id_bit = (1 << 8), },	/* SW3 */
	{.scancode = KEY_F7, .btn_id_bit = (1 << 9), },	/* SW7 */
	{.scancode = KEY_F19, .btn_id_bit = (1 << 10), },	/* SW10 */
	{.scancode = KEY_F13, .btn_id_bit = (1 << 12), },	/* SW13 */

	/* KEY_OUT2 */
	{.scancode = KEY_F2, .btn_id_bit = (1 << 16), },	/* SW2 */
	{.scancode = KEY_F6, .btn_id_bit = (1 << 17), },	/* SW6 */
	{.scancode = KEY_F9, .btn_id_bit = (1 << 18), },	/* SW9 */
	{.scancode = KEY_F12, .btn_id_bit = (1 << 20), },	/* SW12 */

	/* KEY_OUT3 */
	{.scancode = KEY_F1, .btn_id_bit = (1 << 26), },	/* SW1 */
	{.scancode = KEY_F5, .btn_id_bit = (1 << 28), },	/* SW5 */

	{.scancode = 0, .btn_id_bit = 0, },
};

/*
 * mp200_max7324_scan_core
 *    scan button matrix, and judge valid pressing
 */
static void mp200_max7324_scan_core(struct work_struct *work)
{
	int btn_push_cnt = 0;
	unsigned long chg_bcode;
	unsigned long tmp_bcode;
	unsigned long btncode = 0;
	unsigned long val1 = 0, val2 = 0, val3 = 0, val4 = 0;
	struct input_dev *btn_dev = mp200_max7324_dev;

	struct mp200_btn_table *bst;

#ifdef BUTTON_DEBUG
	printk(KERN_INFO "%s(): time=%ld\n", __FUNCTION__, jiffies);
#endif

#define I2C_MAX7324_DELAY_TIME 64

	extio_write(EXTIO_OUT, 0x01, 0x0f);
	udelay(I2C_MAX7324_DELAY_TIME);
	pwc_read(DA9052_STATUSD_REG , (unsigned int *)&val1);

	extio_write(EXTIO_OUT, 0x02, 0x0f);
	udelay(I2C_MAX7324_DELAY_TIME);
	pwc_read(DA9052_STATUSD_REG , (unsigned int *)&val2);

	extio_write(EXTIO_OUT, 0x04, 0x0f);
	udelay(I2C_MAX7324_DELAY_TIME);
	pwc_read(DA9052_STATUSD_REG , (unsigned int *)&val3);

	extio_write(EXTIO_OUT, 0x08, 0x0f);
	udelay(I2C_MAX7324_DELAY_TIME);
	pwc_read(DA9052_STATUSD_REG , (unsigned int *)&val4);

	btncode = ((val4 << 24) | (val3 << 16) | (val2 << 8) | val1) & BTN_DATA_MASK;

#ifdef BUTTON_DEBUG
	printk(KERN_INFO "%s(): button bit =0x%08lx\n", __FUNCTION__, btncode);
#endif

	/* check out pressed button number */
	bst = mp200_max7324_code_tbl;
	tmp_bcode = btncode;

	while ((bst->btn_id_bit != 0) && (tmp_bcode != 0)) {
		if ((btncode & bst->btn_id_bit) != 0) {
			btn_push_cnt++;
			tmp_bcode &= ~(bst->btn_id_bit);
		}
		bst++;
	}

	chg_bcode = last_bcode ^ btncode;

	/* if total pressed button number is invalid, ignore it */
	if (btn_push_cnt > 1) {
		tmp_bcode = (chg_bcode & btncode);
		black_bcode |= (tmp_bcode & ~last_bcode);
	}

	/* which button released ? */
	tmp_bcode = (chg_bcode & last_bcode);
	if ((tmp_bcode & ~black_bcode) != 0) {
		bst = mp200_max7324_code_tbl;
		while (bst->btn_id_bit != 0) {
			if ((tmp_bcode & bst->btn_id_bit) != 0) {	/* check out which button */
				/* button release */
				input_report_key(btn_dev, bst->scancode, 0);
#ifdef BUTTON_DEBUG
				printk(KERN_INFO "%s(): RELEASE %d time=%ld\n",
				       __FUNCTION__, bst->scancode, jiffies);
#endif
				break;
			}
			bst++;
		}
	}

	/* which button pressed ? */
	tmp_bcode = (chg_bcode & btncode);
	if ((tmp_bcode & ~black_bcode) != 0) {
		bst = mp200_max7324_code_tbl;
		while (bst->btn_id_bit != 0) {
			if ((tmp_bcode & bst->btn_id_bit) != 0) {	/* check out which button */
				/* button press */
				input_report_key(btn_dev, bst->scancode, 1);
#ifdef BUTTON_DEBUG
				printk(KERN_INFO "%s(): PRESS %d time=%ld\n",
				       __FUNCTION__, bst->scancode, jiffies);
#endif
				break;
			}
			bst++;
		}
	}

	/* make a record for legal button */
	last_bcode = btncode & ~black_bcode;

	/* try to clear black bit */
	black_bcode &= btncode;

	if (btncode != 0) {
		queue_delayed_work(mp200_max7324_workqueue, &mp200_max7324_work, scan_interval);
	} else {
		/* button int unmask */
		pwc_write(DA9052_CONTROLC_REG, 0x04, 0x1C);	/* DEBOUNCING time 10msec */
		extio_write(EXTIO_OUT, 0xf, 0x0f);	/* button scan all on */

		/* event clear in enable_irq(), when set_irq_type(IRQ_TYPE_LEVEL_HIGH or IRQ_TYPE_LEVEL_LOW) */
		/* pwc_write(DA9052_EVENTD_REG, 0x17, 0x17); */

		enable_irq(INT_PWC_GPIO_KEY_DATA0);
		enable_irq(INT_PWC_GPIO_KEY_DATA1);
		enable_irq(INT_PWC_GPIO_KEY_DATA2);
		enable_irq(INT_PWC_GPIO_KEY_DATA3);
	}

	return;
}

/*
 * mp200_max7324_interrupt
 *     button interrupt handler for all buttons
 */
static irqreturn_t
mp200_max7324_interrupt(int irq, void *dev_id)
{
#ifdef BUTTON_DEBUG
	printk(KERN_INFO "%s(): irq %d, time=%ld\n",
					__FUNCTION__, irq - INT_PWC_BASE,  jiffies);
#endif

	/* button int unmask */
	disable_irq(INT_PWC_GPIO_KEY_DATA0);
	disable_irq(INT_PWC_GPIO_KEY_DATA1);
	disable_irq(INT_PWC_GPIO_KEY_DATA2);
	disable_irq(INT_PWC_GPIO_KEY_DATA3);

	pwc_write(DA9052_CONTROLC_REG, 0x00,  0x1C);	/* DEBOUNCING time 0msec */

	queue_delayed_work(mp200_max7324_workqueue, &mp200_max7324_work, scan_delay);

	return IRQ_HANDLED;
}


static int mp200_max7324_start(void)
{
	int err;
	struct mp200_btn_table *bst;
	int irq = 0;

	/* input device & irq init */
	mp200_max7324_dev = input_allocate_device();
	if (!mp200_max7324_dev) {
		printk(KERN_ERR "mp200_max7324.c: Not enough memory for input device\n");
		return -ENOMEM;
	}

	/* work queue */
	INIT_DELAYED_WORK(&mp200_max7324_work, mp200_max7324_scan_core);

	/* setup input device */
	set_bit(EV_KEY, mp200_max7324_dev->evbit);
	set_bit(EV_REP, mp200_max7324_dev->evbit);

	bst = mp200_max7324_code_tbl;
	while (bst->btn_id_bit != 0) {
		set_bit(bst->scancode, mp200_max7324_dev->keybit);
		bst++;
	}

	mp200_max7324_dev->name = MP200_BTN_NAME;

	err = input_register_device(mp200_max7324_dev);
	if (err < 0) {
		printk(KERN_ERR "Unable to register mp200_max7324 input device\n");
		goto out_input_free_device;
	}

	gpio_set_value(GPIO_EXT1_P12, 1);

	pwc_write(DA9052_CONTROLC_REG, 0x04, 0x1C);	/* DEBOUNCING time 10msec */

	extio_write(EXTIO_OUT, 0x0f, 0x0f);	/* button scan all on */

	pwc_write(DA9052_EVENTD_REG, 0x17, 0x17);	/* key interrupt clear */

	for (; irq < (sizeof(irq_pwc_gpio_key) / sizeof(irq_pwc_gpio_key[0])); irq++) {
		/* set interrupt high level */
		set_irq_type(irq_pwc_gpio_key[irq], IRQ_TYPE_LEVEL_HIGH);

		/* request interrupt handler */
		err = request_irq(irq_pwc_gpio_key[irq], mp200_max7324_interrupt, IRQF_DISABLED, MP200_BTN_NAME, NULL);
		if (err != 0) {
			printk(KERN_INFO
			       "%s(): reuest_irq(%#x, %p, IRQF_DISABLED, %s, 0) failed (%d)\n",
			       __FUNCTION__, irq_pwc_gpio_key[irq], mp200_max7324_interrupt, MP200_BTN_NAME, err);
			goto out_free_irq;
		}
	}

	mp200_max7324_initialized = 1;

	return 0;

out_free_irq:
	for (irq--; irq >= 0; irq--) {
		free_irq(irq_pwc_gpio_key[irq], 0);
	}
	input_unregister_device(mp200_max7324_dev);

out_input_free_device:
	input_free_device(mp200_max7324_dev);

	return err;
}

static int wait_count = 0;
static void mp200_max7324_start_wait(struct work_struct *work)
{
	int ret;

	if (mp200_extio_initialized == 0) {
		if (wait_count++ < 10) {
			queue_delayed_work(mp200_max7324_workqueue, &mp200_max7324_work, 10);
		}
	} else {
		ret = mp200_max7324_start();
	}
}


/*
 * mp200_max7324_init
 */
static int __init mp200_max7324_init(void)
{
	int err;

	printk(KERN_INFO "Starting button.\n");

	/* work queue */
	mp200_max7324_workqueue = create_singlethread_workqueue(MP200_BTN_NAME);

	if (mp200_extio_initialized == 0) {
		INIT_DELAYED_WORK(&mp200_max7324_work, mp200_max7324_start_wait);
		queue_delayed_work(mp200_max7324_workqueue, &mp200_max7324_work, 10);
	} else {
		err = mp200_max7324_start();
		if (err != 0) {
			return err;
		}
	}
	/* set scan delay */
	scan_delay = scan_delay * HZ / 1000;
	if (scan_delay < 1) {
		scan_delay = BTN_CHAT_GUARDTIME * HZ / 1000;
	}

	/* set scan interval */
	scan_interval = scan_interval * HZ / 1000;
	if (scan_interval < 1) {
		scan_interval = BTN_SCAN_INTERVAL * HZ / 1000;
	}

	return 0;
}

/*
 * mp200_max7324_exit
 */
static void __exit mp200_max7324_exit(void)
{
	int irq;

	extio_write(EXTIO_OUT, 0, 0x0f);	/* button scan all off */

	if (mp200_max7324_initialized) {
		for (irq = 0; irq < (sizeof(irq_pwc_gpio_key) / sizeof(irq_pwc_gpio_key[0])); irq++) {
			free_irq(irq_pwc_gpio_key[irq], 0);
		}
		input_unregister_device(mp200_max7324_dev);
		input_free_device(mp200_max7324_dev);
	}

	destroy_workqueue(mp200_max7324_workqueue);
}

#ifndef MODULE
/*
 * boot command line parameters
 * button=scan_delay[mS],scan_interval[mS]
 */
static int __init mp200_max7324_setup(char *str)
{
	int inkb[3];

	get_options(str, ARRAY_SIZE(inkb), inkb);

	if ((inkb[0] >= 1) && (inkb[1] > 0)) {
		scan_delay = (unsigned int)inkb[1];
	}
	if ((inkb[0] >= 2) && (inkb[2] > 0)) {
		scan_interval = (unsigned int)inkb[2];
	}
	return 0;
}

__setup("button=", mp200_max7324_setup);
#endif

module_init(mp200_max7324_init);
module_exit(mp200_max7324_exit);

module_param(scan_delay, uint, 0644);
MODULE_PARM_DESC(scan_delay, "mp200 button chattering guard time [ms]");

module_param(scan_interval, uint, 0644);
MODULE_PARM_DESC(scan_interval, "mp200 button scan interval time [ms]");

MODULE_LICENSE("GPL");


