/*
 * File Name       : drivers/media/video/mp200/mp200_v4l2_rot.h
 * Function        : Video for Linux driver for EM1 driver core
 * Release Version :
 * Release Date    : 2009.10.21
 *
 * Copyright (C) NEC Electronics Corporation 2009
 *
 *
 * This program is free software;you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by Free Softwere
 * Foundation; either version 2 of License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warrnty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; If not, write to the Free Software Foundation, Inc., 59 Temple
 * Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */

#ifndef MP200_V4L2_IPU__H
#define MP200_V4L2_IPU__H


struct mp200_ipu_req_v4l2;


/*===============================================================*/
/* IPU driver request information structure                      */
/*===============================================================*/
struct mp200_ipu_req_v4l2 {
	/* sequence No */
	unsigned long sequence;
	/* Status of IPU request (used by IPU driver) */
	int status;
	/* callback function */
	ipu_callback_func callback;
	/* optional data (used by caller) */
	unsigned long data;
	int ipu_hw_func;	/* H/W Function (processing demand) */
				/*     IMG Function   : IPU_HW_FUNC_IMG */
				/*     ROT Function   : IPU_HW_FUNC_ROT */
				/*     GDMA Function  : IPU_HW_FUNC_GDMA */
				/*     1PASS Function : IPU_HW_FUNC_1PASS */
				/*     2PASS Function : IPU_HW_FUNC_2PASS */

	/* IPU register setting informations */
	struct mp200_ipu_reg_info ipu_reg_info[MAX_SIZE_MP200_IPU_REQ];
		/* repeat until end condition (max 230 pairs)        */
		/*  end condition :                                  */
		/*     set 0x00000001 to IMG_REQ/ROT_REQ/GDMA_REQ    */
		/*                    (request to start IPU process) */
		/*     reg=IMG_REQ/ROT_REQ/GDMA_REQ data=0x00000001, */
		/*                             or reg=-1(0xFFFFFFFF) */
};


/*******************************************************/
/* request data structure defintion. to IPU-ROT driver */
#define RotRequest(DATA, REG) \
	do { \
		rot_buf->ipu_request.ipu_reg_info[index_req].reg  = REG;  \
		rot_buf->ipu_request.ipu_reg_info[index_req].data = DATA; \
		index_req++; \
	} while (0)


/*******************************************************/
/* request data structure defintion. to IPU-IMG driver */
#define ImgRequest(DATA, REG) \
	do { \
		img_buf->ipu_request.ipu_reg_info[index_req].reg  = REG;  \
		img_buf->ipu_request.ipu_reg_info[index_req].data = DATA; \
		index_req++; \
	} while (0)


/*********************************/
/* ROT register data  <ROT_MODE> */
#define ROT_MODE_SRCENDIAN_BIG    0x00000000
#define ROT_MODE_SRCENDIAN_LITTLE 0x00000010

#define ROT_MODE_DSTENDIAN_BIG    0x00000000
#define ROT_MODE_DSTENDIAN_LITTLE 0x00000008

#define ROT_MODE_SRCYUV_422       0x00000000
#define ROT_MODE_SRCYUV_420       0x00000004

#define ROT_MODE_ROT_MOD_0        0x00000000
#define ROT_MODE_ROT_MOD_90       0x00000001
#define ROT_MODE_ROT_MOD_180      0x00000002
#define ROT_MODE_ROT_MOD_270      0x00000003


#endif /* MP200_V4L2_IPU__H */
