/*
 *  File Name	    : mp200_cam.h
 *  Function	    : CAMERA I/F Driver local header
 *  Release Version : Ver 0.01
 *  Release Date    : 2007/10/03
 *
 *  Copyright (C) NEC Electronics Corporation 2005
 *
 *  This program is free software;you can redistribute it and/or modify it
 *  under the terms of the GNU General Public License as published by Free
 *  Softwere Foundation; either version 2 of License, or (at your option) any
 *  later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY;
 *  without even the implied warrnty of MERCHANTABILITY or FITNESS FOR A
 *  PARTICULAR PURPOSE.
 *  See the GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License along
 *  with this program;
 *  If not, write to the Free Software Foundation, Inc., 59 Temple Place -
 *  Suite 330, Boston, MA 02111-1307, USA.
 *
 */
#ifndef MP200_CAM_H
#define MP200_CAM_H

#include <linux/types.h>
#include <linux/videodev2.h>
#include <mach/mp200_cam.h>
#include <linux/proc_fs.h>
#include <mach/ipu.h>

#if MP200_CAM_NON_PMMU
#else
/*#include <asm/arch/mp200_pmmu.h>*/
#endif

/* ipu externs */
extern int mp200_ipu_regist(struct mp200_ipu_req *p);
extern int mp200_ipu_cancel(unsigned long sequence, int flag);
#if MP200_CAM_NON_PMMU
#else
extern void *mp200_ipu_pmmu_map(int ipu_hw_func, unsigned long size);
extern int mp200_ipu_pmmu_unmap(int ipu_hw_func, void *va, unsigned long size);
#endif

#define MP200_CAM_MAJ_VER       0
#define MP200_CAM_MIN_VER       1
#define MP200_CAM_PATCH_VER     0

struct mp200_cam_prepare {
	__u32 actions:1;
	__u32 reset:1;
	__u32 syncmode:1;        /* CA_CSR SYNCMODE */
	__u32 synctype:1;        /* CA_CSR SYNCTYPE */
	__u32 data_id:1;         /* CA_CSR DATA_ID */
	__u32 vs_det:1;          /* CA_CSR VS_DET */
	__u32 hs_det:1;          /* CA_CSR HS_DET */
	__u32 clk_edge:1;        /* CA_CSR CLK_EDGE */
	__u32 data_det:1;        /* CA_CSR DATA_DET */
	__u32 vs_pol:1;          /* CA_CSR VS_POL */
	__u32 hs_pol:1;          /* CA_CSR HS_POL */
#if 1 /* XXX */
	__u32 width;             /* Image width in pixels. */
	__u32 height;            /* Image height in pixels. */
	struct v4l2_rect c;      /* Cropping rectangle */
#endif
	struct v4l2_rect bounds; /* Defines the window within capturing */
};

struct mp200_cam_hw_operations {
	__u8 name[32];
	int (*vidioc_queryctrl)(struct file *file, void *fh,
				struct v4l2_queryctrl *a);
	int (*vidioc_g_ctrl)(struct file *file, void *fh,
			     struct v4l2_control *a);
	int (*vidioc_s_ctrl)(struct file *file, void *fh,
			     struct v4l2_control *a);
	int (*vidioc_querymenu)(struct file *file, void *fh,
				struct v4l2_querymenu *a);
	int (*prepare)(struct mp200_cam_prepare *);
	int (*trigger)(int);
#if 1 /* XXX */
	int (*sync)(struct mp200_cam_prepare *);
#else
	int (*sync)(int);
#endif
	int (*stream_on)(int);
	int (*stream_off)(int);
	int (*startup)(int);
	int (*shutdown)(int);
	void *private;
	int (*unregister)(int);
};

extern int mp200_cam_hw_register(struct mp200_cam_hw_operations *hw);

#if 0
#define err(format, arg...) printk(KERN_ERR format, ## arg)
#define info(format, arg...) printk(KERN_INFO format, ## arg)
#define warn(format, arg...) printk(KERN_WARNING format, ## arg)
#define emerg(format, arg...) printk(KERN_EMERG format, ## arg)
#define assert(expr) do {} while (0)
#else
#define err(format, arg...) printk(KERN_INFO format, ## arg)
#define info(format, arg...) printk(KERN_INFO format, ## arg)
#define warn(format, arg...) printk(KERN_INFO format, ## arg)
#define emerg(format, arg...) printk(KERN_INFO format, ## arg)
#define assert(expr) \
	if (unlikely(!(expr))) {				 \
		printk(KERN_ERR "Assertion failed! %s,%s,%s,line=%d\n",	  \
		       # expr, __FILE__, __func__, __LINE__);      	   \
	}
#endif

/*** DEBUG code by the making ->*/
#ifdef MP200_CAM_MAKING_DEBUG

#ifdef MP200_CAM_GLOBAL_DEFINE
int debug = 2;
char *ddump;
int ddump_cnt;
int ddump_cnt_max = 750;
int ddump_page = 1;
int entry_cnt;
#else
extern int debug;
extern char *ddump;
static int ddump_cnt_max;
extern int ddump_cnt;
extern int ddump_page;
extern int entry_cnt;
#endif

#define DUMP_MAX (4096 * 4)

#include <linux/moduleparam.h>

#define FNC_ENTRY	\
	if (debug == 1 || debug >= 9) {	\
		char buf[20]; \
		memset(buf, 0, 20); \
		memset(buf, ' ', entry_cnt); \
		entry_cnt++; \
		printk(KERN_INFO "%s((%d:entry )):%s\n", buf, entry_cnt, \
		       __func__); \
	}
#define FNC_EXIT(r)	\
	if (debug == 1 || debug >= 9) {	\
		int i = entry_cnt; \
		char buf[20]; \
		entry_cnt--; \
		memset(buf, 0, 20); \
		memset(buf, ' ', entry_cnt); \
		printk(KERN_INFO "%s((%d:exit:%d)):%s:%d\n", buf, i, \
		       r, __func__, __LINE__); \
	}
#define d0b(fmt, args...)	\
	{ \
		char buf[20]; \
		memset(buf, 0, 20); \
		memset(buf, ' ', entry_cnt + 1); \
		printk(KERN_INFO "%s%s:%d: " fmt, buf, \
		       __func__, __LINE__, ## args); \
	}
#define d1b(fmt, args...)	\
	if (debug == 1 || debug >= 9) {	\
		char buf[20]; \
		memset(buf, 0, 20); \
		memset(buf, ' ', entry_cnt + 1); \
		printk(KERN_INFO "%s%s:%d: " fmt, buf, \
		       __func__, __LINE__, ## args); \
	}
#define d2b(fmt, args...)	\
	if (debug == 2 || debug >= 9) {	\
		char buf[20]; \
		memset(buf, 0, 20); \
		memset(buf, ' ', entry_cnt + 1); \
		printk(KERN_INFO "%s%s:%d: " fmt, buf, \
		       __func__, __LINE__, ## args); \
	}
#define d3b(fmt, args...)	\
	if (debug == 3 || debug >= 9) {	\
		printk(KERN_INFO " --M%d-- " fmt, debug, ## args); \
	}
#define d4b(fmt, args...)	\
	if (debug == 4 || debug >= 9) {	\
		printk(KERN_INFO " --M%d-- " fmt, debug, ## args); \
	}
#define d5b(fmt, args...)	\
	if (debug == 5 || debug >= 9) {	\
		printk(KERN_INFO " --M%d-- " fmt, debug, ## args); \
	}
#define d6b(fmt, args...)	\
	if (debug == 6 || debug >= 9) {	\
		printk(KERN_INFO " --M%d-- " fmt, debug, ## args); \
	}

#define debug_page	\
	{ \
		if (DUMP_MAX <= (ddump_cnt + ddump_cnt_max + 128)) { \
			ddump_cnt = 0; \
			ddump_page++; \
		} \
		ddump_cnt += snprintf(ddump + ddump_cnt, \
				      DUMP_MAX - ddump_cnt,  \
				      "-- p.%d +%d -->\n", ddump_page, \
				      ddump_cnt);		       \
	}

#define debug_snprintf(fmt, args...)	\
	{ \
		debug_page; \
		ddump_cnt += snprintf(ddump + ddump_cnt, \
				      DUMP_MAX - ddump_cnt, fmt, ## args); \
	}

#define debug_dump(func, args...)	\
	{ \
		int i; \
		debug_page; \
		i = func(ddump + ddump_cnt, DUMP_MAX - ddump_cnt, ## args); \
		if (ddump_cnt_max < i) { \
			ddump_cnt_max = i; \
		} \
		ddump_cnt += i;	\
		ddump_cnt += snprintf(ddump + ddump_cnt, \
				      DUMP_MAX - ddump_cnt, "<-- %d --\n", i); \
	}
#else
#define FNC_ENTRY do { } while (0);
#define FNC_EXIT(r) do { } while (0);
#define d0b(fmt, args...) do { } while (0);
#define d1b(fmt, args...) do { } while (0);
#define d2b(fmt, args...) do { } while (0);
#define d3b(fmt, args...) do { } while (0);
#define d4b(fmt, args...) do { } while (0);
#define d5b(fmt, args...) do { } while (0);
#define d6b(fmt, args...) do { } while (0);
#define debug_dump(func, args...) do { } while (0);
#define debug_snprintf(fmt, args...) do { } while (0);
#endif
/*<- end of DEBUG code ***/
#endif

