/***********************************************************************/
/*                                                                     */
/***********************************************************************/
//#include <stdlib.h>

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/cdev.h>
#include <mach/gpio.h>

#include <linux/delay.h>

#include <asm/uaccess.h>
#include <linux/errno.h>
#include <linux/fs.h>

#include <linux/uaccess.h>
#include <linux/proc_fs.h>
#include <mach/pwc.h>

#include <mach/system.h>

static int   devmajor = 11;
static char* devname  = "pocketsvd";

static int device_open_num;

static dev_t pocketsvd_dev;
static struct cdev pocketsvd_cdev;


#define POCKETSVD_PROC_ENTRY	"driver/pocketsvd"

#define MAXBUF 1024
static unsigned char procbuf[ MAXBUF +1 ];
static struct semaphore pocketsvd_proc_sem;

static int		Charging = 0;

//0xFF000000	bit31-24	DIALOG VDD_RES(2.5-4.5V)
//0x00FF0000	bit23-16	DIALOG STATUS_B
//				bit19	CHG_END(0=Charging)
//0x00008000	bit15	Charging(Iavg>=Iend)
//0x00000400	bit10	ICHG result(0=not charging, 1=charging)
//0x00000200	bit9	SD Card(0=removed, 1=inserted)
//0x00000100	bit8	Connect SW(0=ON, 1=OFF)
//0x000000FF	bit0-7	DIALOG STATUS_A
//				bit4	VBUS (0=OFF, 1=DETECT)
//				bit0	POWER SW(0=ON, 1=OFF)

static int pocketsvd_proc_read( char* page, char** start, off_t offset, int count, int* eof, void* data )
{
	unsigned int gpio_num;
	int sw;
	int len = 0;
	unsigned int pwc;
	unsigned int val;
	int ret;

	if(count<8){
		return -EFAULT;
	}

//	printk( KERN_INFO "%s: called\n", __func__ );

	if ( down_interruptible( &pocketsvd_proc_sem ) ) {
		printk( KERN_INFO "%s : down_interruptible for read failed\n", __func__ );
		return -ERESTARTSYS;
	}

	val = 0;

	//Connect SW
	gpio_num = GPIO_P76;				// 0=ON, 1=OFF
	sw = 0x00;
	sw = gpio_get_value(gpio_num);
	val |= 0x00000100 * sw;

	//SD Card
	if( inw( 0xe100041c ) & 0x0020 ){	// 0=removed, 1=inserted
		val |= 0x00000200;
	}

	//ICHGɂ[ď
	if( Charging ){						// 0=not charging, 1=charging
		val |= 0x00000400;
	}

	//DIALOG STATUS_A
	ret = pwc_read(DA9052_STATUSA_REG, &pwc);
	if(ret){
		val = 0xffffffff;
	} else {
		val |= (pwc<<0);
	}

	//DIALOG STATUS_B
	ret = pwc_read(DA9052_STATUSB_REG, &pwc);
	if(ret){
		val = 0xffffffff;
	} else {
		val |= (pwc<<16);
	}

	//DIALOG VDD_RES
	ret = pwc_read(DA9052_VDDRES_REG, &pwc);
	if(ret){
		val = 0xffffffff;
	} else {
		val |= (pwc<<24);
	}

	//[dH
	ret = pwc_read(DA9052_ICHGAV_REG, &pwc);
	if(ret){
		val = 0xffffffff;
	} else {
		if(pwc>=0x0C){
			val |= 0x00008000;
		}
	}

	len += sprintf( page, "%08x\n", val);

	up( &pocketsvd_proc_sem );

	*eof = 1;
	return len;
}

static int pocketsvd_proc_write( struct file* filp, const char* buffer, unsigned long count, void* data )
{
	int copy_len;
	unsigned int val,val2;

	if ( down_interruptible( &pocketsvd_proc_sem ) ) {
		printk( KERN_INFO "%s : down_interruptible for write failed\n", __func__ );
		return -ERESTARTSYS;
	}

	if ( count > MAXBUF )
		copy_len = MAXBUF;
	else
		copy_len = count;

	if ( copy_from_user( &procbuf[0], buffer, copy_len ) ) {
		up( &pocketsvd_proc_sem );
		printk( KERN_INFO "%s : copy_from_user failed\n", __func__ );
		return -EFAULT;
	}

	if( !strncmp(procbuf, "PD", 2) ) {
		printk( "PocketServer system shutdown\n" );

//		pwc_write(DA9052_CONTROLB_REG, 0x80, 0x80);		// SHUTDOWN=1  USB𔲂Ɠd̂ŕs̗p

		pwc_write(DA9052_IRQMASKA_REG, 0x08, 0x08);		// M_VBUS_REM=1
		pwc_write(DA9052_CONTROLA_REG, 0x00, 0x01);		// SYS_EN=0

	}
	else if( !strncmp(procbuf, "RB", 2) ) {
		printk( "PocketServer system reboot\n" );
		arch_reset(0);
	}
	else if( !strncmp(procbuf, "VL:", 3) ) {
		sscanf( &procbuf[3], "%2x", &val );
		if( val >= 0 && val <= 0x1f ) {
			pwc_write( DA9052_CHGCONT_REG, (val<<3), 0xf8 );
		}
	}
	else if( !strncmp(procbuf, "CL:", 3) ) {
		sscanf( &procbuf[3], "%2x", &val );
		if( val >= 0 && val <= 0x3f ) {
			pwc_write( DA9052_BATCHG_REG, val, 0x3f );
		}
	}
	else if( !strncmp(procbuf, "ICHG", 4) ) {

#define	ICHG_AVG_NUM		20

//		int		avgbuf[ICHG_AVG_NUM];
//		int		sum;
		int		i;

		for(i=0; i<ICHG_AVG_NUM; i++){
			int		lcnt;
			pwc_write(DA9052_EVENTB_REG, 0x20, 0x20);
			pwc_reg_write(DA9052_ADCMAN_REG, 0x11);
			lcnt=0;
			do {
				pwc_read(DA9052_ADCMAN_REG, &val);
				lcnt++;
				if(lcnt>2000){	//ł܂邱Ƃ̂ōĎs
					pwc_write(DA9052_EVENTB_REG, 0x20, 0x20);
					pwc_reg_write(DA9052_ADCMAN_REG, 0x11);
					lcnt=0;
				}
			} while (val & 0x10);
			pwc_read(DA9052_ADCRESH_REG, &val);
			pwc_read(DA9052_ADCRESL_REG, &val2);

printk("CHK [%d] :  %03X\n",i, (val<<2) + (val2&0x03));

#if 0
			avgbuf[i] = (val<<2) + (val2&0x03);
#else
			if(val!=0 || (val2&0x03)!=0){
				break;
			}
#endif
			msleep(20);
		}
#if 0
		sum = 0;
		for(i=0;i<ICHG_AVG_NUM;i++){
			sum += avgbuf[i];
		}
		val = sum * 1000 / 1024 / ICHG_AVG_NUM;
		printk( "AVG_VAL  : %d[mA]\n", val );
		if(val<40){
			Charging = 0;	//40mAȉ[dĂȂICHG_AV̒l͖
		} else {
			Charging = 1;	//40mA偨[dĂ遨ICHG_AV
		}
#else
		if(i==ICHG_AVG_NUM){
			printk("charge check : not charging\n");
			Charging = 0;	//I[0[dĂȂICHG_AV̒l͖
		} else {
			printk("charge check : now charging\n");
			Charging = 1;	//I[0łȂ[dĂ遨ICHG_AV
		}
#endif
	}
// Debug
	else if( !strncmp(procbuf, "CHGREG", 6) ) {
		unsigned int	val,reg1,reg2,reg3,reg4;
		struct timeval tv;

		do_gettimeofday(&tv);

		pwc_write(DA9052_EVENTB_REG, 0x20, 0x20);
		pwc_reg_write(DA9052_ADCMAN_REG, 0x11);
		do {
			pwc_read(DA9052_ADCMAN_REG, &val);
		} while (val & 0x10);
		pwc_read(DA9052_ADCRESH_REG, &val);
		pwc_read(DA9052_ADCRESL_REG, &val2);
		val = (val<<2) + (val2&0x03);

		pwc_read(DA9052_STATUSA_REG, &reg1);
		pwc_read(DA9052_STATUSB_REG, &reg2);
		pwc_read(DA9052_ICHGAV_REG, &reg3);
		pwc_read(DA9052_VDDRES_REG, &reg4);

		printk("%d %02X %02X %d %d %d\n",
			(int)(tv.tv_sec*1000+tv.tv_usec/1000),
			reg1,reg2,
			val*1000/1023,reg3*1000/255,
			reg4*2000/255+2500);
	}

	else if( procbuf[0] == 'R' ) {
		int i;
		unsigned int dreg;
		if( sscanf( &procbuf[1], "%d", &dreg ) == 1 ) {
			for( i = 1; procbuf[i] >= '0' && procbuf[i] <= '9'; ++i ) ;

			if( sscanf( &procbuf[i], ":%2x", &val ) == 1 ) { // write
				printk( "Dialog write:[R%d] [%02x]\n", dreg, val );
				pwc_write( dreg, val, 0xff );
			}
			else {	// read
				pwc_read( dreg, &val );
				printk( "Dialog Read:[R%d] [%02x]\n", dreg, val );
			}
		}
	}
// Debug end

	up( &pocketsvd_proc_sem );
	return copy_len;
}

static int pocketsvd_proc_init( void )
{
	struct proc_dir_entry* entry;

	entry = create_proc_entry( POCKETSVD_PROC_ENTRY, 
		       S_IFREG | S_IRUGO | S_IWUGO, NULL );

	if ( entry != NULL ) {
		entry->read_proc  = pocketsvd_proc_read;
		entry->write_proc = pocketsvd_proc_write;
	}
	else {
		printk( KERN_INFO "%s : create_proc_entry failed\n", __func__ );
		return -EBUSY;
	}

	sema_init( &pocketsvd_proc_sem, 1 );
//	printk( KERN_INFO "%s : loaded  into kernel\n", __func__ );

	return 0;
}
static void pocketsvd_proc_cleanup( void )
{
    remove_proc_entry( POCKETSVD_PROC_ENTRY, NULL );
//    printk( KERN_INFO "%s : removed from kernel\n", __func__ );
}


static int pocketsvd_open( struct inode* inode, struct file* filp )
{
//	printk("pocketsvd_open()  called\n" );

	device_open_num++;

	return 0;
}


static int pocketsvd_release( struct inode* inode, struct file* filp )
{
//	printk("pocketsvd_release() called\n" );

	device_open_num--;

	return 0;
}


static ssize_t pocketsvd_read( struct file* filp, char* buf, size_t count, loff_t* pos )
{
	unsigned int gpio_num;
	int sw;

	gpio_num = GPIO_P76;
	sw = 0x00;
	sw = gpio_get_value(gpio_num);

#if 1
	*buf = (char)sw;
#else
	// debug print ver
//	*buf = (char)sw+0x30;
	printk("PushSW2 GPIO_P76 value %x\n",sw);
#endif

	return sizeof(char);
}

#if 0
static ssize_t pocketsvd_write( struct file* filp, const char* buf, size_t count, loff_t* pos )
{
//	ignore
	return 0;
}
#endif

static struct file_operations gpio_fops = 
{
	THIS_MODULE,			// owner
	NULL,					// llseek
	pocketsvd_read,			// read
	NULL,					// write
	NULL,					// readdir
	NULL,					// poll
	NULL,					// ioctl
	NULL,					// mmap
	pocketsvd_open,			// open
	NULL,					// flush
	pocketsvd_release,		// release
	NULL,					// fsync
	NULL,					// fasync
	NULL,					// lock
	NULL,					// readv
	NULL,					// writev
	NULL,					// sendpage
	NULL,					// get_unmapped_area
};


static int __init pocketsvd_init(void)
{
	int ret;

	ret = pocketsvd_proc_init();
	if( ret ) {
		return ret;
	}

	printk("psv driver module loaded\n");

	return 0;
}


static void __exit pocketsvd_exit(void)
{
	if ( device_open_num != 0 ) {
		printk("pocketsv driver module unregister failed\n");
		return;
	}

	remove_proc_entry(POCKETSVD_PROC_ENTRY, NULL);

	printk("psv driver module unregister\n");
}


module_init(pocketsvd_init);
module_exit(pocketsvd_exit);

MODULE_LICENSE("GPL");

