/*
 *  File Name		: linux/drivers/mmc/host/mp200_mmc_sdio.c
 *  Function		: MMC
 *  Release Version : Ver 1.01
 *  Release Date	: 2009/02/04
 *
 *  Copyright (C) NEC Electronics Corporation 2009
 *
 *
 *  This program is free software;you can redistribute it and/or modify it under the terms of
 *  the GNU General Public License as published by Free Softwere Foundation; either version 2
 *  of License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 *  without even the implied warrnty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *  See the GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License along with this program;
 *  If not, write to the Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
 *  MA 02111-1307, USA.
 *
 */

#include <linux/module.h>
#include <linux/init.h>
#include <linux/interrupt.h>
#include <linux/delay.h>
#include <linux/dma-mapping.h>
#include <linux/platform_device.h>
#include <linux/mutex.h>
#include <linux/workqueue.h>
#include <linux/timer.h>
#include <linux/mmc/mmc.h>
#include <linux/mmc/host.h>
#include <linux/mmc/sd.h>
#include <linux/mmc/sdio.h>
#include <linux/mmc/core.h>
#include <linux/mmc/card.h>
#include <linux/io.h>
#include <linux/sched.h>

#include <asm/mach/irq.h>
#include <mach/dma.h>
#include <mach/hardware.h>

#include <mach/smu.h>
#include <mach/pmu.h>
#include <mach/pwc.h>
#include <mach/gpio.h>
//#include <linux/wakelock.h>
#include "../wakelock.h"

#ifdef CONFIG_PM
#include <linux/notifier.h>
#include <linux/pm.h>
#endif

#define MMC_WP_DISABLE

#include "mp200_mmc.h"
/*
#ifndef CONFIG_ENABLE_SUBMICON_COMMUNICATION_DRIVER
#include <mach/subcon_request_if.h>
#else
#include <mach/subcon_if.h>
#endif
*/

void sdio_wake_lock(void);
void sdio_wake_unlock(void);

#define DRIVER_NAME	"mp200_mmc_sdio"

#define MP200_MMC_BASE	(IO_ADDRESS(MP200_SDIB_BASE))
#define MP200_MMC_DMASD		(MP200_SDIB_BASE + 0x300)

#define MMC_CLK_ON      outl((inl(MP200_MMC_CLK_CTRL)|MMC_CLOCK_EN), MP200_MMC_CLK_CTRL);
#define MMC_CLK_OFF	outl((inl(MP200_MMC_CLK_CTRL)&~MMC_CLOCK_EN), MP200_MMC_CLK_CTRL);
#define MMC_GATECLK_ON	mp200_pmu_open_clockgate(MP200_CLOCKGATE_SDIB_CLK);sdio_wake_lock();
#define MMC_GATECLK_OFF	mp200_pmu_close_clockgate(MP200_CLOCKGATE_SDIB_CLK);sdio_wake_unlock();


//#define DEBUG_PRINT
#ifdef DEBUG_PRINT
#define FUNC_PRINT(FMT,ARGS...)		printk(KERN_INFO "%s(): " FMT, __FUNCTION__ , ##ARGS)
#else
#define FUNC_PRINT(FMT,ARGS...)		
#endif

/*#define SDIO_CLK_2M */

#define DEV_OFF      (0)
#define DEV_ON       (1)
#define RESET_OFF    (1)
#define RESET_ON     (0)
#define WLAN_PWR_ON  (0)
#define WLAN_PWR_OFF (1)
#define CARD_DETECT  (1)
#define CARD_REMOVE  (0)
#define SUBCON_PWR_DATA_WLAN         0x04

static spinlock_t intlock;
int clk_on = 1;
int card_present_now = CARD_DETECT;
int card_present_old = CARD_DETECT;
int request_on = 0; /* リクエスト中を示すフラグ */
int clk_off_pending = 0; /* CLK_OFF実行延期を示すフラグ */

int crp_state = DEV_ON;

static struct mp200_mmc_host *host_data;

int powersave_mode	= 0;	/* 省電力モードのレベル		*/
int pre_powersave_mode	= 0;	/* 省電力モードの変更前のレベル	*/
int bus_wake_flag	= 0;	/* バス起床フラグ */
int base_powersave_mode;        /* 省電力モードのシステム設定値  */
struct wake_lock sdio_idle_lock;
int flag_lock = 0;
//////////////////////

EXPORT_SYMBOL(powersave_mode);
EXPORT_SYMBOL(pre_powersave_mode);
EXPORT_SYMBOL(bus_wake_flag);
EXPORT_SYMBOL(base_powersave_mode);

/* 以下のマクロはdhd_sdio.cでも定義している */
#define PS_MODE_LEVEL_0		0	/* 省電力モードOFF        */
#define PS_MODE_LEVEL_1		1	/* クロックレベル1の省電力	*/
#define PS_MODE_LEVEL_2		2	/* クロックレベル2の省電力	*/

void mp200_sdio_card_check(struct mp200_mmc_host *host, int flag);
void mp200_sdio_dev_clock_on(int);
void mp200_sdio_dev_clock_off(int);

//extern TCOM_FUNC   pcom_uart_write( unsigned long dev, TCOM_RW_DATA *data );


static inline u32 mp200_mmc_info_read(void)
{
	u32 info;

FUNC_PRINT(">> mp200_mmc_info_read\n");
	info = inl(MP200_MMC_INFO1) & 0xffff;
	info |= ((inl(MP200_MMC_INFO2) & 0xffff) << 16);

FUNC_PRINT("<< mp200_mmc_info_read\n");
	return info;
}

/*
 * Notify the core about command completion
 */
static void
mp200_mmc_data_done(struct mp200_mmc_host *host, struct mmc_command *cmd)
{
	struct mmc_data *data = host->data;

FUNC_PRINT(">> mp200_mmc_data_done\n");
	outl(MMC_STOP_STOP, MP200_MMC_STOP);
	outl(MMC_CC_EXT_MODE_FIFO, MP200_MMC_CC_EXT_MODE);

	dma_unmap_sg(mmc_dev(host->mmc), data->sg, host->dma_num, host->dma_dir);

	if (!host->data->error) {
		host->data->bytes_xfered = host->data_len;
	}
	host->data_len = 0;

	host->data = NULL;
	host->cmd = NULL;
	host->req = NULL;
	host->stop = NULL;

	/* Clear interrupt */
	outl(0, MP200_MMC_INFO1);
	outl(0, MP200_MMC_INFO2);
	host->info_mask = MMC_INFO_MASK;
	outl(MMC_INFO_MASK & 0xffff, MP200_MMC_INFO1_MASK);
if (!clk_on) {
	MMC_CLK_OFF;
	MMC_GATECLK_OFF;
}
	if (clk_off_pending) {
		FUNC_PRINT("Exec CLK_OFF [L%d] (response received) : clk_off_pending=(0)\n", __LINE__);
		crp_state = DEV_OFF;
		MMC_CLK_OFF;
		MMC_GATECLK_OFF;
		clk_off_pending = 0;
	}
	request_on = 0;

	mmc_request_done(host->mmc, cmd->mrq);
FUNC_PRINT("<< mp200_mmc_data_done\n");
}

/*
 * Notify the core about command completion
 */
static void
mp200_mmc_cmd_done(struct mp200_mmc_host *host, struct mmc_command *cmd)
{
	u32 val;

FUNC_PRINT(">> mp200_mmc_cmd_done\n");
	if (cmd->flags & MMC_RSP_PRESENT) {
		if (cmd->flags & MMC_RSP_136) {
			/* response type 2 */
			val = inl(MP200_MMC_RSP0);
			cmd->resp[3] = val << 8;
			val = inl(MP200_MMC_RSP1);
			cmd->resp[3] |= (val & 0xff) << 24;
			cmd->resp[2] = val >> 8;
			val = inl(MP200_MMC_RSP2);
			cmd->resp[2] |= val << 8;
			val = inl(MP200_MMC_RSP3);
			cmd->resp[2] |= (val & 0xff) << 24;
			cmd->resp[1] = val >> 8;
			val = inl(MP200_MMC_RSP4);
			cmd->resp[1] |= val << 8;
			val = inl(MP200_MMC_RSP5);
			cmd->resp[1] |= (val & 0xff) << 24;
			cmd->resp[0] = val >> 8;
			val = inl(MP200_MMC_RSP6);
			cmd->resp[0] |= val << 8;
			val = inl(MP200_MMC_RSP7);
			cmd->resp[0] |= (val & 0xff) << 24;
			FUNC_PRINT("%s: CMD=%d RSP %x %x %x %x\n", __FUNCTION__, cmd->opcode, cmd->resp[0], cmd->resp[1], cmd->resp[2], cmd->resp[3]);
		} else {
			/* response types 1, 1b, 3, 4, 5, 6 */
			cmd->resp[0] = inl(MP200_MMC_RSP0);
			cmd->resp[0] |= (inl(MP200_MMC_RSP1) << 16);
			FUNC_PRINT("%s: CMD=%d RSP %x\n", __FUNCTION__, cmd->opcode, cmd->resp[0]);
		}
	}

	if( cmd->flags & MMC_RSP_BUSY ) {
		/* wait DATA0 == 1 */
		while( !(mp200_mmc_info_read() & MMC_INFO_DAT0) ) {
		}
	}

	if (host->app_mode) {
		host->app_mode = 0;
	} else if (cmd->opcode == 55) {
		host->app_mode = 1;
	}

	if (host->data == NULL) {
		/* Clear interrupt */
		outl(0, MP200_MMC_INFO1);
		outl(0, MP200_MMC_INFO2);
if (!clk_on) {
		MMC_CLK_OFF;
		MMC_GATECLK_OFF;
}
		host->req = NULL;
		host->cmd = NULL;
		if (clk_off_pending) {
			FUNC_PRINT("Exec CLK_OFF [L%d] (response received) : clk_off_pending=(0)\n", __LINE__);
			crp_state = DEV_OFF;
			MMC_CLK_OFF;
			MMC_GATECLK_OFF;
			clk_off_pending = 0;
		}
		request_on = 0;
		mmc_request_done(host->mmc, cmd->mrq);
	} else if (host->data->error) {
		mp200_mmc_data_done(host, host->cmd);
	} else {
		val = inl(MP200_MMC_INFO1_MASK) | MMC_INFO_EOC;
		host->info_mask = val;
		outl(val, MP200_MMC_INFO1_MASK);
	}
FUNC_PRINT("<< mp200_mmc_cmd_done\n");
}

/*
 * Configure the Responce type
 */
static void
mp200_mmc_start_cmd(struct mp200_mmc_host *host, struct mmc_command *cmd, u16 cmddat)
{
	u32 val;

FUNC_PRINT(">> mp200_mmc_start_cmd\n");
	host->cmd = cmd;

	cmddat |= cmd->opcode;

	if(cmd->opcode == SD_IO_RW_DIRECT || cmd->opcode == SD_IO_RW_EXTENDED)
	{
		cmddat |= MMC_CMD_SDIO;
	}

	switch (mmc_resp_type(cmd)) {
	case MMC_RSP_R1: /* short CRC, OPCODE */
		cmddat |= MMC_CMD_RSP_R1;
		break;
	case MMC_RSP_R1B:/* short CRC, OPCODE, BUSY */
		cmddat |= MMC_CMD_RSP_R1B;
		break;
	case MMC_RSP_R2: /* long 136 bit + CRC */
		cmddat |= MMC_CMD_RSP_R2;
		break;
	case MMC_RSP_R3: /* short */
		cmddat |= MMC_CMD_RSP_R3;
		break;
	}

	if (host->app_mode) {
		cmddat |= MMC_CMD_ACMD;
	}
	
	val = inl(MP200_MMC_STOP) & ~MMC_STOP_STOP;
	outl(val, MP200_MMC_STOP);

	outl(cmd->arg & 0xffff, MP200_MMC_ARG0);
	outl(cmd->arg >> 16,    MP200_MMC_ARG1);

	FUNC_PRINT("cmd=%d arg=%08X flags=%x retries=%x cmddat=%08X\n",
			cmd->opcode, cmd->arg, cmd->flags, cmd->retries, cmddat);

	/* Send command */
	outl(cmddat, MP200_MMC_CMD);

//if ((cmd->opcode != 52) && (cmd-> opcode != 53))
//printk("cmd:%d\n", cmd->opcode);

FUNC_PRINT("<< mp200_mmc_start_cmd\n");
}

/*
 * DMA clean up for command errors
 */
static void mp200_mmc_dma_cleanup(struct mp200_mmc_host *host)
{
	struct mmc_data *data = host->data;

FUNC_PRINT(">> mp200_mmc_dma_cleanup\n");
	if (host->dma_ch != -1) {
		dma_unmap_sg(mmc_dev(host->mmc), data->sg, host->dma_num, host->dma_dir);
		mp200_reset_dma(host->dma_ch);
		host->dma_ch = -1;
	}
FUNC_PRINT("<< mp200_mmc_dma_cleanup\n");
}

/*
 * MMC controller IRQ handler
 */
static irqreturn_t
mp200_mmc_irq(int irq, void *dev_id)
{
	struct mp200_mmc_host *host = (struct mp200_mmc_host *)dev_id;
	u32 info, end_trans = 0;

//	printk("%s: %x\n", __FUNCTION__, info);

	u16 info_sdio;

FUNC_PRINT(">> mp200_mmc_irq\n");
	info_sdio = inl(MP200_SDIO_INFO1);
	if(info_sdio & MMC_SDIO_INFO_IOIRQ)
	{
		outl(0, MP200_SDIO_MODE);
		outl(info_sdio&~MMC_SDIO_INFO_IOIRQ,MP200_SDIO_INFO1);
		mmc_signal_sdio_irq(host->mmc);
	}

	info = mp200_mmc_info_read() & ~host->info_mask;

	if (host->cmd == NULL && host->data == NULL) {
		outl(0, MP200_MMC_INFO1);
		outl(0, MP200_MMC_INFO2);
FUNC_PRINT("<< mp200_mmc_irq1\n");
		return IRQ_HANDLED;
	}

//printk("SDIO_INFO1: %08X\n", info_sdio);
//printk("SD  _INFO1: %08X\n", info);
	if (info & MMC_INFO_ALLERR) {
		end_trans = 1;
		if (info & MMC_INFO_CTO) {
			printk("%s: CMD=%d CMD Timeout(:-)\n", __FUNCTION__, host->cmd->opcode);
			host->cmd->error = -ETIMEDOUT;
		} else if (info & MMC_INFO_CRCERR) {
			printk("%s: CMD=%d CRC error\n", __FUNCTION__, host->cmd->opcode);
			host->cmd->error = -EILSEQ;
		} else {
			printk("%s: CMD=%d Other error.(0x%x)\n", __FUNCTION__, host->cmd->opcode, info);
			host->cmd->error = -EILSEQ;
		}
		if (host->data) {
			if (info & MMC_INFO_DTO) {
				printk("%s: CMD=%d DATA Timeout\n", __FUNCTION__, host->cmd->opcode);
				host->data->error = -ETIMEDOUT;
			} else {
				printk("%s: CMD=%d Other error.(0x%x)\n", __FUNCTION__, host->cmd->opcode, info);
				host->data->error = -EILSEQ;
			}
			mp200_mmc_dma_cleanup(host);
		}
	}

	if ((info & MMC_INFO_EOC) || end_trans) {
		mp200_mmc_cmd_done(host, host->cmd);
FUNC_PRINT("<< mp200_mmc_irq2\n");
		return IRQ_HANDLED;
	}
	if ((info & MMC_INFO_RWEND)) {
		mp200_mmc_data_done(host, host->cmd);
FUNC_PRINT("<< mp200_mmc_irq3\n");
		return IRQ_HANDLED;
	}
FUNC_PRINT("<< mp200_mmc_irq\n");
	return IRQ_HANDLED;
}

static void mp200_mmc_dma_cb(void *data, int intsts, int intrawsts)
{
	struct mp200_mmc_host *host = (struct mp200_mmc_host *)data;
	u32 val;

FUNC_PRINT(">> mp200_mmc_dma_cb\n");
	host->dma_count++;

	if (host->dma_count >= host->dma_num) {
		val = inl(MP200_MMC_INFO1_MASK) & ~MMC_INFO_RWEND;
		host->info_mask = val;
		outl(val, MP200_MMC_INFO1_MASK);
		host->dma_ch = -1;
	} else {
		if (host->dma_ch == host->dma_rx_ch) {
			mp200_start_dma(host->dma_rx_ch, MP200_MMC_DMASD, host->dma_len[host->dma_count],
				host->dma_address[host->dma_count], MP200_DMAC_INT_ERROR_EN | MP200_DMAC_INT_LENG_EN);
		} else {
			mp200_start_dma(host->dma_tx_ch, host->dma_address[host->dma_count], host->dma_len[host->dma_count],
				MP200_MMC_DMASD, MP200_DMAC_INT_ERROR_EN | MP200_DMAC_INT_LENG_EN);
		}
	}
FUNC_PRINT("<< mp200_mmc_dma_cb\n");
}

/*
 * Configure block leangth for MMC/SD cards and intiate the transfer.
 */
static void
mp200_mmc_setup_data(struct mp200_mmc_host *host, struct mmc_data *data)
{
	int i;

FUNC_PRINT(">> mp200_mmc_setup_data\n");
	host->data = data;

	host->data_len = data->blksz * data->blocks;

	FUNC_PRINT("MMC: DATA: size=%d num=%d flags=0x%x\n", data->blksz, data->blocks, data->flags);

	outl(data->blksz,  MP200_MMC_SIZE);
	outl(data->blocks, MP200_MMC_SECCNT);

	host->dma_dir = (data->flags & MMC_DATA_WRITE) ?
				DMA_TO_DEVICE : DMA_FROM_DEVICE;

	host->dma_num = dma_map_sg(mmc_dev(host->mmc), data->sg,
				data->sg_len, host->dma_dir);

	for (i = 0; i < host->dma_num; i++) {
		host->dma_address[i] = sg_dma_address(&data->sg[i]);
		host->dma_len[i] = sg_dma_len(&data->sg[i]);
	}

	outl(MMC_CC_EXT_MODE_SD_DMA, MP200_MMC_CC_EXT_MODE);

	host->dma_count = 0;

	if (data->flags & MMC_DATA_READ) {
		host->dma_ch = host->dma_rx_ch;
		mp200_start_dma(host->dma_rx_ch, MP200_MMC_DMASD, host->dma_len[0],
			host->dma_address[0], MP200_DMAC_INT_ERROR_EN | MP200_DMAC_INT_LENG_EN);
	} else {
		host->dma_ch = host->dma_tx_ch;
		mp200_start_dma(host->dma_tx_ch, host->dma_address[0], host->dma_len[0],
			MP200_MMC_DMASD, MP200_DMAC_INT_ERROR_EN | MP200_DMAC_INT_LENG_EN);
	}

/*
	printk("MMC: dma_num=%d sg_len=%d\n", host->dma_num, data->sg_len);
*/
FUNC_PRINT("<< mp200_mmc_setup_data\n");
}

/*
 * Request function. for read/write operation
 */
static void
mp200_mmc_request(struct mmc_host *mmc, struct mmc_request *req)
{
	struct mp200_mmc_host *host = mmc_priv(mmc);
	u16 cmddat = 0;
	u32 info;

FUNC_PRINT(">> mp200_mmc_request, %d, %d\n", pre_powersave_mode, powersave_mode);
	host->req = req;
	host->stop = req->stop;

	request_on = 1;
	MMC_GATECLK_ON;
	MMC_CLK_ON;
//	if (crp_state == DEV_OFF) {
//		mp200_sdio_initialize(host, 0);
//		crp_state = DEV_ON;
	  
		if ((pre_powersave_mode == PS_MODE_LEVEL_2) ||
		  (powersave_mode == PS_MODE_LEVEL_2)) {
			bus_wake_flag = 1;
		}
//		mp200_sdio_start_clock(host);
//		mp200_sdio_setwidth(host);
//	}

/*gpio_set_value(GPIO_USB_RST, 1);*/
	info = mp200_mmc_info_read();
/*gpio_set_value(GPIO_USB_RST, 0);*/

	/* Clear stat */
	outl(0, MP200_MMC_INFO1);
	outl(0, MP200_MMC_INFO2);
	req->cmd->error = 0;

	if (req->data) {
		req->data->error = 0;
		mp200_mmc_setup_data(host, req->data);

		cmddat |= MMC_CMD_DATA;

		if (host->data->flags & MMC_DATA_READ) {
			cmddat |= MMC_CMD_READ;
		}

		if (host->data->blocks > 1) {
			cmddat |= MMC_CMD_MULTI;
			outl(MMC_STOP_MULTI, MP200_MMC_STOP);
		}
	}

//	printk("cmd=%d arg=%08X flags=%x retries=%x cmddat=%08X\n",
//			cmd->opcode, cmd->arg, cmd->flags, cmd->retries, cmddat);

	mp200_mmc_start_cmd(host, req->cmd, cmddat);
FUNC_PRINT("<< mp200_mmc_request\n");
}

static void mp200_mmc_power(struct mp200_mmc_host *host, u32 power_on)
{
FUNC_PRINT(">> mp200_mmc_power, %d\n", power_on);
	if (power_on) {
		MMC_GATECLK_ON;
		MMC_CLK_ON;

		mp200_pmu_unreset_device(MP200_RESETDEVICE_SDIB);

		outl(MMC_SDWAITCTRL_VAL,     SMU_AB1_SDIBWAITCTRL);
		outl(MMC_SDREADCTRL_VAL,     SMU_AB1_SDIBREADCTRL);
		outl(MMC_USER_VAL,           MP200_MMC_USER);
		outl(MMC_USER2_VAL,          MP200_MMC_USER2);

		outl(MMC_SOFT_RST_RESET,     MP200_MMC_SOFT_RST);
		outl(MMC_SOFT_RST_UNRESET,   MP200_MMC_SOFT_RST);
		outl(MMC_CC_EXT_MODE_FIFO,   MP200_MMC_CC_EXT_MODE);
		outl(MMC_INFO_MASK,          MP200_MMC_INFO1_MASK);
		outl(MMC_INFO_MASK >> 16,    MP200_MMC_INFO2_MASK);

		outl(0, MP200_MMC_INFO1);
		outl(0, MP200_MMC_INFO2);
		outl(0, MP200_SDIO_INFO1);
		outl(MMC_SDIO_INFO1_MASK, MP200_SDIO_INFO1_MASK);
		outl(MMC_STOP_STOP, MP200_MMC_STOP);
		outl(0x80EE, MP200_MMC_OPTION);
		outl(MMC_SDIO_MODE_IOMODE, MP200_SDIO_MODE);
		MMC_CLK_OFF;
		MMC_GATECLK_OFF;
//		enable_irq(host->irq);
	} else {
//		disable_irq(host->irq);
		host->clock = 0;
		host->bus_width = 0;
		MMC_GATECLK_ON;
		MMC_CLK_ON;
		mp200_pmu_reset_device(MP200_RESETDEVICE_SDIB);
		MMC_CLK_OFF;
		MMC_GATECLK_OFF;
	}
FUNC_PRINT("<< mp200_mmc_power\n");
}


/*
 * Configuring clock values.
 */
static void
mp200_mmc_set_ios(struct mmc_host *mmc, struct mmc_ios *ios)
{
	struct mp200_mmc_host *host = mmc_priv(mmc);
	u32 val, wDiv;

FUNC_PRINT(">> mp200_mmc_set_ios\n");
/*
	printk("%s: %d %x %d %d\n", __FUNCTION__,
		ios->clock, ios->power_mode, ios->bus_width, ios->timing);
*/
	if ((host->bus_width != ios->bus_width) && (host->power_mode != MMC_POWER_OFF)) {
		MMC_GATECLK_ON;
		MMC_CLK_ON;
		val = inl(MP200_MMC_OPTION);
		if (ios->bus_width == MMC_BUS_WIDTH_4) {
			host->bus_width = MMC_BUS_WIDTH_4;
			val &= ~MMC_OPTION_WIDTH4;
		} else {
			host->bus_width = MMC_BUS_WIDTH_1;
			val |= MMC_OPTION_WIDTH4;
		}
		outl(val, MP200_MMC_OPTION);
		MMC_CLK_OFF;
		MMC_GATECLK_OFF;
	}

	if ((ios->clock != host->clock) && (host->power_mode != MMC_POWER_OFF)) {
		MMC_GATECLK_ON;
		MMC_CLK_ON;
		/* Stop clock */
		outl(0, MP200_MMC_CLK_CTRL);

		if (ios->clock >= 40000000) {
			wDiv = MMC_CLOCK_HIGH   | MMC_CLOCK_OFFEN;
			outl(MMC_USER2_HIGH_VAL, MP200_MMC_USER2);
		} else if (ios->clock >= 20000000) {
			wDiv = MMC_CLOCK_DIV4   | MMC_CLOCK_OFFEN;
		} else if (ios->clock >= 10000000) {
			wDiv = MMC_CLOCK_DIV8   | MMC_CLOCK_OFFEN;
		} else if (ios->clock >=  5000000) {
			wDiv = MMC_CLOCK_DIV16  | MMC_CLOCK_OFFEN;
		} else if (ios->clock >=  2500000) {
			wDiv = MMC_CLOCK_DIV32  | MMC_CLOCK_OFFEN;
		} else if (ios->clock >=  1250000) {
			wDiv = MMC_CLOCK_DIV64  | MMC_CLOCK_OFFEN;
		} else if (ios->clock >=   625000) {
			wDiv = MMC_CLOCK_DIV128 | MMC_CLOCK_OFFEN;
		} else {
			wDiv = MMC_CLOCK_DIV256;
		}

		wDiv |= MMC_CLOCK_EN;
//		wDiv &= ~MMC_CLOCK_OFFEN;

		/* Start clock */
		outl(wDiv, MP200_MMC_CLK_CTRL);
		MMC_CLK_OFF;
		MMC_GATECLK_OFF;
		host->clock = ios->clock;
	}

	if (host->power_mode != ios->power_mode) {
		switch (ios->power_mode) {
		case MMC_POWER_OFF:
			mp200_mmc_power(host, 0);
			FUNC_PRINT("power off\n");
       		break;
		case MMC_POWER_UP:
			mp200_mmc_power(host, 1);
			FUNC_PRINT("power up\n");
       		break;
		case MMC_POWER_ON:
			FUNC_PRINT("power on\n");
       		break;
		}
		host->power_mode = ios->power_mode;
	}
FUNC_PRINT("<< mp200_mmc_set_ios\n");
}

#ifndef MMC_WP_DISABLE
static int mp200_mmc_get_ro(struct mmc_host *mmc)
{
	int ret;

FUNC_PRINT(">> mp200_mmc_get_ro\n");
	ret = gpio_get_value(GPIO_SDWP);
	dev_dbg(mmc_dev(mmc), "write protect = %d\n", ret);

FUNC_PRINT("<< mp200_mmc_get_ro\n");
	return ret;
}
#endif

static void mp200_enable_sdio_irq(struct mmc_host *host, int enable)
{
	unsigned long flags;

FUNC_PRINT(">> mp200_enable_sdio_irq, %d\n", enable);
	spin_lock_irqsave(&intlock, flags);

MMC_GATECLK_ON;
MMC_CLK_ON;
/*gpio_set_value(GPIO_DTV_RST, 1);*/
	if(enable)
	{
		outl(MMC_SDIO_MODE_IOMODE, MP200_SDIO_MODE);
clk_on = 1;
	}
	else
	{
		outl(0,MP200_SDIO_MODE);
clk_on = 0;
	}
/*gpio_set_value(GPIO_DTV_RST, 0);*/

	spin_unlock_irqrestore(&intlock, flags);
FUNC_PRINT("<< mp200_enable_sdio_irq\n");
}


static struct mmc_host_ops mp200_sdio_ops = {
	.request	= mp200_mmc_request,
	.set_ios	= mp200_mmc_set_ios,
#ifndef MMC_WP_DISABLE
	.get_ro		= mp200_mmc_get_ro,
#endif
	.enable_sdio_irq = mp200_enable_sdio_irq,
};

static int
mp200_mmc_probe(struct platform_device *pdev)
{
	struct mmc_host *mmc;
	struct mp200_mmc_host *host = NULL;
	int ret = 0;

FUNC_PRINT(">> mp200_probe\n");
	mmc = mmc_alloc_host(sizeof(struct mp200_mmc_host), &pdev->dev);
	if (!mmc) {
		ret = -ENOMEM;
		goto err;
	}

	mmc->ops   = &mp200_sdio_ops;
	mmc->f_min = 400000;
#ifdef SDIO_CLK_10M
	mmc->f_max = 10000000;
#else
//	mmc->f_max = 50000000;
	mmc->f_max = 25000000;
#endif
	mmc->ocr_avail = MMC_VDD_28_29;
	mmc->caps = MMC_CAP_4_BIT_DATA |
				MMC_CAP_MMC_HIGHSPEED | MMC_CAP_SD_HIGHSPEED |
				MMC_CAP_SDIO_IRQ;

	/* MMC core transfer sizes tunable parameters */
	mmc->max_hw_segs   = MP200_MAX_SEGS;
	mmc->max_phys_segs = MP200_MAX_SEGS;
	mmc->max_seg_size  = MP200_MAX_SEGS * 512;
	mmc->max_req_size  = MP200_MAX_SEGS * 512;
	mmc->max_blk_size  = 512;
	mmc->max_blk_count = 65535;

	host      = mmc_priv(mmc);
	host->mmc = mmc;
	host->irq = INT_SDIBSD;
	host->power_mode = MMC_POWER_OFF;
	host->info_mask = MMC_INFO_MASK;
	mutex_init(&host->mutex);

	spin_lock_init(&intlock);

	/* DMA setting */
	host->dma_rx_ch = MP200_DMAC_P2M_LCH4;
	host->dma_tx_ch = MP200_DMAC_M2P_LCH4;

	ret = mp200_request_dma(host->dma_rx_ch, DRIVER_NAME,
				mp200_mmc_dma_cb, (void *)host, &host->dma_rx_regs);
	if (ret < 0) {
		goto err;
	}

	ret = mp200_request_dma(host->dma_tx_ch, DRIVER_NAME,
				mp200_mmc_dma_cb, (void *)host, &host->dma_tx_regs);
	if (ret < 0) {
		goto err_dma;
	}

	/* Request IRQ for MMC operations */
	ret = request_irq(host->irq, mp200_mmc_irq, IRQF_DISABLED, pdev->name, host);
	if (ret) {
		FUNC_PRINT("Unable to IRQ");
		goto err_irq;
	}
//	disable_irq(host->irq);

	platform_set_drvdata(pdev, host);
	mmc_add_host(mmc);

#ifdef MV100
#if defined(CONFIG_PANA_BRB) /* -> bootloader */
	/* Rest port for WLAN */
//	gpio_direction_output(GPIO_WLAN_RST, 0);
#endif
#endif

	host_data = host;

FUNC_PRINT("<< mp200_probe\n");
	return 0;

err_irq:
	mp200_free_dma(host->dma_tx_ch);
err_dma:
	mp200_free_dma(host->dma_rx_ch);

err:
	printk("%s: error!!(%d)\n", __FUNCTION__, ret);
	if (host)
		mmc_free_host(mmc);
	return ret;

}

static int
mp200_mmc_remove(struct platform_device *pdev)
{
	struct mp200_mmc_host *host = platform_get_drvdata(pdev);

FUNC_PRINT(">> mp200_remove\n");
	platform_set_drvdata(pdev, NULL);
	if (host) {
		flush_scheduled_work();

		mmc_remove_host(host->mmc);

		mp200_mmc_power(host, 0);
		free_irq(host->irq, host);

		mp200_free_dma(host->dma_rx_ch);
		mp200_free_dma(host->dma_tx_ch);

		mmc_free_host(host->mmc);
	}

FUNC_PRINT("<< mp200_remove\n");
	return 0;
}

static unsigned int mmc_save[IDX_REGS_NUM];

static void mmc_reg_save(void)
{
	mmc_save[IDX_CLK_CTRL] = inl(MP200_MMC_CLK_CTRL);
	mmc_save[IDX_OPTION] = inl(MP200_MMC_OPTION);
	mmc_save[IDX_SDIO_INFO1_MASK] = inl(MP200_SDIO_INFO1_MASK);
	mmc_save[IDX_INFO2_MASK] = inl(MP200_MMC_INFO2_MASK);
	mmc_save[IDX_INFO1_MASK] = inl(MP200_MMC_INFO1_MASK);
	mmc_save[IDX_USER2] = inl(MP200_MMC_USER2);
	mmc_save[IDX_USER] = inl(MP200_MMC_USER);

	return;
}

static void mmc_reg_restore(void)
{
	outl(mmc_save[IDX_USER], MP200_MMC_USER);
	outl(mmc_save[IDX_USER2], MP200_MMC_USER2);

	outl(MMC_SOFT_RST_RESET,     MP200_MMC_SOFT_RST);
	outl(MMC_SOFT_RST_UNRESET,   MP200_MMC_SOFT_RST);
	outl(MMC_CC_EXT_MODE_FIFO,   MP200_MMC_CC_EXT_MODE);

	outl(0, MP200_MMC_INFO1);
	outl(0, MP200_MMC_INFO2);
	outl(0, MP200_SDIO_INFO1);
	outl(mmc_save[IDX_INFO1_MASK], MP200_MMC_INFO1_MASK);
	outl(mmc_save[IDX_INFO2_MASK], MP200_MMC_INFO2_MASK);
	outl(mmc_save[IDX_SDIO_INFO1_MASK], MP200_SDIO_INFO1_MASK);

	outl(mmc_save[IDX_OPTION], MP200_MMC_OPTION);
	outl(mmc_save[IDX_CLK_CTRL], MP200_MMC_CLK_CTRL);

	return;
}

//extern int mmc_remove_done;
static int
mp200_mmc_suspend(struct platform_device *pdev, pm_message_t state)
{
	struct mp200_mmc_host *host = platform_get_drvdata(pdev);

FUNC_PRINT(">> mp200_mmc_suspend\n");
	if (!host || host->suspended) {
FUNC_PRINT("<< mp200_mmc_suspend1\n");
		return 0;
	}
	if (!mp200_sleep_while_idle) {
//printk("mp200_mmc_suspend1\n");
		/* detect interrupt disable */
		disable_irq(INT_SDIBSD);
		mmc_suspend_host(host->mmc, state);
		host->suspended = 1;
	} else //if (mmc_remove_done) {
	  if (host->clock) {
//printk("mp200_mmc_suspend2\n");
		MMC_GATECLK_ON;
		MMC_CLK_ON;
		mmc_reg_save();
		MMC_CLK_OFF;
		MMC_GATECLK_OFF;
		host->suspended = 1;
	} else {
//printk("mp200_mmc_suspend3\n");
	}
//	} else {
//printk("mp200_mmc_suspend4\n");
/*
printk("[CLK_CTRL]        = %08X\n", inl(MP200_MMC_CLK_CTRL));
printk("[OPTION]          = %08X\n", inl(MP200_MMC_OPTION));
printk("[SDIO_INFO1_MASK] = %08X\n", inl(MP200_SDIO_INFO1_MASK));
printk("[MMC_INFO2_MASK]  = %08X\n", inl(MP200_MMC_INFO2_MASK));
printk("[MMC_INFO1_MASK]  = %08X\n", inl(MP200_MMC_INFO1_MASK));
printk("[USER2]           = %08X\n", inl(MP200_MMC_USER2));
printk("[USER]            = %08X\n", inl(MP200_MMC_USER));
printk("[SOFT_RST_RESET]  = %08X\n", inl(MP200_MMC_SOFT_RST));
printk("[SOFT_RST_UNRESET]= %08X\n", inl(MP200_MMC_SOFT_RST));
printk("[CC_EXT_MODE_FIFO]= %08X\n", inl(MP200_MMC_CC_EXT_MODE));

printk("[INFO1]           = %08X\n", inl(MP200_MMC_INFO1));
printk("[INFO2]           = %08X\n", inl(MP200_MMC_INFO2));
printk("[SDIO_INFO1]      = %08X\n", inl(MP200_SDIO_INFO1));
*/
//	}
FUNC_PRINT("<< mp200_mmc_suspend\n");
	return 0;
}

/* Routine to resume the MMC device */
static int
mp200_mmc_resume(struct platform_device *pdev)
{
	struct mp200_mmc_host *host = platform_get_drvdata(pdev);

FUNC_PRINT(">> mp200_mmc_resume\n");
	if (!host || !host->suspended) {
FUNC_PRINT("<< mp200_mmc_resume1\n");
		return 0;
	}
	if (!mp200_sleep_while_idle) {
//printk("mp200_mmc_resume1\n");
		/* detect interrupt enable */

		mmc_resume_host(host->mmc);
		enable_irq(INT_SDIBSD);
	} else //if (mmc_remove_done) {
	  if (host->clock) {
//printk("mp200_mmc_resume2\n");
		MMC_GATECLK_ON;
		MMC_CLK_ON;
		mmc_reg_restore();
		MMC_CLK_OFF;
		MMC_GATECLK_OFF;
	} else {
//printk("mp200_mmc_resume3\n");
	}
//	} else {
//printk("mp200_mmc_resume4\n");
//		MMC_GATECLK_ON;
//		MMC_CLK_ON;
//	}
	host->suspended = 0;

FUNC_PRINT("<< mp200_mmc_resume\n");
	return 0;
}


static struct platform_driver mp200_sdio_driver = {
	.probe		= mp200_mmc_probe,
	.remove		= mp200_mmc_remove,
	.suspend	= mp200_mmc_suspend,
	.resume		= mp200_mmc_resume,
	.driver		= {
			.name = DRIVER_NAME,
	},
};
static int __init mp200_sdio_init(void)
{
FUNC_PRINT(">> mp200_sdio_init\n");
#ifdef MV100
#if defined(CONFIG_PANA_BRB) /* -> bootloader */
	gpio_direction_output(GPIO_WLAN_RST, 0x0);
	gpio_direction_output(GPIO_WLAN_REGON, 0x0);

__raw_writel((__raw_readl(CHG_PINSEL_G80) | 0x01000000), CHG_PINSEL_G80);
#endif/*defined(CONFIG_PANA_BRB)*/

	/* WLAN RESET(L), BT_RESET(L) */
	gpio_set_value(GPIO_WLAN_RST, 0);
	gpio_set_value(GPIO_BT_RST, 0);
	msleep(100);

	/* WLAN REGON(H) */
	gpio_set_value(GPIO_WLAN_REGON, 1);
	msleep(150);

	/* WLAN RESET(H), BT_RESET(H) */
	gpio_set_value(GPIO_WLAN_RST, 1);
	gpio_set_value(GPIO_BT_RST, 1);
	msleep(150);

	/* BT_RESET(L) */
	gpio_set_value(GPIO_BT_RST, 0);
	msleep(100);
	
#else /* MV100 end / PS start */

#define GPIO_WLAN_WAKE			GPIO_P1 // Out
#define GPIO_WLAN_HOST			GPIO_P2 // Out
#define GPIO_WLAN_ENABLE		GPIO_P3 // Out
#define GPIO_BT_ENABLE			GPIO_P4 // Out
#define GPIO_WLAN_CLK_ENABLE	GPIO_P9 // In

	/* WLAN ENABLE(L), BT_ENABLE(L) */
	gpio_set_value(GPIO_WLAN_ENABLE, 0);
	gpio_set_value(GPIO_BT_ENABLE, 0);

__raw_writel((__raw_readl(CHG_PINSEL_G80) | 0x01000000), CHG_PINSEL_G80);

	msleep(100);

	/* WLAN ENABLE(H), BT_ENABLE(H) */
	gpio_set_value(GPIO_WLAN_ENABLE, 1);
//	gpio_set_value(GPIO_BT_ENABLE, 1);
	msleep(150);

#endif /* PS end */

	wake_lock_init(&sdio_idle_lock, WAKE_LOCK_IDLE, "sdio_drv");
	flag_lock = 0;

	/* Register the MMC driver */
	if (platform_driver_register(&mp200_sdio_driver)) {
		printk(KERN_ERR ": Failed to register MMC driver\n");
		wake_lock_destroy(&sdio_idle_lock);
		return -ENODEV;
	}

FUNC_PRINT("<< mp200_sdio_init\n");
	return 0;
}

static void __exit mp200_sdio_cleanup(void)
{
FUNC_PRINT(">> mp200_sdio_cleanup\n");
	/* Unregister MMC driver */
	platform_driver_unregister(&mp200_sdio_driver);
	wake_lock_destroy(&sdio_idle_lock);

	/* K0通信 */
FUNC_PRINT("<< mp200_sdio_cleanup\n");
}

module_init(mp200_sdio_init);
module_exit(mp200_sdio_cleanup);

MODULE_DESCRIPTION("MP200 High Speed Multimedia Card driver");
MODULE_LICENSE("GPL");
MODULE_ALIAS(DRIVER_NAME);
MODULE_AUTHOR("NEC Electronics");


void mp200_sdio_dev_clock_on(int card_det_flag)
{
 FUNC_PRINT(">> mp200_sdio_dev_clock_on, flag:%d, state;%d\n", card_det_flag, crp_state);
	if (crp_state == DEV_OFF) {
		MMC_GATECLK_ON;
		MMC_CLK_ON;
		crp_state = DEV_ON;
		if (card_det_flag) {
			mp200_sdio_card_check(host_data, 1);
		}
	}
FUNC_PRINT("<< mp200_sdio_dev_clock_on\n");
}
EXPORT_SYMBOL(mp200_sdio_dev_clock_on);

void mp200_sdio_dev_clock_off(int card_ejc_flag)
{
  FUNC_PRINT(">> mp200_sdio_dev_clock_off, flag:%d, state:%d\n", card_ejc_flag, crp_state);
	if (crp_state != DEV_OFF) {
		MMC_GATECLK_ON;
		MMC_CLK_ON;
		if (card_ejc_flag) {
			mp200_sdio_card_check(host_data, 2);
		}
		if (request_on) {
			clk_off_pending = 1;
			FUNC_PRINT("Pending CLK_OFF (waiting for response) : clk_off_pending=(1)\n");
		} else {
			crp_state = DEV_OFF;
			MMC_CLK_OFF;
			MMC_GATECLK_OFF;
		}
	}
FUNC_PRINT("<< mp200_sdio_dev_clock_off\n");
}
EXPORT_SYMBOL(mp200_sdio_dev_clock_off);

//extern void mmc_end_rescan(void);
void mp200_sdio_card_check(struct mp200_mmc_host *host, int flag)
{
	unsigned long flags;
//	int count = 0;
FUNC_PRINT(">> mp200_sdio_card_check, flag:%d, state:%d\n", flag, crp_state);
	switch (flag) {
	case 0:
	case 1:
		spin_lock_irqsave(&intlock, flags);
		card_present_now = CARD_DETECT;
		spin_unlock_irqrestore(&intlock, flags);
		if (card_present_old != card_present_now) {
			mmc_detect_change(host->mmc, 0);
		}
	  	break;
	case 2: // 終了時に排出状態に変更する
		spin_lock_irqsave(&intlock, flags);
		card_present_now = CARD_REMOVE;
		spin_unlock_irqrestore(&intlock, flags);
		//mmc_remove_done = 0;
		mmc_detect_change(host->mmc, 0);
/*
#ifdef DEBUG_PRINT
		while (mmc_remove_done != 1) {
			msleep(100);
			count++;
		}
#else
		while ((mmc_remove_done != 1) && (count < 20)) {
			msleep(10);
			count++;
		}
#endif
		printk("### count:%d, %d, %d, %d\n", count, mmc_remove_done, clk_off_pending, crp_state);*/
//mmc_end_rescan();
		break;
	case 3: // insmod時に排出状態に変更する。MMCにはここでは非通知
		spin_lock_irqsave(&intlock, flags);
		card_present_now = CARD_REMOVE;
		spin_unlock_irqrestore(&intlock, flags);
		break;
	default:
		break;
	}

        card_present_old = card_present_now;

FUNC_PRINT("<< mp200_sdio_card_check\n");
}


void mp200_sdio_card_eject(void)
{
	msleep(50);
	crp_state = DEV_OFF;
	mp200_sdio_card_check(host_data, 2);
}
EXPORT_SYMBOL(mp200_sdio_card_eject);

void sdio_wake_lock(void)
{
    if (!flag_lock) {
        wake_lock(&sdio_idle_lock);
        flag_lock = 1;
    }
}

void sdio_wake_unlock(void)
{
    if (flag_lock) {
        wake_unlock(&sdio_idle_lock);
        flag_lock = 0;
    }
}
