/*
 * File Name       : /drivers/video/mp200/mp200_fb.c
 * Function        : FraemBuffer Driver corresponding to V4L2
 * Release Version :
 * Release Date    : 2009.10.21
 *
 * Copyright (C) NEC Electronics Corporation 2009
 *
 *
 * This program is free software;you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by Free Softwere
 * Foundation; either version 2 of License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warrnty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; If not, write to the Free Software Foundation, Inc., 59 Temple
 * Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */

/*
 *
 *  framebuffer driver for D77601 -- based on vfb.c
 *
 */

/*
 *  linux/drivers/video/vfb.c -- Virtual frame buffer device
 *
 *      Copyright (C) 2002 James Simmons
 *
 *	Copyright (C) 1997 Geert Uytterhoeven
 *
 *  This file is subject to the terms and conditions of the GNU General Public
 *  License. See the file COPYING in the main directory of this archive for
 *  more details.
 */


/********************************************************
 *  Include Files                                       *
 *******************************************************/
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/version.h>
#include <linux/errno.h>
#include <linux/string.h>
#include <linux/mm.h>
#include <linux/tty.h>
#include <linux/slab.h>
#include <linux/vmalloc.h>
#include <linux/delay.h>
#include <linux/interrupt.h>
#include <linux/fb.h>
#include <linux/init.h>
#include <linux/ioport.h>
#include <linux/timer.h>
#include <linux/sched.h>

#include <linux/platform_device.h>
#if defined(CONFIG_PM) || defined(CONFIG_DPM)
#include <mach/pm.h>
#endif /* CONFIG_PM || CONFIG_DPM */

#include <linux/uaccess.h>
#include <mach/hardware.h>
#include <mach/irqs.h>
#include <mach/em1_mem.h>
#include <mach/fbcommon.h>

#include "mp200_common.h"
#include "mp200_fb.h"
#include "mp200_imc.h"
#include "mp200_lcd.h"
#include "mp200_lcdhw.h"
#include "nts/mp200_nts.h"



/********************************************************
 *  Definitions                                         *
 *******************************************************/
#define _DEBUG_FB 0x00 /* 00008421(bit) */
		       /* 0x01: debug function in
			* 0x02: debug function out
			* 0x04: debug buffer
			* 0x08: debug frame buffer status
			* 0x10: debug update timer
			*/

#define TIMEOUT_MSLEEP	0


/********************************************************
 *  Definitions                                         *
 *******************************************************/
#define DEV_NAME "mp200_fb"


/********************************************************
 *  Macros                                              *
 *******************************************************/
#define printk_err(fmt, arg...) \
	do {                     \
		printk(KERN_ERR DEV_NAME ": %s: " fmt, __func__, ## arg); \
	} while (0)

#define printk_wrn(fmt, arg...) \
	do {                     \
		printk(KERN_WARNING DEV_NAME ": %s: " fmt, __func__, ## arg); \
	} while (0)

#define printk_info(fmt, arg...) \
	do {                      \
		printk(KERN_INFO DEV_NAME ": " fmt, ## arg); \
	} while (0)

#if _DEBUG_FB
#define printk_dbg(level, fmt, arg...) \
	do {                            \
		if (level > 0) \
			printk(KERN_DEBUG DEV_NAME ": %s: " fmt, __func__, ## arg); \
	} while (0)
#else
#define printk_dbg(level, fmt, arg...) \
	;
#endif


/********************************************************
 *  Definitions                                         *
 *******************************************************/
/* init function */
#define FB_INIT_ERROR__SMEM_NOT_RESERVE		-1
#define FB_INIT_ERROR__SMEM_NOT_REMAP		-2
#define FB_INIT_ERROR__FB_NOT_ALLOC		-3
#define FB_INIT_ERROR__FB_CMAP_NOT_ALLOC	-4
#define FB_INIT_ERROR__FB_NOT_REGIST		-5
#define FB_INIT_ERROR__LCD_NOT_INITIALIZE	-6

/* frame buffer page */
#define FB_NO_A				0
#define FB_NO_B				1

/* frame buffer status */
#define IS_NODATA			0
#define NOW_WRITING_TO_FB		1
#define ALREADY_DISPLAY			2
#define ENABLE_WRITE			3

/* timer */
#define TIMER_OFF			0
#define TIMER_ON			1
#define TIMEOUT_TIME			60 /* (ms) */
#define FB_TIMER_STOP			0
#define FB_TIMER_START			1

/* CHK_SCRN timer */
#define RECHK_TIME			10  /* (ms) */

/* timer default setting */
#define DEFALT_UPDATE_TIMER_LCD		TIMER_OFF
#ifdef CONFIG_FB_MP200_NTS
#define DEFALT_UPDATE_TIMER_NTS		TIMER_ON
#endif /* CONFIG_FB_MP200_NTS */

/* mmap */
#define MP200_FB_MMAP_MAX_SIZE	FB_FRAME_BUFFER_SIZE


/* Number of FB devices */
#ifdef CONFIG_FB_MP200_NTS
#define EM1_FB_DEVICES			2
#else	/* CONFIG_FB_MP200_NTS */
#define EM1_FB_DEVICES			1
#endif	/* CONFIG_FB_MP200_NTS */


#if 0
void ............Structure................() {}
#endif
/********************************************************
 *  Structure                                           *
 *******************************************************/
/* image infomation from 2D API */
typedef struct fb_other_image_info{
	int  iMaskColFlg;
	int  iMaskCol;
	int  iAlpha;
	int  iInvFlg;
	int  iPage;
	int  iUpDate;
} FB_OTHER_IMAGE_INFO;


/* image infomation from 2D API */
typedef struct _mp200_fb_par {
	struct fb_info          *fb_inf;
	atomic_t                 use_count;
	struct semaphore         sem_fb_image_info;	/* semafore for
							   fb_image_info */
	struct semaphore         sem_fb_other_info_A;	/* semafore for
							   fb_other_info_A */
	struct semaphore         sem_fb_other_info_B;	/* semafore for
							   fb_other_info_B */
	struct semaphore         sem_tvmode;	/* semafore for tvmode */
	spinlock_t               fb_lock;	/* spin lock fb timer flag */
#if (TIMEOUT_MSLEEP == 0)
	wait_queue_head_t        wait_fb_chk_scrn;
#endif
	struct workqueue_struct *mp200_fb_workqueue;
	struct work_struct       wk_timeout_update_bottom_half;
	u32                      pseudo_palette[256];
	int                      devflag; /* 0 : /dev/fb0   1 : /dev/fb1  */
} mp200_fb_par;


/* screen infomation default data */
static struct fb_var_screeninfo mp200_fb_default = {
	.xres           = FRONT_WIDTH,
	.yres           = FRONT_HEIGHT,
	.xres_virtual   = FRONT_WIDTH_V,
	.yres_virtual   = FRONT_HEIGHT_V,
	.xoffset        = 0,
	.yoffset        = 0,
	.grayscale      = 0,
	.bits_per_pixel = 16,
	.red            = {11, 5, 0},
	.green          = {5, 6, 0},
	.blue           = {0, 5, 0},
	.transp         = {0, 0, 0},
	.activate       = 0,
	.height         = -1,
	.width          = -1,
	.pixclock       = 0,
	.left_margin    = 0,
	.right_margin   = 0,
	.upper_margin   = 0,
	.lower_margin   = 0,
	.hsync_len      = 24,
	.vsync_len      = 4,
	.sync           = FB_SYNC_HOR_HIGH_ACT | FB_SYNC_VERT_HIGH_ACT,
	.vmode          = FB_VMODE_NONINTERLACED,
};

#ifdef CONFIG_FB_MP200_NTS
/* screen  tv infomation default data */
static struct fb_var_screeninfo mp200_fb_tv_default = {
	.xres           = NTS_WIDTH,
	.yres           = NTS_HEIGHT,
	.xres_virtual   = NTS_WIDTH_V,
	.yres_virtual   = NTS_HEIGHT_V,
	.xoffset        = 0,
	.yoffset        = 0,
	.grayscale      = 0,
	.bits_per_pixel = 16,
	.red            = {11, 5, 0},
	.green          = {5, 6, 0},
	.blue           = {0, 5, 0},
	.transp         = {0, 0, 0},
	.activate       = 0,
	.height         = -1,
	.width          = -1,
	.pixclock       = 0,
	.left_margin    = 0,
	.right_margin   = 0,
	.upper_margin   = 0,
	.lower_margin   = 0,
	.hsync_len      = 24,
	.vsync_len      = 4,
	.sync           = FB_SYNC_HOR_HIGH_ACT | FB_SYNC_VERT_HIGH_ACT,
	.vmode          = FB_VMODE_NONINTERLACED,
};
#endif /* CONFIG_FB_MP200_NTS */


/* screen infomation fix data */
static struct fb_fix_screeninfo mp200_fb_fix __initdata =
{
	.id        = "MP200 FB",
	.type      = FB_TYPE_PACKED_PIXELS,
	.visual    = FB_VISUAL_TRUECOLOR,
	.xpanstep  = 1,
	.ypanstep  = 1,
	.ywrapstep = 1,
	.accel     = FB_ACCEL_NONE,
};

typedef struct mp200_fb_struct{
	int                   is_first;
	/* image infomation to LCD */
	FB_IMAGE_INFO         fb_image_info;
	/* frame buffer A */
	FB_OTHER_IMAGE_INFO   fb_other_info_A;
	/* frame buffer B */
	FB_OTHER_IMAGE_INFO   fb_other_info_B;

	/* Disp Buffer (2D) */
	ulong   DispBufA;
	ulong   DispBufB;
	ulong   DispBufLength;
	char   *DispBufBV;
	char   *DispBufAV;

	/* Timer */
	struct timer_list     mp200_fb_timer;
	int                   iTimeoutTime;
	mp200_fb_par          *mp200_fb_dev;
	/* flags */
	int                   iFBStatus_A;
	int                   iFBStatus_B;
	int                   iNowDispPage;   /* now on display page */
	int                   iNextDispPage;  /* next display page   */
	int                   iMixModeFlg;
	int                   iUpdateTimerFlag;
	int                   iTimerStatusFlag;
	int                   iDpmSuspendFlag;
} MP200_FB_STRUCT;


typedef struct mp200_fb_drvdata{
	struct fb_info *info[2];
} MP200_FB_DRVDATA;


#if 0
void ............Variables................() {}
#endif
/********************************************************
 *  Variables                                           *
 *******************************************************/
static MP200_FB_STRUCT    fb_data[2];
static EM1_TV_MODE        tvmode;
static MP200_FB_DRVDATA   drvdata;
/* Smem */
static  ulong             Smem       = SMEM_START;
static  ulong             SmemLength = SMEM_LENGTH;
	char             *SmemV;
static  ulong             DispBufAllLength = MP200_FB_MMAP_MAX_SIZE;

/********************************************************
 *  Prototype declarations of local function            *
 *******************************************************/
static int mp200_fb_ioc_chk_scrn(struct fb_info *info, unsigned long arg);
static int mp200_fb_ioc_get_modes(struct fb_info *info, unsigned long arg);
static int mp200_fb_ioc_set_modes(struct fb_info *info, unsigned int cmd,
 unsigned long arg);
static int mp200_fb_ioc_update_scrn(struct fb_info *info, unsigned int cmd,
 unsigned long arg);
static int mp200_fb_ioc_update_timer(struct fb_info *info, unsigned long arg);
#ifdef CONFIG_FB_MP200_NTS
static int mp200_fb_ioc_set_output(struct fb_info *info, unsigned long arg);
static int mp200_fb_ioc_get_output(struct fb_info *info, unsigned long arg);
#endif /* CONFIG_FB_MP200_NTS */

static int set_image_info_to_2DAPI(int iFBNo, scrn_modes *fb_scrn_modes,
 struct fb_info *info);
static int chk_fb_status(int iFBStatus);
static int chk_input_modes(scrn_modes fb_scrn_modes);
static int set_image_info_from_2DAPI(int iFBNo, scrn_modes *fb_scrn_modes,
 struct fb_info *info);
static int set_image_info_to_LCD(int iFBNo, struct fb_info *info);
static int chk_input_update_timer(update_timer *fb_timer);
#ifdef CONFIG_FB_MP200_NTS
static void set_tv_size(EM1_TV_MODE mode, mp200_fb_par *par);
static void set_tv_var(EM1_TV_MODE mode, mp200_fb_par *par);
#endif /* CONFIG_FB_MP200_NTS */
static int mp200_fb_open(struct fb_info *info, int user);
static int mp200_fb_release(struct fb_info *info, int user);
static  u_long get_line_length(int xres_virtual, int bpp);
static int mp200_fb_check_var(struct fb_var_screeninfo *var,
 struct fb_info *info);
static int mp200_fb_set_par(struct fb_info *info);
static int mp200_fb_setcolreg(u_int regno, u_int red, u_int green, u_int blue,
 u_int transp, struct fb_info *info);
static int mp200_fb_blank(int blank_mode, struct fb_info *info);
static int mp200_fb_pan_display(struct fb_var_screeninfo *var,
 struct fb_info *info);
static int mp200_fb_remove(struct platform_device *device);
static int  __init mp200_fb_probe(struct platform_device *device);
static void __init mp200_fb_probe_error(int iErrorMode,
 struct fb_info *info_0, struct fb_info *info_1, int fb_num);
static void set_buffer_address(int fb_num);
static void set_info(struct fb_info *info, mp200_fb_par **par, int fb_num);
static void set_flag_init(int fb_num);
static void set_val_init(struct fb_info *info, int fb_num);
static void timer_init(struct fb_info *info);
static void timeout_update(unsigned long data);
#if defined(CONFIG_PM) || defined(CONFIG_DPM)
static int mp200_fb_suspend(struct platform_device *dev, pm_message_t state);
static int mp200_fb_suspend_sub(struct platform_device *dev,
 pm_message_t state, int fb_num);
static int mp200_fb_resume(struct platform_device *dev);
static int mp200_fb_resume_sub(struct platform_device *dev, int fb_num);
#endif /* CONFIG_PM || CONFIG_DPM */

/* workqueue function */
static void timeout_update_bottom_half_do(struct work_struct *num);


/********************************************************
 *  Function Structure                                  *
 *******************************************************/
/* fb ops */
static struct fb_ops mp200_fb_ops = {
	.owner          = THIS_MODULE,
	.fb_open        = mp200_fb_open,
	.fb_release     = mp200_fb_release,
	.fb_check_var   = mp200_fb_check_var,
	.fb_set_par     = mp200_fb_set_par,
	.fb_setcolreg   = mp200_fb_setcolreg,
	.fb_blank       = mp200_fb_blank,
	.fb_pan_display = mp200_fb_pan_display,
	.fb_fillrect    = cfb_fillrect,  /* fb common function */
	.fb_copyarea    = cfb_copyarea,  /* fb common function */
	.fb_imageblit   = cfb_imageblit, /* fb common function */
#ifdef CONFIG_FRAMEBUFFER_CONSOLE
	.fb_cursor      = soft_cursor,   /* fb common function */
#endif /* CONFIG_FRAMEBUFFER_CONSOLE */
	.fb_ioctl       = mp200_fb_ioctl,
	.fb_mmap        = mp200_fb_mmap,
};

/* platform driver */
static struct platform_driver mp200_fb_driver = {
	.probe       = mp200_fb_probe,
	.remove      = mp200_fb_remove,

#if defined(CONFIG_PM) || defined(CONFIG_DPM)
	.suspend     = mp200_fb_suspend,
	.resume      = mp200_fb_resume,
#endif /* CONFIG_PM || CONFIG_DPM */

	.driver.name = DEV_NAME,
	.driver.bus  = &platform_bus_type,
};


#if 0
void ............ioctl................() {}
#endif
/********************************************************
 *  Function Definitions                                *
 *******************************************************/
/******************************************************************************
* MODULE   : mp200_fb_ioctl
* FUNCTION : frame buffer driver ioctl
* RETURN   :  0      : success
*            -1      :
*            -EFAULT :
*            -EBUSY  :
*            -ENXIO  :
*            -ENOIOCTLCMD:
* NOTE     : none
* CREATE   : 2006.05.26 H.Nagashima
* UPDATE   : 2008.11.25 J.Manabe
******************************************************************************/
int mp200_fb_ioctl(struct fb_info *info, unsigned int cmd, unsigned long arg)
{
	switch (cmd) {
	case MP200_FB_CHKSCRN:
		/* check frame buffer status        */
		return mp200_fb_ioc_chk_scrn(info, arg);

	case MP200_FB_GET_MODES:
		/* get frame buffer infomation      */
		return mp200_fb_ioc_get_modes(info, arg);

	case MP200_FB_SET_MODES:
		/* Set or Update frame buffer infomation   */
		return mp200_fb_ioc_set_modes(info, cmd, arg);

	case MP200_FB_UPDATE_SCRN:
		return mp200_fb_ioc_update_scrn(info, cmd, arg);

	case MP200_FB_UPDATE_TIMER:
		/* update timer infomation          */
		return mp200_fb_ioc_update_timer(info, arg);

#ifdef CONFIG_FB_MP200_NTS
	case MP200_FB_SET_OUTPUT:
		return mp200_fb_ioc_set_output(info, arg);

	case MP200_FB_GET_OUTPUT:
		return mp200_fb_ioc_get_output(info, arg);
#endif /* CONFIG_FB_MP200_NTS */

	default:
		printk_dbg((_DEBUG_FB & 0x01),
		 "ioctl(*** unknown request ***)\n");
		return -ENOIOCTLCMD;
	}
}


/******************************************************************************
* MODULE   : mp200_fb_ioc_chk_scrn
* FUNCTION : frame buffer driver ioctl
* RETURN   :  0      : success
*            -EFAULT :
* NOTE     : none
* CREATE   : 2008.04.03 M.Haguro
* UPDATE   : 2008.12.10 J.Manabe
******************************************************************************/
static int mp200_fb_ioc_chk_scrn(struct fb_info *info, unsigned long arg)
{
#if (TIMEOUT_MSLEEP == 0)
	int             iWaitTime  = 0;
	int             iTotalTime = 0;
	scrn_mode       FBScrnMode;
	mp200_fb_par   *par  = (mp200_fb_par *)info->par;
#else /* TIMEOUT_MSLEEP==1 */
	int             iTotalTime = 0;
	scrn_mode       FBScrnMode;
	struct timeval  tStartTime;
	struct timeval  tNowTime;
	mp200_fb_par   *par  = (mp200_fb_par *)info->par;

	/* get start time */
	do_gettimeofday(&tStartTime);
#endif
	printk_dbg((_DEBUG_FB & 0x01), "ioctl(CHKSCRN) <start>\n");

	down(&par->sem_tvmode);
	/* TVMODE Check */
	if ((par->devflag == EM1_FB_DEVICE_NTS)
	 && (tvmode == EM1_TV_MODE_OFF)) {
		up(&par->sem_tvmode);
		return -EIO;
	}

	/* get scrn mode info */
	if (copy_from_user(&FBScrnMode, (scrn_mode *) arg, sizeof(scrn_mode))) {
		up(&par->sem_tvmode);
		return -EFAULT;
	}

	/* check frame buffer status */
	FBScrnMode.page0 = chk_fb_status(fb_data[par->devflag].iFBStatus_A);
	FBScrnMode.page1 = chk_fb_status(fb_data[par->devflag].iFBStatus_B);

	/* if arg->timeout=0, then break */
	if (FBScrnMode.timeout <= 0) {
		if (copy_to_user((scrn_mode *) arg, &FBScrnMode, sizeof(scrn_mode))) {
			up(&par->sem_tvmode);
			return -EFAULT;
		}
		up(&par->sem_tvmode);
		return 0;
	}

	/* if both A and B states are "WAITING_FOR_ON_DISPLAY", */
	/*    then waits during the time specified by "timeout" */
	if (FBScrnMode.page0 == WAITING_FOR_ON_DISPLAY
	 && FBScrnMode.page1 == WAITING_FOR_ON_DISPLAY) {
#if (TIMEOUT_MSLEEP == 0)
		/* loop until timeout or enable write */
		do {
			/* timeout time (msec to jiffies) */
			iWaitTime = (RECHK_TIME + 1000/HZ-1) / (1000/HZ);

			wait_event_interruptible_timeout(par->wait_fb_chk_scrn, ((fb_data[par->devflag].iFBStatus_A == IS_NODATA) || (fb_data[par->devflag].iFBStatus_A == ENABLE_WRITE) || (fb_data[par->devflag].iFBStatus_B == IS_NODATA)    || (fb_data[par->devflag].iFBStatus_B == ENABLE_WRITE)), iWaitTime);

			iTotalTime += RECHK_TIME;

			/* re-check frame buffer status */
			FBScrnMode.page0 = chk_fb_status(fb_data[par->devflag].iFBStatus_A);
			FBScrnMode.page1 = chk_fb_status(fb_data[par->devflag].iFBStatus_B);

			/* when over timeout time, return */
			if (iTotalTime > FBScrnMode.timeout)
				break;

		} while (FBScrnMode.page0 == WAITING_FOR_ON_DISPLAY
		 && FBScrnMode.page1 == WAITING_FOR_ON_DISPLAY);
#else /* TIMEOUT_MSLEEP==1 */
		/* loop until timeout or enable write */
		do {
			/* sleep 10ms */
			msleep(RECHK_TIME);

			/* get time */
			do_gettimeofday(&tNowTime);
			if (tNowTime.tv_sec > tStartTime.tv_sec) {
				iTotalTime = (tNowTime.tv_sec - tStartTime.tv_sec) * 1000;
				if (tNowTime.tv_usec > tStartTime.tv_usec)
					iTotalTime += (tNowTime.tv_usec - tStartTime.tv_usec) / 1000;
				else
					iTotalTime -= (tStartTime.tv_usec - tNowTime.tv_usec) / 1000;

			else
				iTotalTime = (tNowTime.tv_usec - tStartTime.tv_usec) / 1000;

			/* re-check frame buffer status */
			FBScrnMode.page0 = chk_fb_status(fb_data[par->devflag].iFBStatus_A);
			FBScrnMode.page1 = chk_fb_status(fb_data[par->devflag].iFBStatus_B);

			/* when over timeout time, return */
			if (iTotalTime > FBScrnMode.timeout)
				break;

		} while (FBScrnMode.page0 == WAITING_FOR_ON_DISPLAY
		 && FBScrnMode.page1 == WAITING_FOR_ON_DISPLAY);
#endif
	}
	printk_dbg((_DEBUG_FB & 0x08), "ioctl(CHKSCRN): A(%d) B(%d)\n", FBScrnMode.page0, FBScrnMode.page1);

	/* return scrn mode info */
	if (copy_to_user((scrn_mode *) arg, &FBScrnMode, sizeof(scrn_mode))) {
		up(&par->sem_tvmode);
		return -EFAULT;
	}
	if (FBScrnMode.page0 == WAITING_FOR_ON_DISPLAY
	 && FBScrnMode.page1 == WAITING_FOR_ON_DISPLAY) {
		up(&par->sem_tvmode);
		return FBScrnMode.timeout;
	}

	up(&par->sem_tvmode);

	printk_dbg((_DEBUG_FB & 0x02), "ioctl(CHKSCRN) <end>\n");
	return 0;
}


/******************************************************************************
* MODULE   : mp200_fb_ioc_get_modes
* FUNCTION : frame buffer driver ioctl
* RETURN   :  0      : success
*            -EFAULT :
* NOTE     : none
* CREATE   : 2008.04.03 M.Haguro
* UPDATE   : 2008.12.10 J.Manabe
******************************************************************************/
static int mp200_fb_ioc_get_modes(struct fb_info *info, unsigned long arg)
{
	scrn_modes    FBScrnModes;
	mp200_fb_par *par  = (mp200_fb_par *)info->par;

	printk_dbg((_DEBUG_FB & 0x01), "ioctl(GET_MODES) <start>\n");

	down(&par->sem_tvmode);
	/* TVMODE Check */
	if ((par->devflag == EM1_FB_DEVICE_NTS)
	 && (tvmode == EM1_TV_MODE_OFF)) {
		up(&par->sem_tvmode);
		return -EIO;
	}

	/* set now on display infomation */
	set_image_info_to_2DAPI(fb_data[par->devflag].iNowDispPage,
	 &FBScrnModes, info);

	/* copy to user scrn mode infomation */
	if (copy_to_user((scrn_modes *) arg, &FBScrnModes, sizeof(scrn_modes))) {
		up(&par->sem_tvmode);
		return -EFAULT;
	}

	up(&par->sem_tvmode);

	printk_dbg((_DEBUG_FB & 0x02), "(ioctl(GET_MODES) <end>\n");
	return 0;
}


/******************************************************************************
* MODULE   : mp200_fb_ioc_set_modes
* FUNCTION : frame buffer driver ioctl
* RETURN   :  0      : success
*            -1      :
*            -EBUSY  :
*            -ENXIO  :
*            -EFAULT :
* NOTE     : none
* CREATE   : 2008.04.03 M.Haguro
* UPDATE   :
******************************************************************************/
static int mp200_fb_ioc_set_modes(struct fb_info *info, unsigned int cmd,
 unsigned long arg)
{
	return mp200_fb_ioc_update_scrn(info, cmd, arg);
}


/******************************************************************************
* MODULE   : mp200_fb_ioc_update_scrn
* FUNCTION : frame buffer driver ioctl
* RETURN   :  0      : success
*            -1      :
*            -EBUSY  :
*            -ENXIO  :
*            -EFAULT :
* NOTE     : none
* CREATE   : 2008.04.03 M.Haguro
* UPDATE   : 2008.12.10 J.Manabe
******************************************************************************/
static int mp200_fb_ioc_update_scrn(struct fb_info *info, unsigned int cmd,
 unsigned long arg)
{
	scrn_modes    FBSrcnModes;
	mp200_fb_par *par = (mp200_fb_par *)info->par;

	printk_dbg((_DEBUG_FB & 0x01), "ioctl(UPDATE_SCRN) <start>\n");

	down(&par->sem_tvmode);
	/* TVMODE Check */
	if ((par->devflag == EM1_FB_DEVICE_NTS)
	 && (tvmode == EM1_TV_MODE_OFF)) {
		up(&par->sem_tvmode);
		return -EIO;
	}
	/* copy fb image info from 2D API */
	if (copy_from_user(&FBSrcnModes, (scrn_modes *) arg,
	 sizeof(scrn_modes))) {
		up(&par->sem_tvmode);
		return -EFAULT;
	}

	/* check fb image infomation from 2D API */
	if (chk_input_modes(FBSrcnModes)) {
		up(&par->sem_tvmode);
		return -1;
	}

	down(&par->sem_fb_image_info);		/* lock fb_image_info */

	/* set fb image infomation from 2D API */
	set_image_info_from_2DAPI(FBSrcnModes.page, &FBSrcnModes, info);

	/* set fb image infomaition to LCD */
	set_image_info_to_LCD(FBSrcnModes.page, info);

	fb_data[par->devflag].iNextDispPage = FBSrcnModes.page;

	/* when cmd is MP200_FB_SET_MODES, set fb image update flag off */
	if (cmd == MP200_FB_SET_MODES)
		fb_data[par->devflag].fb_image_info.update_flag = FB_UPDATE_OFF;
	else {
		/* when cmd is MP200_FB_UPDATE_MODES,
		   set fb image update flag on */
		/* absolutely update data */
		if (FBSrcnModes.update == UPDATE_FLAG_ON)
			fb_data[par->devflag].fb_image_info.update_flag = FB_ABSOLUTERY_UPDATE;
		else {
			/* when ONLY_2D_MODE, update data */
			fb_data[par->devflag].fb_image_info.update_flag = FB_UPDATE_ON;
		}
	}

	/* set image infomation to LCD driver */
	printk_dbg((_DEBUG_FB & 0x10), "ioctl(UPDATE_SCRN): request LCD\n");

#ifdef CONFIG_FB_MP200_NTS
	if (par->devflag == EM1_FB_DEVICE_LCD) {
#endif /* CONFIG_FB_MP200_NTS */
		printk_dbg((_DEBUG_FB & 0x02), "toLCD: PaddrYRGB(%lx)\n",
		 fb_data[par->devflag].fb_image_info.image_data.yrgbaddr);
		if (mp200_lcd_set_fb_image(&fb_data[par->devflag].fb_image_info)) {
			up(&par->sem_fb_image_info); /* unlock fb_image_info */
			up(&par->sem_tvmode);
			return -EBUSY;
		}
#ifdef CONFIG_FB_MP200_NTS
	} else {
		printk_dbg((_DEBUG_FB & 0x02), "toNTSC: PaddrYRGB(%lx)\n", fb_data[par->devflag].fb_image_info.image_data.yrgbaddr);
		if (mp200_nts_set_fb_image(&fb_data[par->devflag].fb_image_info)) {
			up(&par->sem_fb_image_info); /* unlock fb_image_info */
			up(&par->sem_tvmode);
			return -EBUSY;
		}
	}
#endif /* CONFIG_FB_MP200_NTS */

	up(&par->sem_fb_image_info);		/* unlock fb_image_info */

	if (FBSrcnModes.page == DISP_BUFA)
		info->screen_base = (char __iomem *)fb_data[par->devflag].DispBufAV;
	else
		info->screen_base = (char __iomem *)fb_data[par->devflag].DispBufBV;

	up(&par->sem_tvmode);

	printk_dbg((_DEBUG_FB & 0x02), "ioctl(UPDATE_SCRN) <end>\n");
	return 0;
}


/******************************************************************************
* MODULE   : mp200_fb_ioc_update_timer
* FUNCTION : frame buffer driver ioctl
* RETURN   :  0      : success
*            -EFAULT :
* NOTE     : none
* CREATE   : 2008.04.03 M.Haguro
* UPDATE   : 2008.12.10 J.Manabe
******************************************************************************/
static int mp200_fb_ioc_update_timer(struct fb_info *info, unsigned long arg)
{
	unsigned long flags;
	update_timer   FbUpdateTimer;
	mp200_fb_par *par = (mp200_fb_par *)info->par;

	printk_dbg((_DEBUG_FB & 0x01), "ioctl(UPDATE_TIMER) <start>\n");

	down(&par->sem_tvmode);
	/* TVMODE Check */
	if ((par->devflag == EM1_FB_DEVICE_NTS)
	 && (tvmode == EM1_TV_MODE_OFF)) {
		up(&par->sem_tvmode);
		return -EIO;
	}

	/* copy from update timer info from 2D API */
	if (copy_from_user(&FbUpdateTimer, (update_timer *) arg,
	 sizeof(update_timer))) {
		up(&par->sem_tvmode);
		return -EFAULT;
	}

	/* fb timer stop */
	spin_lock_irqsave(&fb_data[par->devflag].mp200_fb_dev->fb_lock, flags);
	del_timer(&fb_data[par->devflag].mp200_fb_timer);
	fb_data[par->devflag].iTimerStatusFlag = FB_TIMER_STOP;
	spin_unlock_irqrestore(&fb_data[par->devflag].mp200_fb_dev->fb_lock,
	 flags);

	/* check input value */
	if (chk_input_update_timer(&FbUpdateTimer)) {
		up(&par->sem_tvmode);
		return -1;
	}

	/* set update structure info */
	fb_data[par->devflag].iUpdateTimerFlag = FbUpdateTimer.timerflag;
	fb_data[par->devflag].iTimeoutTime     = FbUpdateTimer.timeout;

	/* fb timer start */
	if ((fb_data[par->devflag].iUpdateTimerFlag == TIMER_ON)
	 && (fb_data[par->devflag].iMixModeFlg == ONLY_2D_MODE)) {
		spin_lock_irqsave(&fb_data[par->devflag].mp200_fb_dev->fb_lock, flags);
		if (fb_data[par->devflag].iTimerStatusFlag == FB_TIMER_STOP) {
			printk_dbg((_DEBUG_FB & 0x10),
			 "ioctl(UPDATE_TIMER): add_timer\n");
			fb_data[par->devflag].mp200_fb_timer.expires = jiffies
			 + (fb_data[par->devflag].iTimeoutTime * HZ / 1000);
			add_timer(&fb_data[par->devflag].mp200_fb_timer);
			fb_data[par->devflag].iTimerStatusFlag = FB_TIMER_START;
		}
		spin_unlock_irqrestore(&fb_data[par->devflag].mp200_fb_dev->fb_lock, flags);
	}

	up(&par->sem_tvmode);

	printk_dbg((_DEBUG_FB & 0x02), "ioctl(UPDATE_TIMER) <end>\n");
	return 0;
}


#ifdef CONFIG_FB_MP200_NTS
/******************************************************************************
* MODULE   : mp200_fb_ioc_set_output
* FUNCTION : frame buffer driver ioctl
* RETURN   :  0      : success
*            -EINVAL : input value error
*            -EIO    :
* NOTE     : none
* CREATE   : 2009.01.06 J.Manabe
* UPDATE   :
******************************************************************************/
static int mp200_fb_ioc_set_output(struct fb_info *info, unsigned long arg)
{
	int ret = 0;
	EM1_TV_MODE mode;
	mp200_fb_par *par  = (mp200_fb_par *)info->par;

	printk_dbg((_DEBUG_FB & 0x01), "<start>\n");

	/* fb device check */
	if (par->devflag != EM1_FB_DEVICE_NTS) {
		printk_dbg((_DEBUG_FB & 0x01), "fb1 IF only!\n");
		return -EIO;
	}

	/* copy fb image info from 2D API */
	if (copy_from_user(&mode, (EM1_TV_MODE *) arg, sizeof(EM1_TV_MODE)))
		return -EFAULT;

	down(&par->sem_tvmode);

	/* MODE CHECK */
	if ((mode != EM1_TV_MODE_OFF) &&
		(mode != EM1_TV_MODE_NTS) &&
		(mode != EM1_TV_MODE_PAL)) {
		up(&par->sem_tvmode);
		return -EINVAL;
	}

	/* ERROR CHECK */
	if (tvmode == mode) {
		up(&par->sem_tvmode);
		return 0;
	} else if ((tvmode == EM1_TV_MODE_NTS) && (mode == EM1_TV_MODE_PAL)) {
		up(&par->sem_tvmode);
		return -EIO;
	} else if ((tvmode == EM1_TV_MODE_PAL) && (mode == EM1_TV_MODE_NTS)) {
		up(&par->sem_tvmode);
		return -EIO;
	}

	/* NTS SET */
	if (mode == EM1_TV_MODE_OFF) {
		fb_data[par->devflag].iFBStatus_A = IS_NODATA;
		fb_data[par->devflag].iFBStatus_B = IS_NODATA;

		del_timer(&fb_data[par->devflag].mp200_fb_timer);
		fb_data[par->devflag].iTimerStatusFlag = FB_TIMER_STOP;

		/* NTS SET(OFF) */
		mp200_nts_release(NTS_ACTIVE_FB);
	} else {
		del_timer(&fb_data[par->devflag].mp200_fb_timer);
		fb_data[par->devflag].iTimerStatusFlag = FB_TIMER_STOP;

		/* NTS SET(NTS/PAL) */
		ret = mp200_nts_reserve(NTS_ACTIVE_FB, (int)mode);
		if (ret < 0) {
			printk_dbg((_DEBUG_FB & 0x01),
			 "nts_reserve err [%d]\n", ret);
			up(&par->sem_tvmode);
			return ret;
		}

		/* Set tv size */
		set_tv_size(mode, par);

		/* timer set */
		if (fb_data[par->devflag].iMixModeFlg == ONLY_2D_MODE) {
			if (fb_data[par->devflag].iUpdateTimerFlag == TIMER_ON) {
				printk_dbg((_DEBUG_FB & 0x10), "add_timer()\n");
				/* start timer interrupt */
				add_timer(&fb_data[par->devflag].mp200_fb_timer);
				fb_data[par->devflag].iTimerStatusFlag = FB_TIMER_START;
			} else {
				printk_dbg((_DEBUG_FB & 0x10),
				 "queue_work()\n");
				queue_work(par->mp200_fb_workqueue,
				 &par->wk_timeout_update_bottom_half);
			}
		}
	}
	set_tv_var(mode, par);

	/* set output mode */
	tvmode = mode;
	up(&par->sem_tvmode);

	printk_dbg((_DEBUG_FB & 0x01), "< end >\n");
	return 0;
}


/******************************************************************************
* MODULE   : mp200_fb_ioc_get_output
* FUNCTION : frame buffer driver ioctl
* RETURN   :  0      : success
*            -EFAULT :
* NOTE     : none
* CREATE   : 2008.12.25 J.Manabe
* UPDATE   :
******************************************************************************/
static int mp200_fb_ioc_get_output(struct fb_info *info, unsigned long arg)
{
	EM1_TV_MODE mode;
	mp200_fb_par *par = (mp200_fb_par *)info->par;
	printk_dbg((_DEBUG_FB & 0x01), "<start>\n");

	/* fb device check */
	if (par->devflag != EM1_FB_DEVICE_NTS) {
		printk_dbg((_DEBUG_FB & 0x01), "fb1 IF only!\n");
		return -EIO;
	}

	down(&par->sem_tvmode);
	mode = tvmode;
	/* copy to user EM1_TV_MODE infomation */
	if (copy_to_user((EM1_TV_MODE *) arg, &mode, sizeof(EM1_TV_MODE))) {
		up(&par->sem_tvmode);
		return -EFAULT;
	}
	up(&par->sem_tvmode);

	printk_dbg((_DEBUG_FB & 0x01), "< end >\n");
	return 0;
}
#endif /* CONFIG_FB_MP200_NTS */


#if 0
void ............set_status................() {}
#endif
/******************************************************************************
* MODULE   : set_image_info_to_2DAPI
* FUNCTION : set other image infomation (frame bufferA/B) to 2D API
* RETURN   : 0 : success
* NOTE     : none
* CREATE   : 2006.05.26 H.Nagashima
* UPDATE   : 2008.11.14 J.Manabe
******************************************************************************/
static int set_image_info_to_2DAPI(int iFBNo, scrn_modes *fb_scrn_modes, struct fb_info *info)
{
	mp200_fb_par *par = (mp200_fb_par *)info->par;

	/* set fb A other image infomation */
	if (iFBNo == FB_NO_A) {
		down(&par->sem_fb_other_info_A); /* lock fb_other_info_A */
		fb_scrn_modes->tc_enable = fb_data[par->devflag].fb_other_info_A.iMaskColFlg;
		fb_scrn_modes->t_color   = fb_data[par->devflag].fb_other_info_A.iMaskCol;
		fb_scrn_modes->alpha     = fb_data[par->devflag].fb_other_info_A.iAlpha;
		fb_scrn_modes->rot       = fb_data[par->devflag].fb_other_info_A.iInvFlg;
		fb_scrn_modes->page      = fb_data[par->devflag].fb_other_info_A.iPage;
		fb_scrn_modes->update    = fb_data[par->devflag].fb_other_info_A.iUpDate;
		up(&par->sem_fb_other_info_A);	/* unlock fb_other_info_A */
	} else {
	/* set fb B other image infomation */
		down(&par->sem_fb_other_info_B); /* lock fb_other_info_B */
		fb_scrn_modes->tc_enable = fb_data[par->devflag].fb_other_info_B.iMaskColFlg;
		fb_scrn_modes->t_color   = fb_data[par->devflag].fb_other_info_B.iMaskCol;
		fb_scrn_modes->alpha     = fb_data[par->devflag].fb_other_info_B.iAlpha;
		fb_scrn_modes->rot       = fb_data[par->devflag].fb_other_info_B.iInvFlg;
		fb_scrn_modes->page      = fb_data[par->devflag].fb_other_info_B.iPage;
		fb_scrn_modes->update    = fb_data[par->devflag].fb_other_info_B.iUpDate;
		up(&par->sem_fb_other_info_B);	/* unlock fb_other_info_B */
	}
	return 0;
}


/******************************************************************************
* MODULE   : chk_fb_status
* FUNCTION : check frame buffer (A/B) status
* RETURN   : 0 : update enable
*            1 : update disable
* NOTE     : none
* CREATE   : 2006.05.26 H.Nagashima
* UPDATE   : 2008.11.14 J.Manabe
******************************************************************************/
static int chk_fb_status(int iFBStatus)
{
	int iRet;

	/* check frame buffer status */
	if (iFBStatus == IS_NODATA || iFBStatus == ENABLE_WRITE)
		iRet = OUT_OF_DISPLAY;
	else
		iRet = WAITING_FOR_ON_DISPLAY;

	return iRet;
}


/******************************************************************************
* MODULE   : chk_input_modes
* FUNCTION : check input scrn_modes value from 2D API
* RETURN   : 0       : correct
*            -EINVAL : incorrect
* NOTE     : none
* CREATE   : 2006.06.05 H.Nagashima
* UPDATE   :
******************************************************************************/
static int chk_input_modes(scrn_modes fb_scrn_modes)
{
	/* check mask color flag */
	if ((fb_scrn_modes.tc_enable != FB_MASK_COLOR_DISP_OFF)
	 && (fb_scrn_modes.tc_enable != FB_MASK_COLOR_DISP_ON)) {
		printk_wrn("mask color flag incorrect  tc_enable = %d\n",
		 fb_scrn_modes.tc_enable);
		return -EINVAL;
	}

	/* check mask color */
	if ((fb_scrn_modes.t_color < 0x00)
	 || (0xFFFF < fb_scrn_modes.t_color)) {
		printk_wrn("mask color incorrect  t_color = %d\n",
		 fb_scrn_modes.t_color);
		return -EINVAL;
	}

	/* check alpha */
	if ((fb_scrn_modes.alpha < 0x00) || (0xFF < fb_scrn_modes.alpha)) {
		printk_wrn("alpha incorrect  alpha = %d\n",
		 fb_scrn_modes.alpha);
		return -EINVAL;
	}

	/* check rot */
	if ((fb_scrn_modes.rot != NO_INVERSE)
	 && (fb_scrn_modes.rot != UDRL_INVERSE)) {
		printk_wrn("rot incorrect  rot = %d\n", fb_scrn_modes.rot);
		return -EINVAL;
	}

	/* check page */
	if ((fb_scrn_modes.page != FB_NO_A)
	 && (fb_scrn_modes.page != FB_NO_B)) {
		printk_wrn("page incorrect  page = %d\n", fb_scrn_modes.page);
		return -EINVAL;
	}

	/* check update */
	if ((fb_scrn_modes.update != UPDATE_FLAG_OFF)
	 && (fb_scrn_modes.update != UPDATE_FLAG_ON)) {
		printk_wrn("update incorrect  update = %d\n",
		 fb_scrn_modes.update);
		return -EINVAL;
	}

	return 0;
}


/******************************************************************************
* MODULE   : set_image_info_from_2DAPI
* FUNCTION : set image infomation from LCD to fb local
* RETURN   : 0 : success
* NOTE     : none
* CREATE   : 2006.05.26 H.Nagashima
* UPDATE   :
******************************************************************************/
static int set_image_info_from_2DAPI(int iFBNo, scrn_modes *fb_scrn_modes,
 struct fb_info *info)
{
	mp200_fb_par *par = (mp200_fb_par *)info->par;

	/* set FB_A image infomation from 2D API */
	if (iFBNo == FB_NO_A) {
		fb_data[par->devflag].iFBStatus_A = NOW_WRITING_TO_FB;

		down(&par->sem_fb_other_info_A);/* lock fb_other_info_A */
		fb_data[par->devflag].fb_other_info_A.iMaskColFlg = fb_scrn_modes->tc_enable;
		fb_data[par->devflag].fb_other_info_A.iMaskCol    = fb_scrn_modes->t_color;
		fb_data[par->devflag].fb_other_info_A.iAlpha      = fb_scrn_modes->alpha;
		fb_data[par->devflag].fb_other_info_A.iInvFlg     = fb_scrn_modes->rot;
		fb_data[par->devflag].fb_other_info_A.iPage       = fb_scrn_modes->page;
		fb_data[par->devflag].fb_other_info_A.iUpDate     = fb_scrn_modes->update;
		up(&par->sem_fb_other_info_A);	/* unlock fb_other_info_A */
	} else {
	/* set FB_B image infomation from 2D API */
		fb_data[par->devflag].iFBStatus_B = NOW_WRITING_TO_FB;

		down(&par->sem_fb_other_info_B); /* lock fb_other_info_B */
		fb_data[par->devflag].fb_other_info_B.iMaskColFlg = fb_scrn_modes->tc_enable;
		fb_data[par->devflag].fb_other_info_B.iMaskCol    = fb_scrn_modes->t_color;
		fb_data[par->devflag].fb_other_info_B.iAlpha      = fb_scrn_modes->alpha;
		fb_data[par->devflag].fb_other_info_B.iInvFlg     = fb_scrn_modes->rot;
		fb_data[par->devflag].fb_other_info_B.iPage       = fb_scrn_modes->page;
		fb_data[par->devflag].fb_other_info_B.iUpDate     = fb_scrn_modes->update;
		up(&par->sem_fb_other_info_B);	/* unlock fb_other_info_B */
	}
	printk_dbg((_DEBUG_FB & 0x08), "[> set_2DAPI:   A(%d) B(%d)\n",
	 fb_data[par->devflag].iFBStatus_A, fb_data[par->devflag].iFBStatus_B);

	return 0;
}


/******************************************************************************
* MODULE   : set_image_info_to_LCD
* FUNCTION : set image infomation to LCD
* RETURN   : 0 : success
* NOTE     : none
* CREATE   : 2006.05.26 H.Nagashima
* UPDATE   : 2008.11.14 J.Manabe
******************************************************************************/
static int set_image_info_to_LCD(int iFBNo, struct fb_info *info)
{
	mp200_fb_par *par = (mp200_fb_par *)info->par;

	/* set fb A infomation */
	if (iFBNo == DISP_BUFA) {
		printk_dbg((_DEBUG_FB & 0x04), "DispBufA(%08lx)\n",
		 fb_data[par->devflag].DispBufA);
		fb_data[par->devflag].fb_image_info.image_data.yrgbaddr = fb_data[par->devflag].DispBufA;

		/* set other image infomation */
		down(&par->sem_fb_other_info_A);/* lock fb_other_info_A */
		fb_data[par->devflag].fb_image_info.maskcolr     = fb_data[par->devflag].fb_other_info_A.iMaskCol;
		fb_data[par->devflag].fb_image_info.maskcolrflg  = fb_data[par->devflag].fb_other_info_A.iMaskColFlg;
		fb_data[par->devflag].fb_image_info.alpha        = fb_data[par->devflag].fb_other_info_A.iAlpha;
		fb_data[par->devflag].fb_image_info.invflg       = fb_data[par->devflag].fb_other_info_A.iInvFlg;
		fb_data[par->devflag].fb_image_info.mix_buf_page = fb_data[par->devflag].fb_other_info_A.iPage;
		up(&par->sem_fb_other_info_A);	/* unlock fb_other_info_A */
	} else {
	/* set fb B infomation */
		printk_dbg((_DEBUG_FB & 0x04), "DispBufB(%08lx)\n", fb_data[par->devflag].DispBufB);
		fb_data[par->devflag].fb_image_info.image_data.yrgbaddr = fb_data[par->devflag].DispBufB;

		/* set other image infomation */
		down(&par->sem_fb_other_info_B); /* lock fb_other_info_B */
		fb_data[par->devflag].fb_image_info.maskcolr     = fb_data[par->devflag].fb_other_info_B.iMaskCol;
		fb_data[par->devflag].fb_image_info.maskcolrflg  = fb_data[par->devflag].fb_other_info_B.iMaskColFlg;
		fb_data[par->devflag].fb_image_info.alpha        = fb_data[par->devflag].fb_other_info_B.iAlpha;
		fb_data[par->devflag].fb_image_info.invflg       = fb_data[par->devflag].fb_other_info_B.iInvFlg;
		fb_data[par->devflag].fb_image_info.mix_buf_page = fb_data[par->devflag].fb_other_info_B.iPage;
		up(&par->sem_fb_other_info_B);	/* unlock fb_other_info_B */
	}

	return 0;
}


/******************************************************************************
* MODULE   : chk_input_update_timer
* FUNCTION : check update timer value from 2D API
* RETURN   : 0       : success
	   : -EINVAL : input value error
* NOTE     : none
* CREATE   : 2006.05.26 H.Nagashima
* UPDATE   :
******************************************************************************/
static int chk_input_update_timer(update_timer *fb_timer)
{
	/* check update mode value */
	if ((fb_timer->timerflag != TIMER_OFF)
	 && (fb_timer->timerflag != TIMER_ON)) {
		printk_wrn("timerflag value incorrect  timerflag = %d\n",
		 fb_timer->timerflag);
		return -EINVAL;
	}

	/* check update mode value */
	if (fb_timer->timeout < 33) {
		printk_wrn("timeout value incorrect  timeout = %d\n",
		 fb_timer->timeout);
		return -EINVAL;
	}

	return 0;
}


#ifdef CONFIG_FB_MP200_NTS
/******************************************************************************
* MODULE   : set_tv_size
* FUNCTION : tvmode change
* RETURN   : 0 : success
* NOTE     : none
* CREATE   : 2008.12.04 J.Manabe
* UPDATE   :
******************************************************************************/
static void set_tv_size(EM1_TV_MODE mode, mp200_fb_par *par)
{
	printk_dbg((_DEBUG_FB & 0x01), "<start>\n");

	down(&par->sem_fb_image_info);

	switch (mode) {
	case EM1_TV_MODE_NTS:
		fb_data[par->devflag].fb_image_info.image_data.hsize    = NTSC_WIDTH;
		fb_data[par->devflag].fb_image_info.image_data.vsize    = NTSC_HEIGHT;
		fb_data[par->devflag].fb_image_info.image_data.size     = NTSC_WIDTH * BYTES_PER_PIXEL;
		break;
	case EM1_TV_MODE_PAL:
		fb_data[par->devflag].fb_image_info.image_data.hsize    = PAL_WIDTH;
		fb_data[par->devflag].fb_image_info.image_data.vsize    = PAL_HEIGHT;
		fb_data[par->devflag].fb_image_info.image_data.size     = PAL_WIDTH * BYTES_PER_PIXEL;
		break;
	default:
	/* EM1_TV_MODE_OFF */
	break;
	}

	up(&par->sem_fb_image_info);

	printk_dbg((_DEBUG_FB & 0x01), "< end >\n");
}


/******************************************************************************
* MODULE   : set_tv_var
* FUNCTION : tvmode change
* RETURN   : 0 : success
* NOTE     : none
* CREATE   : 2009.1.19 M.haguro
* UPDATE   :
******************************************************************************/
static void set_tv_var(EM1_TV_MODE mode, mp200_fb_par *par)
{
	printk_dbg((_DEBUG_FB & 0x01), "<start>\n");

	down(&par->sem_fb_image_info);

	switch (mode) {
	case EM1_TV_MODE_NTS:
		par->fb_inf->var = mp200_fb_tv_default;
		par->fb_inf->var.xres         = NTSC_WIDTH;
		par->fb_inf->var.yres         = NTSC_HEIGHT;
		par->fb_inf->var.xres_virtual = NTS_WIDTH_V;
		par->fb_inf->var.yres_virtual = NTS_HEIGHT_V;
		par->fb_inf->var.bits_per_pixel = 16;
		break;
	case EM1_TV_MODE_PAL:
		par->fb_inf->var = mp200_fb_tv_default;
		par->fb_inf->var.xres         = PAL_WIDTH;
		par->fb_inf->var.yres         = PAL_HEIGHT;
		par->fb_inf->var.xres_virtual = NTS_WIDTH_V;
		par->fb_inf->var.yres_virtual = NTS_HEIGHT_V;
		par->fb_inf->var.bits_per_pixel = 16;
		break;
	default:
		/* EM1_TV_MODE_OFF */
		goto failed;
	}
	mp200_fb_set_par(par->fb_inf);

failed:
	up(&par->sem_fb_image_info);

	printk_dbg((_DEBUG_FB & 0x01), "< end >\n");
}
#endif /* CONFIG_FB_MP200_NTS */



#if 0
void ............fb_ops................() {}
#endif
/******************************************************************************
* MODULE   : mp200_fb_mmap
* FUNCTION : mmap
* RETURN   : 0       : success
*          : -EINVAL : input value error
*          : -EAGAIN : run again
* NOTE     : none
* CREATE   :
* UPDATE   : 2006.11.16 H.Nagashima
* UPDATE   : 2008.11.27 J.Manabe
******************************************************************************/
int mp200_fb_mmap(struct fb_info *info, struct vm_area_struct *vma)
{
	unsigned long start;
	unsigned long off;
	u32 len;

	if (vma->vm_pgoff > (~0UL >> PAGE_SHIFT))
		return -EINVAL;

	off = vma->vm_pgoff << PAGE_SHIFT;
	if (off < DispBufAllLength) {
		start = (unsigned long)Smem;
		len = PAGE_ALIGN((start & ~PAGE_MASK) + DispBufAllLength);
	} else
		return -EINVAL;

	start &= PAGE_MASK;
	if ((vma->vm_end - vma->vm_start + off) > len)
		return -EINVAL;

	off += start;
	vma->vm_pgoff = off >> PAGE_SHIFT;

	/* Accessing memory will be done non-cached. */
	vma->vm_page_prot = __pgprot(pgprot_val(vma->vm_page_prot)
	 | L_PTE_BUFFERABLE);
	vma->vm_page_prot = __pgprot(pgprot_val(vma->vm_page_prot)
	 & ~L_PTE_CACHEABLE);

	/* To stop the swapper from even considering these pages */
	vma->vm_flags |= (VM_IO | VM_RESERVED);

	if (remap_pfn_range(vma, vma->vm_start, vma->vm_pgoff,
	 vma->vm_end - vma->vm_start, vma->vm_page_prot)) {
		return -EAGAIN;
	}

	return 0;
}


/******************************************************************************
* MODULE   : mp200_fb_open
* FUNCTION : open qfb driver
* RETURN   : 0      : success
*            -EBUSY : LCD initialize failed
* NOTE     : none
* CREATE   :
* UPDATE   : 2006.05.29 H.Nagashima
* UPDATE   : 2008.11.14 J.Manabe
******************************************************************************/
static int mp200_fb_open(struct fb_info *info, int user)
{
	mp200_fb_par *par      = (mp200_fb_par *)info->par;

	if (fb_data[par->devflag].is_first) {
		if (par->devflag == EM1_FB_DEVICE_LCD) {
			/* initialize LCD driver */
			if (init_lcdhw() != 0) {
				printk_wrn("initialize LCD failed.\n");
				return -EBUSY;
			}

			printk_dbg((_DEBUG_FB & 0x10), "queue_work\n");
			queue_work(par->mp200_fb_workqueue,
			 &par->wk_timeout_update_bottom_half);
		}
		timer_init(info);
		fb_data[par->devflag].is_first = 0;
	}

	atomic_inc(&par->use_count);

	return 0;
}


/******************************************************************************
* MODULE   : mp200_fb_release
* FUNCTION : close qfb driver
* RETURN   : 0       : success
*            -EINVAL :
* NOTE     : none
* CREATE   :
* UPDATE   : 2006.05.29 H.Nagashima
* UPDATE   : 2009.01.06 J.Manabe
******************************************************************************/
static int mp200_fb_release(struct fb_info *info, int user)
{
	mp200_fb_par *par = (mp200_fb_par *)info->par;
	u32           count;
	unsigned long flags;

	down(&par->sem_tvmode);

	down(&par->sem_fb_image_info);		/* lock fb_image_info */
	count = atomic_read(&par->use_count);
	if (count == 0) {
		up(&par->sem_fb_image_info);	/* unlock fb_image_info */
		up(&par->sem_tvmode);
		return -EINVAL;
	}

	if (count == 1) {
		if (par->devflag == EM1_FB_DEVICE_LCD) {
			fb_data[par->devflag].iUpdateTimerFlag = DEFALT_UPDATE_TIMER_LCD;
			fb_data[par->devflag].iTimeoutTime     = TIMEOUT_TIME;
			del_timer(&fb_data[par->devflag].mp200_fb_timer);
			fb_data[par->devflag].iTimerStatusFlag = FB_TIMER_STOP;
			if ((fb_data[par->devflag].iUpdateTimerFlag == TIMER_ON) && (fb_data[par->devflag].iMixModeFlg == ONLY_2D_MODE)) {
				fb_data[par->devflag].mp200_fb_timer.expires = jiffies + (fb_data[par->devflag].iTimeoutTime * HZ / 1000);
				add_timer(&fb_data[par->devflag].mp200_fb_timer);
				fb_data[par->devflag].iTimerStatusFlag = FB_TIMER_START;
			}
		} else {
			spin_lock_irqsave(&fb_data[par->devflag].mp200_fb_dev->fb_lock, flags);
			/* default timer on */
			fb_data[par->devflag].iUpdateTimerFlag = TIMER_ON;
			fb_data[par->devflag].iTimeoutTime     = TIMEOUT_TIME;
			del_timer(&fb_data[par->devflag].mp200_fb_timer);
			fb_data[par->devflag].iTimerStatusFlag = FB_TIMER_STOP;
			spin_unlock_irqrestore(&fb_data[par->devflag].mp200_fb_dev->fb_lock, flags);

#ifdef CONFIG_FB_MP200_NTS
			/* NTS SET(OFF) */
			mp200_nts_release(NTS_ACTIVE_FB);
#endif /* CONFIG_FB_MP200_NTS */
			tvmode = EM1_TV_MODE_OFF;

			fb_data[par->devflag].iFBStatus_A = IS_NODATA;
			fb_data[par->devflag].iFBStatus_B = IS_NODATA;
		}
	}

	atomic_dec(&par->use_count);
	up(&par->sem_fb_image_info);		/* unlock fb_image_info */
	up(&par->sem_tvmode);
	return 0;
}


/******************************************************************************
* MODULE   : get_line_length
* FUNCTION : get line length
* RETURN   : length
* NOTE     : none
* CREATE   :
* UPDATE   :
******************************************************************************/
static u_long get_line_length(int xres_virtual, int bpp)
{
	u_long length;

	length = xres_virtual * bpp;
	length = (length + 31) & ~31;
	length >>= 3;
	return length;
}


/******************************************************************************
* MODULE   : mp200_fb_check_var
* FUNCTION : variable check
* RETURN   : 0       : success
*            -EINVAL : input value incorrect
*            -ENOMEM :
* NOTE     : none
* CREATE   :
* UPDATE   : 2006.05.29 H.Nagashima
* UPDATE   : 2008.11.14 J.Manabe
******************************************************************************/
static int mp200_fb_check_var(struct fb_var_screeninfo *var,
 struct fb_info *info)
{
	u_long line_length;
	mp200_fb_par *par  = (mp200_fb_par *)info->par;

	if (var->vmode & FB_VMODE_CONUPDATE) {
		var->vmode |= FB_VMODE_YWRAP;
		var->xoffset = info->var.xoffset;
		var->yoffset = info->var.yoffset;
	}

	/* Some very basic checks */
	if (!var->xres)
		var->xres = 1;
	if (!var->yres)
		var->yres = 1;
	if (var->xres > var->xres_virtual)
		var->xres_virtual = var->xres;
	if (var->yres > var->yres_virtual)
		var->yres_virtual = var->yres;
	if (var->bits_per_pixel <= 1)
		var->bits_per_pixel = 1;
	else if (var->bits_per_pixel <= 8)
		var->bits_per_pixel = 8;
	else if (var->bits_per_pixel <= 16)
		var->bits_per_pixel = 16;
	else if (var->bits_per_pixel <= 24)
		var->bits_per_pixel = 24;
	else if (var->bits_per_pixel <= 32)
		var->bits_per_pixel = 32;
	else
		return -EINVAL;

	if (var->xres_virtual < var->xoffset + var->xres)
		var->xres_virtual = var->xoffset + var->xres;

	if (var->yres_virtual < var->yoffset + var->yres)
		var->yres_virtual = var->yoffset + var->yres;

	/* Memory limit */
	line_length = get_line_length(var->xres_virtual, var->bits_per_pixel);
	if (line_length * var->yres_virtual > fb_data[par->devflag].DispBufLength)
		return -ENOMEM;

	/* check var */
	switch (var->bits_per_pixel) {
	case 1:	/* FALL THROUGH */
	case 8:
		{
			var->red.offset    = 0;
			var->red.length    = 8;
			var->green.offset  = 0;
			var->green.length  = 8;
			var->blue.offset   = 0;
			var->blue.length   = 8;
			var->transp.offset = 0;
			var->transp.length = 0;
			break;
		}
	case 16:		/* RGBA 5551 */
		{
			if (var->transp.length) {
				var->red.offset    = 0;
				var->red.length    = 5;
				var->green.offset  = 5;
				var->green.length  = 5;
				var->blue.offset   = 10;
				var->blue.length   = 5;
				var->transp.offset = 15;
				var->transp.length = 1;
			} else {		/* RGB 565 */
				var->red.offset    = 11;
				var->red.length    = 5;
				var->green.offset  = 5;
				var->green.length  = 6;
				var->blue.offset   = 0;
				var->blue.length   = 5;
				var->transp.offset = 0;
				var->transp.length = 0;
			}
			break;
		}
	case 24:		/* RGB 888 */
		{
			var->red.offset    = 0;
			var->red.length    = 8;
			var->green.offset  = 8;
			var->green.length  = 8;
			var->blue.offset   = 16;
			var->blue.length   = 8;
			var->transp.offset = 0;
			var->transp.length = 0;
			break;
		}
	case 32:		/* RGBA 8888 */
		{
			var->red.offset    = 0;
			var->red.length    = 8;
			var->green.offset  = 8;
			var->green.length  = 8;
			var->blue.offset   = 16;
			var->blue.length   = 8;
			var->transp.offset = 24;
			var->transp.length = 8;
			break;
		}
	}
	var->red.msb_right    = 0;
	var->green.msb_right  = 0;
	var->blue.msb_right   = 0;
	var->transp.msb_right = 0;

	return 0;
}

/******************************************************************************
* MODULE   : mp200_fb_set_par
* FUNCTION : set qfb parameter
* RETURN   : 0 : success
* NOTE     : none
* CREATE   :
* UPDATE   : 2006.05.29 H.Nagashima
******************************************************************************/
static int mp200_fb_set_par(struct fb_info *info)
{
	info->fix.line_length = get_line_length(info->var.xres_virtual,
	 info->var.bits_per_pixel);
	return 0;
}


/******************************************************************************
* MODULE   : mp200_fb_setcolreg
* FUNCTION : set qfb color
* RETURN   : 0 : success
*            1 :
* NOTE     : none
* CREATE   :
* UPDATE   : 2006.05.29 H.Nagashima
******************************************************************************/
static int mp200_fb_setcolreg(u_int regno, u_int red, u_int green, u_int blue,
 u_int transp, struct fb_info *info)
{
	/* No. of hw registers */
	if (regno >= 256)
		return 1;

	/* grayscale works only partially under directcolor */
	if (info->var.grayscale) {
		/* grayscale = 0.30*R + 0.59*G + 0.11*B */
		red = green = blue = (red * 77 + green * 151 + blue * 28) >> 8;
	}

#define CNVT_TOHW(val, width) ((((val)<<(width))+0x7FFF-(val))>>16)
	switch (info->fix.visual) {
	case FB_VISUAL_TRUECOLOR:	/* FALL THROUGH */
	case FB_VISUAL_PSEUDOCOLOR:
	{
		red = CNVT_TOHW(red, info->var.red.length);
		green = CNVT_TOHW(green, info->var.green.length);
		blue = CNVT_TOHW(blue, info->var.blue.length);
		transp = CNVT_TOHW(transp, info->var.transp.length);
		break;
	}
	case FB_VISUAL_DIRECTCOLOR:
	{
		red = CNVT_TOHW(red, 8);	/* expect 8 bit DAC */
		green = CNVT_TOHW(green, 8);
		blue = CNVT_TOHW(blue, 8);
		/* hey, there is bug in transp handling... */
		transp = CNVT_TOHW(transp, 8);
		break;
	}
	}
#undef CNVT_TOHW
	/* Truecolor has hardware independent palette */
	if (info->fix.visual == FB_VISUAL_TRUECOLOR) {
		u32 v;

		if (regno >= 16)
			return 1;

		v = (red << info->var.red.offset)
		  | (green << info->var.green.offset)
		  | (blue << info->var.blue.offset)
		  | (transp << info->var.transp.offset);
		switch (info->var.bits_per_pixel) {
		case 16:	/* FALL THROUGH */
		case 24:	/* FALL THROUGH */
		case 32:
		{
			((u32 *) (info->pseudo_palette))[regno] = v;
			break;
		}
		case 8:		/* FALL THROUGH */
		default:
		{
			break;
		}
		}
	}
	return 0;
}

/******************************************************************************
* MODULE   : mp200_fb_blank
* FUNCTION : FBIOBLANK
* RETURN   : 0       : success
*            -EINVAL : input value incorrect
* NOTE     : none
* CREATE   : 2009.9.14 Y.Utsuhara
* UPDATE   :
******************************************************************************/
static int mp200_fb_blank(int blank_mode, struct fb_info *info)
{
	mp200_fb_par *par  = (mp200_fb_par *)info->par;

	if (par->devflag == EM1_FB_DEVICE_LCD)
		return mp200_lcd_blank(blank_mode);
	else
		return mp200_nts_blank(blank_mode);

}

/******************************************************************************
* MODULE   : mp200_fb_pan_display
* FUNCTION :
* RETURN   : 0       : success
*            -EINVAL : input value incorrect
* NOTE     : none
* CREATE   :
* UPDATE   : 2006.05.29 H.Nagashima
******************************************************************************/
static int mp200_fb_pan_display(struct fb_var_screeninfo *var,
 struct fb_info *info)
{
	if (var->vmode & FB_VMODE_YWRAP) {
		if (var->yoffset < 0
		 || var->yoffset >= info->var.yres_virtual || var->xoffset) {
			return -EINVAL;
		}
	} else {
		if (var->xoffset + var->xres > info->var.xres_virtual ||
		    var->yoffset + var->yres > info->var.yres_virtual) {
			return -EINVAL;
		}
	}
	info->var.xoffset = var->xoffset;
	info->var.yoffset = var->yoffset;
	if (var->vmode & FB_VMODE_YWRAP)
		info->var.vmode |= FB_VMODE_YWRAP;
	else
		info->var.vmode &= ~FB_VMODE_YWRAP;

	return 0;
}


/******************************************************************************
* MODULE   : mp200_fb_callback
* FUNCTION : call back from LCD driver
* RETURN   : none
* NOTE     : public function
* CREATE   : 2006.05.27 H.Nagashima
* UPDATE   : 2008.11.14 J.Manabe
******************************************************************************/
void mp200_fb_callback(int iWrittenFBNo, int iMixModeStatus, int fb_num)
{
	printk_dbg((_DEBUG_FB & 0x01), "fb(%d) status(%d) from(%d)\n",
	 iWrittenFBNo, iMixModeStatus, fb_num);
	fb_data[fb_num].iMixModeFlg = iMixModeStatus;

	/* fb timer stop */
	del_timer(&fb_data[fb_num].mp200_fb_timer);
	fb_data[fb_num].iTimerStatusFlag = FB_TIMER_STOP;

	/* change FB (A/B) status */
	if (iWrittenFBNo == FB_NO_A) {
		fb_data[fb_num].iNowDispPage = FB_NO_A;
		fb_data[fb_num].iFBStatus_A  = ALREADY_DISPLAY;
		if (fb_data[fb_num].iNextDispPage != FB_NO_B) {
			fb_data[fb_num].iFBStatus_B = ENABLE_WRITE;
#if (TIMEOUT_MSLEEP == 0)
			wake_up_interruptible(&fb_data[fb_num].mp200_fb_dev->wait_fb_chk_scrn);
#endif
		}
	} else {
		fb_data[fb_num].iNowDispPage = FB_NO_B;
		fb_data[fb_num].iFBStatus_B  = ALREADY_DISPLAY;
		if (fb_data[fb_num].iNextDispPage != FB_NO_A) {
			fb_data[fb_num].iFBStatus_A = ENABLE_WRITE;
#if (TIMEOUT_MSLEEP == 0)
			wake_up_interruptible(&fb_data[fb_num].mp200_fb_dev->wait_fb_chk_scrn);
#endif
		}
	}
	printk_dbg((_DEBUG_FB & 0x08), "<] callback:    A(%d) B(%d)\n",
	 fb_data[fb_num].iFBStatus_A, fb_data[fb_num].iFBStatus_B);

	/* start timer */
	if ((fb_data[fb_num].iUpdateTimerFlag == TIMER_ON) && (fb_data[fb_num].iMixModeFlg == ONLY_2D_MODE)
#if defined(CONFIG_PM) || defined(CONFIG_DPM)
	    && !fb_data[fb_num].iDpmSuspendFlag
#endif /* CONFIG_PM || CONFIG_DPM */
	) {
		printk_dbg((_DEBUG_FB & 0x10), "resume: callback: add_timer\n");
		fb_data[fb_num].mp200_fb_timer.expires = jiffies
		 + (fb_data[fb_num].iTimeoutTime * HZ / 1000);
		add_timer(&fb_data[fb_num].mp200_fb_timer);
		fb_data[fb_num].iTimerStatusFlag = FB_TIMER_START;
	}
}
EXPORT_SYMBOL(mp200_fb_callback);


#if 0
void ............Init_Function................() {}
#endif
/********************************************************
 *  Init Function Definitions                           *
 *******************************************************/
/******************************************************************************
* MODULE   : mp200_fb_init_module
* FUNCTION : frame buffer driver init function
* RETURN   : 0      : success
*            -ENXIO : fail
* NOTE     : none
* CREATE   :
* UPDATE   : 2006.05.27 H.Nagashima
******************************************************************************/
int __init mp200_fb_init_module(void)
{
	int   iRet   = 0;

	iRet = platform_driver_register(&mp200_fb_driver);
	if (!iRet)
		return iRet;

	return 0;
}


/******************************************************************************
* MODULE   : mp200_fb_probe
* FUNCTION : frame buffer driver probe function
* RETURN   : 0       : success
*            -ENOMEM : mem error
*            -EBUSY  : other driver init error
* NOTE     : none
* CREATE   :
* UPDATE   : 2006.05.27 H.Nagashima
* UPDATE   : 2008.12.04 J.Manabe
******************************************************************************/
static int __init mp200_fb_probe(struct platform_device *dev)
{
	struct fb_info         *info = NULL;
	struct fb_info         *info_0 = NULL;
	struct fb_info         *info_1 = NULL;
	mp200_fb_par           *par  = NULL;
	int                     iRet = -ENOMEM;
	int                     fb_num = 0;

	/* reserve region Smem */
	if (!request_mem_region(Smem, SmemLength, "qfb")) {
		printk_wrn("cannot reserve smem\n");
		mp200_fb_probe_error(FB_INIT_ERROR__SMEM_NOT_RESERVE, info_0,
		 info_1, fb_num);
		return -ENOMEM;
	}

	/* remap SmemV */
	SmemV = ioremap_nocache(Smem, SmemLength);
	if (!SmemV) {
		printk_wrn("cannot ioremap_nocache smem\n");
		mp200_fb_probe_error(FB_INIT_ERROR__SMEM_NOT_REMAP, info_0,
		 info_1, fb_num);
		return -ENOMEM;
	}
	printk_dbg((_DEBUG_FB),
	 " Smem    (0x%lx)  SmemV    (0x%p)  SmemLength    (0x%lx)\n",
	 Smem, SmemV, SmemLength);

	for (fb_num = 0; fb_num < EM1_FB_DEVICES; fb_num++) {

		/* alloc framebuffer */
		info = framebuffer_alloc(sizeof(mp200_fb_par), &dev->dev);
		if (!info) {
			printk_wrn("cannot framebuffer alloc\n");
			mp200_fb_probe_error(FB_INIT_ERROR__FB_NOT_ALLOC,
			 info_0, info_1, fb_num);
			return -ENOMEM;
		} else {
			if (fb_num == 0)
				info_0 = info;
			else
				info_1 = info;
		}

		/* flag initialize */
		set_flag_init(fb_num);

		/* set buffer address */
		set_buffer_address(fb_num);

		/* set infomation */
		set_info(info, &par, fb_num);

		/* device flag set */
		par->devflag = fb_num;

		/* alloc cmap */
		iRet = fb_alloc_cmap(&info->cmap, 256, 0);
		if (iRet < 0) {
			printk_wrn("cannot framebuffer alloc cmap\n");
			mp200_fb_probe_error(FB_INIT_ERROR__FB_CMAP_NOT_ALLOC,
			 info_0, info_1, fb_num);
			return -ENOMEM;
		}

		/* regist frame buffer */
		iRet = register_framebuffer(info);
		if (iRet < 0) {
			printk_wrn("cannot regist framebuffer\n");
			mp200_fb_probe_error(FB_INIT_ERROR__FB_NOT_REGIST,
			 info_0, info_1, fb_num);
			return -EBUSY;
		}

		/* spinlock initialize */
		spin_lock_init(&par->fb_lock);

		/* semaphore initialize */
		sema_init(&par->sem_fb_image_info, 1);
		sema_init(&par->sem_fb_other_info_A, 1);
		sema_init(&par->sem_fb_other_info_B, 1);
		sema_init(&par->sem_tvmode, 1);

#if (TIMEOUT_MSLEEP == 0)
		/* wait_event initialize */
		init_waitqueue_head(&par->wait_fb_chk_scrn);
#endif

		/* workqueue initialize */
		par->mp200_fb_workqueue = create_singlethread_workqueue(DEV_NAME);
		INIT_WORK(&par->wk_timeout_update_bottom_half,
		 timeout_update_bottom_half_do);

		/* variable initialize */
		set_val_init(info, fb_num);

		/* set drvdata */
		drvdata.info[fb_num] = info;

		/* for callback function */
		fb_data[fb_num].mp200_fb_dev = par;
	}
	fb_num--;

	/*initialize LCD driver */
	if (mp200_lcd_init_module() != 0) {
		printk_wrn("cannot initialize LCD\n");
		mp200_fb_probe_error(FB_INIT_ERROR__LCD_NOT_INITIALIZE,
		 info_0, info_1, fb_num);
		return -EBUSY;
	}

	/* set drvdate */
	dev_set_drvdata(&dev->dev, &drvdata);

	printk_info("registered device fb\n");
	return 0;
}


/******************************************************************************
* MODULE   : mp200_fb_probe_error
* FUNCTION : When probe error, cleanup
* RETURN   : none
* NOTE     :
* CREATE   : 2006.05.27 H.Nagashima
* UPDATE   :
******************************************************************************/
static void __init  mp200_fb_probe_error(int iErrorMode,
 struct fb_info *info_0, struct fb_info *info_1, int fb_num)
{
	if (fb_num > 0) {
		switch (iErrorMode) {
		case FB_INIT_ERROR__LCD_NOT_INITIALIZE:
			unregister_framebuffer(info_1);
		case FB_INIT_ERROR__FB_NOT_REGIST:
			fb_dealloc_cmap(&info_1->cmap);
		case FB_INIT_ERROR__FB_CMAP_NOT_ALLOC:
			framebuffer_release(info_1);
		case FB_INIT_ERROR__FB_NOT_ALLOC:
		default:
			break;
		}
		unregister_framebuffer(info_0);
		fb_dealloc_cmap(&info_0->cmap);
		framebuffer_release(info_0);
		iounmap(SmemV);
		release_mem_region(Smem, DispBufAllLength);
	} else {
		switch (iErrorMode) {
		case FB_INIT_ERROR__LCD_NOT_INITIALIZE:
			unregister_framebuffer(info_0);
		case FB_INIT_ERROR__FB_NOT_REGIST:
			fb_dealloc_cmap(&info_0->cmap);
		case FB_INIT_ERROR__FB_CMAP_NOT_ALLOC:
			framebuffer_release(info_0);
		case FB_INIT_ERROR__FB_NOT_ALLOC:
			iounmap(SmemV);
		case FB_INIT_ERROR__SMEM_NOT_REMAP:
			release_mem_region(Smem, DispBufAllLength);
		case FB_INIT_ERROR__SMEM_NOT_RESERVE:
		default:
			break;
		}
	}
}


/******************************************************************************
* MODULE   : set_buffer_address
* FUNCTION : set frame buffer (A/B) address
* RETURN   : none
* NOTE     :
* CREATE   : 2006.05.27 H.Nagashima
* UPDATE   : 2008.12.04 J.Manabe
******************************************************************************/
static void set_buffer_address(int fb_num)
{
#ifdef CONFIG_FB_MP200_NTS
	if (fb_num == EM1_FB_DEVICE_LCD) {
#endif /* CONFIG_FB_MP200_NTS */
		/* set & clean DispBufA(2D) address */
		fb_data[fb_num].DispBufAV = SmemV + DISPBUF_A_OFFSET;
		memset_io((char __iomem *)fb_data[fb_num].DispBufAV, 0,
		 fb_data[fb_num].DispBufLength);

		/* set & clean DispBufB(2D) address */
		fb_data[fb_num].DispBufBV = SmemV + DISPBUF_B_OFFSET;
		memset_io((char __iomem *)fb_data[fb_num].DispBufBV, 0,
		 fb_data[fb_num].DispBufLength);
#ifdef CONFIG_FB_MP200_NTS
	} else {
		fb_data[fb_num].DispBufAV = SmemV + NTSCBUF_A_OFFSET;
		memset_io((char __iomem *)fb_data[fb_num].DispBufAV, 0,
		 fb_data[fb_num].DispBufLength);

		/* set & clean DispBufB(2D) address */
		fb_data[fb_num].DispBufBV = SmemV + NTSCBUF_B_OFFSET;
		memset_io((char __iomem *)fb_data[fb_num].DispBufBV, 0,
		 fb_data[fb_num].DispBufLength);
	}
#endif /* CONFIG_FB_MP200_NTS */
}


/******************************************************************************
* MODULE   : set_info
* FUNCTION : set frame buffer driver infomation
* RETURN   : none
* NOTE     :
* CREATE   : 2006.05.30 H.Nagashima
* UPDATE   : 2008.11.13 J.Manabe
******************************************************************************/
static void set_info(struct fb_info *info, mp200_fb_par **par, int fb_num)
{
	*par                  = (mp200_fb_par *)info->par;
	(*par)->fb_inf        = info;

	info->screen_base     = (char __iomem *)fb_data[fb_num].DispBufAV;
	info->screen_size     = fb_data[fb_num].DispBufLength;
	info->fbops           = &mp200_fb_ops;

#ifdef CONFIG_FB_MP200_NTS
	if (fb_num == EM1_FB_DEVICE_LCD)
		info->var     = mp200_fb_default;
	else
		info->var     = mp200_fb_tv_default;

#else
	info->var             = mp200_fb_default;
#endif /* CONFIG_FB_MP200_NTS */

	info->fix             = mp200_fb_fix;
	info->fix.line_length = info->var.xres_virtual
	 * info->var.bits_per_pixel / 8;
	info->fix.smem_start  = fb_data[fb_num].DispBufA;
	info->fix.smem_len    = fb_data[fb_num].DispBufLength;
	info->pseudo_palette  = (*par)->pseudo_palette;
	info->flags           = FBINFO_FLAG_DEFAULT;
}


/******************************************************************************
* MODULE   : set_flag_init
* FUNCTION : flag initialize
* RETURN   : none
* NOTE     :
* CREATE   : 2006.05.30 H.Nagashima
* UPDATE   : 2008.11.14 J.Manabe
******************************************************************************/
static void set_flag_init(int fb_num)
{
	fb_data[fb_num].iFBStatus_A      = IS_NODATA;
	fb_data[fb_num].iFBStatus_B      = IS_NODATA;
	printk_dbg((_DEBUG_FB & 0x08), "A(%d) B(%d)\n",
	 fb_data[fb_num].iFBStatus_A, fb_data[fb_num].iFBStatus_B);
	fb_data[fb_num].iNowDispPage     = DISP_BUFA;
	fb_data[fb_num].iNextDispPage    = DISP_BUFA;

#ifdef CONFIG_FB_MP200_NTS
	if (fb_num == EM1_FB_DEVICE_NTS) {
		fb_data[fb_num].iUpdateTimerFlag = DEFALT_UPDATE_TIMER_NTS;

		/* tvmode default set */
		tvmode = EM1_TV_MODE_OFF;
		/* struct init */
		fb_data[EM1_FB_DEVICE_NTS].DispBufA = SMEM_START
		 + NTSCBUF_A_OFFSET;
		fb_data[EM1_FB_DEVICE_NTS].DispBufB = SMEM_START
		 + NTSCBUF_B_OFFSET;
		fb_data[EM1_FB_DEVICE_NTS].DispBufLength = NTSCBUF_LENGTH;
	} else {
#endif /* CONFIG_FB_MP200_NTS */
		fb_data[fb_num].iUpdateTimerFlag = DEFALT_UPDATE_TIMER_LCD;
		fb_data[fb_num].DispBufA = SMEM_START + DISPBUF_A_OFFSET;
		fb_data[fb_num].DispBufB = SMEM_START + DISPBUF_B_OFFSET;
		fb_data[fb_num].DispBufLength = DISPBUF_LENGTH;
#ifdef CONFIG_FB_MP200_NTS
	}
#endif /* CONFIG_FB_MP200_NTS */

	fb_data[fb_num].iMixModeFlg      = ONLY_2D_MODE;
	fb_data[fb_num].iDpmSuspendFlag  = 0;

	fb_data[fb_num].is_first         = 1;

}


/******************************************************************************
* MODULE   : set_val_init
* FUNCTION : variable initialize
* RETURN   : none
* NOTE     :
* CREATE   : 2006.06.05 H.Nagashima
* UPDATE   : 2008.11.14 J.Manabe
******************************************************************************/
static void set_val_init(struct fb_info *info, int fb_num)
{
	/* fb A val init */
	fb_data[fb_num].fb_other_info_A.iMaskColFlg = TC_COLOR_DISABLE;
	fb_data[fb_num].fb_other_info_A.iMaskCol    = 0x0020;
	fb_data[fb_num].fb_other_info_A.iAlpha      = 0xFF;
	fb_data[fb_num].fb_other_info_A.iInvFlg     = NO_INVERSE;
	fb_data[fb_num].fb_other_info_A.iPage       = DISP_BUFA;
	fb_data[fb_num].fb_other_info_A.iUpDate     = UPDATE_FLAG_OFF;

	/* fb B val init */
	fb_data[fb_num].fb_other_info_B.iMaskColFlg = TC_COLOR_DISABLE;
	fb_data[fb_num].fb_other_info_B.iMaskCol    = 0x0020;
	fb_data[fb_num].fb_other_info_B.iAlpha      = 0xFF;
	fb_data[fb_num].fb_other_info_B.iInvFlg     = NO_INVERSE;
	fb_data[fb_num].fb_other_info_B.iPage       = DISP_BUFB;
	fb_data[fb_num].fb_other_info_B.iUpDate     = UPDATE_FLAG_OFF;

	/* fb data init (to LCD) */
	fb_data[fb_num].fb_image_info.image_data.hsize    = info->var.xres;
	fb_data[fb_num].fb_image_info.image_data.vsize    = info->var.yres;
	fb_data[fb_num].fb_image_info.image_data.size     = info->var.xres_virtual * info->var.bits_per_pixel / 8;
	fb_data[fb_num].fb_image_info.image_data.yrgbaddr = fb_data[fb_num].DispBufA;
	fb_data[fb_num].fb_image_info.image_data.uvaddr   = 0x00000000;
	fb_data[fb_num].fb_image_info.image_data.x        = 0;
	fb_data[fb_num].fb_image_info.image_data.y        = 0;

	/* fb info init (to LCD) */
	fb_data[fb_num].fb_image_info.maskcolrflg  = FB_MASK_COLOR_DISP_OFF;
	fb_data[fb_num].fb_image_info.maskcolr     = 0x0020;
	fb_data[fb_num].fb_image_info.alpha        = 0xFF;
	fb_data[fb_num].fb_image_info.invflg       = NO_INVERSE;
	fb_data[fb_num].fb_image_info.mix_buf_page = DISP_BUFA;
	fb_data[fb_num].fb_image_info.update_flag  = FB_UPDATE_OFF;

	/* timer */
	fb_data[fb_num].iTimeoutTime               = TIMEOUT_TIME;
}


#if 0
void ............Timer_Function................() {}
#endif
/********************************************************
 *  Timer Function Definitions                          *
 *******************************************************/
/******************************************************************************
* MODULE   : timer_init
* FUNCTION : call back from LCD driver
* RETURN   : none
* NOTE     : public function
* CREATE   : 2006.06.07 H.Nagashima
* UPDATE   : 2008.11.14 J.Manabe
******************************************************************************/
static void timer_init(struct fb_info *info)
{
	mp200_fb_par *par = (mp200_fb_par *)info->par;

	init_timer(&fb_data[par->devflag].mp200_fb_timer);
	fb_data[par->devflag].mp200_fb_timer.function = timeout_update;
	fb_data[par->devflag].mp200_fb_timer.entry.next = NULL;
	fb_data[par->devflag].mp200_fb_timer.data = (unsigned long)info;
	down(&par->sem_tvmode);
	if (par->devflag == EM1_FB_DEVICE_LCD || tvmode != EM1_TV_MODE_OFF) {
		if ((fb_data[par->devflag].iUpdateTimerFlag == TIMER_ON)
		 && (fb_data[par->devflag].iMixModeFlg == ONLY_2D_MODE)) {
			printk_dbg((_DEBUG_FB & 0x10), "add_timer\n");
			fb_data[par->devflag].mp200_fb_timer.expires = jiffies
			 + (TIMEOUT_TIME * HZ / 1000);
			add_timer(&fb_data[par->devflag].mp200_fb_timer);
			fb_data[par->devflag].iTimerStatusFlag = FB_TIMER_START;
			printk_dbg((_DEBUG_FB & 0x01), "%d <start>\n",
			 __LINE__);
		}
	}
	up(&par->sem_tvmode);
}


/******************************************************************************
* MODULE   : timeout_update
* FUNCTION : when Timer timeout, update execution
* RETURN   : none
* NOTE     : public function
* CREATE   : 2006.06.07 H.Nagashima
* UPDATE   : 2008.11.14 J.Manabe
******************************************************************************/
static void timeout_update(unsigned long data)
{
	unsigned long        flags;
	struct fb_info      *info;
	mp200_fb_par *par;

	printk_dbg((_DEBUG_FB & 0x10), "\n");

	info = (struct fb_info *)data;
	par  = (mp200_fb_par *)info->par;

	/* fb timer stop */
	spin_lock_irqsave(&fb_data[par->devflag].mp200_fb_dev->fb_lock, flags);
	del_timer(&fb_data[par->devflag].mp200_fb_timer);
	fb_data[par->devflag].iTimerStatusFlag = FB_TIMER_STOP;
	spin_unlock_irqrestore(&fb_data[par->devflag].mp200_fb_dev->fb_lock, flags);

	printk_dbg((_DEBUG_FB & 0x10), "queue_work\n");
	queue_work(par->mp200_fb_workqueue,
	 &par->wk_timeout_update_bottom_half);

	/* start timer */
	if ((fb_data[par->devflag].iUpdateTimerFlag == TIMER_ON)
	 && (fb_data[par->devflag].iMixModeFlg == ONLY_2D_MODE)) {
		spin_lock_irqsave(&fb_data[par->devflag].mp200_fb_dev->fb_lock, flags);
		if (fb_data[par->devflag].iTimerStatusFlag == FB_TIMER_STOP) {
			printk_dbg((_DEBUG_FB & 0x10), "add_timer\n");
			fb_data[par->devflag].mp200_fb_timer.expires = jiffies + (fb_data[par->devflag].iTimeoutTime * HZ / 1000);
			add_timer(&fb_data[par->devflag].mp200_fb_timer);
			fb_data[par->devflag].iTimerStatusFlag = FB_TIMER_START;
		}
		spin_unlock_irqrestore(&fb_data[par->devflag].mp200_fb_dev->fb_lock, flags);
	}
}


/******************************************************************************
* MODULE   : timeout_update_bottom_half_do
* FUNCTION : when Timer timeout, update execution
* RETURN   : none
* NOTE     : private function
* CREATE   : 2006.07.10 Y.Utsuhara
* UPDATE   : 2008.11.14 J.Manabe
******************************************************************************/
static void timeout_update_bottom_half_do(struct work_struct *num)
{
	mp200_fb_par *par;
	struct fb_info *info;

	printk_dbg((_DEBUG_FB & 0x10), "\n");

	par = container_of(num, mp200_fb_par, wk_timeout_update_bottom_half);
	info = par->fb_inf;
	if (down_trylock(&par->sem_fb_image_info))	/* lock fb_image_info */
		return;

	/* set image infomation to LCD */
	set_image_info_to_LCD(fb_data[par->devflag].iNextDispPage, info);

	/* udpate on (normal MP200_FB_UPDATE_SCRN exe) */
	fb_data[par->devflag].fb_image_info.update_flag = FB_UPDATE_ON;

#ifdef CONFIG_FB_MP200_NTS
	if (par->devflag == EM1_FB_DEVICE_LCD) {
#endif /* CONFIG_FB_MP200_NTS */
		/* call fb function */
		printk_dbg((_DEBUG_FB & 0x10), "request LCD\n");
		mp200_lcd_set_fb_image(&fb_data[par->devflag].fb_image_info);
#ifdef CONFIG_FB_MP200_NTS
	} else {
		/* call fb1 function */
		printk_dbg((_DEBUG_FB & 0x10), "request NTS\n");
		mp200_nts_set_fb_image(&fb_data[par->devflag].fb_image_info);
	}
#endif /* CONFIG_FB_MP200_NTS */

	up(&par->sem_fb_image_info);		/* unlock fb_image_info */
}


#if 0
void ............Exit_Function................() {}
#endif
/********************************************************
 *  Exit Function Definitions                           *
 *******************************************************/
/******************************************************************************
* MODULE   : mp200_fb_remove
* FUNCTION : remove
* RETURN   : 0 : success
* NOTE     : none
* CREATE   :
* UPDATE   : 2006.05.27 H.Nagashima
* UPDATE   : 2008.12.09 J.Manabe
******************************************************************************/
static int mp200_fb_remove(struct platform_device *device)
{
	MP200_FB_DRVDATA *drvdata = platform_get_drvdata(device);
	struct fb_info *info;
	int i;

	for (i = 0; i < EM1_FB_DEVICES; i++) {
		/* remove frame buffer */
		info = drvdata->info[i];
		if (info) {
			unregister_framebuffer(info);
			fb_dealloc_cmap(&info->cmap);
			framebuffer_release(info);
		}
	}

	/* lcd remove */
	mp200_lcd_exit_module();

	/* remove Smem */
	iounmap(SmemV);
	release_mem_region(Smem, SmemLength);

	return 0;
}


#if defined(CONFIG_PM) || defined(CONFIG_DPM)
/********************************************************
 *  Suspend/Resume Function Definitions                 *
 *******************************************************/
/******************************************************************************
* MODULE   : mp200_fb_suspend
* FUNCTION : suspend frame buffer driver
* RETURN   :  0 : success
*          : -1 : faile
* NOTE     : none
* CREATE   :
* UPDATE   : 2006.05.29 H.Nagashima
* UPDATE   : 2008.12.05 J.Manabe
******************************************************************************/
static int mp200_fb_suspend(struct platform_device *dev, pm_message_t state)
{
	/*** this module suspend ***/
	if (!fb_data[EM1_FB_DEVICE_LCD].is_first) {
		/*** this module suspend ***/
		/* call fb suspend function  */
		if (mp200_fb_suspend_sub(dev, state, EM1_FB_DEVICE_LCD))
			goto fail_fb_lcd;

		/*** relational module suspend ***/
		/* call LCD suspend function */
		if (mp200_lcd_suspend(dev, state))
			goto fail_lcd;

		/* call IMC suspend function */
		if (mp200_imc_suspend(dev, state))
			goto fail_imc;

	}
#ifdef CONFIG_MP200_NTS
#ifdef CONFIG_FB_MP200_NTS
	/*** this module suspend ***/
	if (!fb_data[EM1_FB_DEVICE_NTS].is_first) {
		/* call fb suspend function  */
		if (mp200_fb_suspend_sub(dev, state, EM1_FB_DEVICE_NTS))
			goto fail_fb_nts;

	}
#endif /* CONFIG_FB_MP200_NTS */
	/*** relational module suspend ***/
	if (mp200_nts_suspend(dev, state)) {	/* call NTS suspend function */
		goto fail_nts;
	}
#endif /* CONFIG_MP200_NTS */
	return 0;

#ifdef CONFIG_MP200_NTS
fail_nts:
#ifdef CONFIG_FB_MP200_NTS
	mp200_fb_resume_sub(dev, EM1_FB_DEVICE_NTS);
fail_fb_nts:
#endif /* CONFIG_FB_MP200_NTS */
	mp200_imc_resume(dev);
#endif /* CONFIG_MP200_NTS */
fail_imc:
	mp200_lcd_resume(dev);
fail_lcd:
	mp200_fb_resume_sub(dev, EM1_FB_DEVICE_LCD);
fail_fb_lcd:
	return -EBUSY;
}


/******************************************************************************
* MODULE   : mp200_fb_suspend_sub
* FUNCTION : suspend frame buffer driver
* RETURN   : 0 : success
* NOTE     : none
* CREATE   :
* UPDATE   : 2007.11.28 M.Haguro
* UPDATE   : 2008.12.04 J.Manabe
******************************************************************************/
static int mp200_fb_suspend_sub(struct platform_device *dev,
 pm_message_t state, int fb_num)
{
	int ret;

	if (state.event == DEV_SUSPEND_IDLE_1) {
		ret = 0;
	} else {/* state.event == PM_EVENT_SUSPEND */
		if (!mp200_sleep_while_idle) {
			/*** this module suspend ***/
			fb_data[fb_num].iDpmSuspendFlag = 1;
			/* stop timer interrupt */
			del_timer(&fb_data[fb_num].mp200_fb_timer);
		}
		ret = 0;
	}
	return ret;
}


/******************************************************************************
* MODULE   : mp200_fb_resume
* FUNCTION : resume frame buffer driver
* RETURN   : 0 : success
* NOTE     : none
* CREATE   :
* UPDATE   : 2006.05.29 H.Nagashima
* UPDATE   : 2008.12.04 J.Manabe
******************************************************************************/
static int mp200_fb_resume(struct platform_device *dev)
{
	if (!fb_data[EM1_FB_DEVICE_LCD].is_first) {
		/*** relathional module resume ***/
		/* call IMC resume function */
		mp200_imc_resume(dev);
		/* call LCD resume function */
		mp200_lcd_resume(dev);
		/* call fb  resume function */
		mp200_fb_resume_sub(dev, EM1_FB_DEVICE_LCD);
	}
#ifdef CONFIG_MP200_NTS
	mp200_nts_resume(dev);	/* call NTS resume function */
#ifdef CONFIG_FB_MP200_NTS
	if (!fb_data[EM1_FB_DEVICE_NTS].is_first) {
		/* call fb  resume function */
		mp200_fb_resume_sub(dev, EM1_FB_DEVICE_NTS);
	}
#endif /* CONFIG_FB_MP200_NTS */
#endif /* CONFIG_MP200_NTS */
	return 0;
}


/******************************************************************************
* MODULE   : mp200_fb_resume_sub
* FUNCTION : resume frame buffer driver
* RETURN   : 0 : success
* NOTE     : none
* CREATE   :
* UPDATE   : 2007.11.28 M.Haguro
* UPDATE   : 2008.12.04 J.Manabe
******************************************************************************/
static int mp200_fb_resume_sub(struct platform_device *dev, int fb_num)
{
	MP200_FB_DRVDATA *drvdata = platform_get_drvdata(dev);
	struct fb_info *info;
	mp200_fb_par *par;

	if (fb_data[fb_num].iDpmSuspendFlag) {
		/*** this module resume ***/
		fb_data[fb_num].iDpmSuspendFlag = 0;

		if (fb_num == EM1_FB_DEVICE_LCD || tvmode != EM1_TV_MODE_OFF) {
			if (fb_data[fb_num].iMixModeFlg == ONLY_2D_MODE) {
				if (fb_data[fb_num].iUpdateTimerFlag == TIMER_ON) {
					printk_dbg((_DEBUG_FB & 0x10),
					 "add_timer\n");
					add_timer(&fb_data[fb_num].mp200_fb_timer);	/* start timer interrupt */
				} else {
					info = drvdata->info[fb_num];
					par  = (mp200_fb_par *)info->par;

					printk_dbg((_DEBUG_FB & 0x10),
					 "queue_work\n");
					queue_work(par->mp200_fb_workqueue, &par->wk_timeout_update_bottom_half);
				}
			}
		}
	}
	return 0;
}
#endif /* CONFIG_PM || CONFIG_DPM */


MODULE_AUTHOR("NEC Electronics Corporation.");
MODULE_DESCRIPTION("EM1 framebuffer driver");
MODULE_LICENSE("GPL");


/* function alias  */
module_init(mp200_fb_init_module);


