/*
 * File Name       : /drivers/video/mp200/mp200_lcd.c
 * Function        : LCD driver corresponding to V4L2
 * Release Version :
 * Release Date    : 2009.10.20
 *
 * Copyright (C) NEC Electronics Corporation 2009
 *
 *
 * This program is free software;you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by Free Softwere
 * Foundation; either version 2 of License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warrnty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; If not, write to the Free Software Foundation, Inc., 59 Temple
 * Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */


/********************************************************
 *  Include Files                                       *
 *******************************************************/
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/version.h>
#include <linux/errno.h>
#include <linux/string.h>
#include <linux/mm.h>
#include <linux/tty.h>
#include <linux/slab.h>
#include <linux/vmalloc.h>
#include <linux/delay.h>
#include <linux/interrupt.h>
#include <linux/fb.h>
#include <linux/init.h>
#include <linux/ioport.h>
#include <linux/timer.h>
#include <linux/sched.h>
#include <linux/device.h>
#include <linux/platform_device.h>

#if defined(CONFIG_PM) || defined(CONFIG_DPM)
#include <mach/pm.h>
#endif /* CONFIG_PM || CONFIG_DPM */

#include <linux/uaccess.h>
#include <mach/hardware.h>
#include <mach/irqs.h>
#include <mach/pmu.h>
#include <mach/em1_mem.h>

#include "mp200_common.h"
#include "mp200_fb.h"

#include "lcdc.h"
#include "mp200_lcd_common.h"
#include "mp200_lcdhw.h"
#include "mp200_lcd.h"

#include "imc.h"
#include "mp200_imc.h"


/********************************************************
 *  Definitions                                         *
 *******************************************************/
#define _DEBUG_LCD  0x00 /* 00008421(bit) */
			 /* 0x01: debug function in
			  * 0x02: debug function out
			  * 0x40: debug FBIOBLANK
			  * 0x80: debug semafore
			  */


/********************************************************
 *  Definitions                                         *
 *******************************************************/
#define DEV_NAME "mp200_lcd"
#define DPM_SUSPEND_FLG_INIT 0
#define INDEX_LCD_VAR_NEXT_INIT 0


/********************************************************
 *  Macros                                              *
 *******************************************************/
#define printk_err(fmt, arg...) \
	do {                     \
		printk(KERN_ERR DEV_NAME ": %s: " fmt, __func__, ## arg); \
	} while (0)

#define printk_wrn(fmt, arg...) \
	do {                     \
		printk(KERN_WARNING DEV_NAME ": %s: " fmt, __func__, ## arg); \
	} while (0)

#define printk_info(fmt, arg...) \
	do {                      \
		printk(KERN_INFO DEV_NAME ": " fmt, ## arg); \
	} while (0)

#if _DEBUG_LCD
#define printk_dbg(level, fmt, arg...) \
	do {                            \
		if (level > 0) \
			printk(KERN_DEBUG DEV_NAME ": %s: " fmt, __func__, ## arg); \
	} while (0)
#else
#define printk_dbg(level, fmt, arg...) \
	;
#endif


/********************************************************
 *  Structure                                           *
 *******************************************************/
/* structure lcd device status */
typedef struct mp200_lcd_dev {
	/* semafore v4l2/fb/img local data */
	struct semaphore      sem_image_data;
	/* semafore IMC refresh */
	struct semaphore      sem_image_refresh;
	/* spin lock LCD timer flag        */
	spinlock_t	      lcd_lock;
} MP200_LCD_DEV;


/********************************************************
 *  Variables                                           *
 *******************************************************/

/* input image data from fb driver */
/* (2D) name of mix frame buffer temp   */
static  unsigned int  uiMixFrameBufferPage_tmp;

/* other img infomation */
static  unsigned int  uiMaskColor_tmp;		/* mask color */
static  unsigned int  uiAlpha_tmp;		/* alpha      */

/* Flags */
static           int  iMixDSPFlg_tmp;		/* DSP mix ON/OFF            */
static           int  iDSPPauseFlg;		/* DSP pause status          */
static  unsigned int  uiMaskColorFlag_tmp;	/* mask color ON/OFF Flag    */
static  unsigned int  uiAbsolutelyUpFlag;	/* absolutely update flag    */
	unsigned int  uiInverseFlag_tmp;	/* inverse mode Flag         */

/*
 * Smem
 */
/* LCDC MMIO */
static  unsigned long LCDCMmio       = LCDC_BASE;
static  unsigned long LCDCMmioLength = LCDC_SIZE;
	char         *LCDCMmioV;

#if 0
/* FrameBuf(LCD) */
/*     unsigned long FrameBuf       = SMEM_START + FRAMEBUF_OFFSET; */
/*     unsigned long FrameBufLength = FRAMEBUF_LENGTH; */
/*     char         *FrameBufV; */
#endif

/* LCD device status */
static MP200_LCD_DEV       lcd_dev;

/*
 * V4L2 dev data
 */
static void *pvDevData_tmp;		/* mp200_v4l2_device               */
static void *pvVideoBuf_tmp;		/* videobuf_buffer                 */

/*
 * timer
 */
static  struct timer_list lcd_timer;
static  int               iTimerFlag;

/* lcd_refresh_flg */
enum imc_refresh_status {
	IMC_R_IDLE         = 0,
	IMC_R_WAIT_REFRESH = 1,
	IMC_R_REFRESHED    = 2,
	IMC_R_DISPLAYED    = 3,
};

typedef struct mp200_lcd_var {
	struct list_head      list;
	/* semafore v4l2/fb/img local flag */
	struct semaphore      sem_image_flag;

	/* input image data from fb driver */
	/* (2D) name of mix frame buffer        */
	unsigned int  uiMixFrameBufferPage;
	/* Flags */
	int  iMixDSPFlgToFB;		/* return FB DSP mix ON/OFF  */
	int  iCallbackV4L2Flg;		/* need to callback to V4L2  */
	/* V4L2 dev data */
	void *pvDevData;		/* mp200_v4l2_device               */
	void *pvVideoBuf;		/* videobuf_buffer                 */

	enum imc_refresh_status lcd_refresh_flg;

#if defined(CONFIG_PM) || defined(CONFIG_DPM)
	int          ctrl_func;
#endif /* CONFIG_PM || CONFIG_DPM */
} MP200_LCD_VAR;
static MP200_LCD_VAR lcd_var[2];
static int index_lcd_var_next = INDEX_LCD_VAR_NEXT_INIT;

static struct list_head list_lcd_var;

/* FBIOBLANK */
typedef struct lcd_blank_state {
	int current_mode;
	int lcd_backlight;
	int lcd_output;
	int lcd_clock;
} LCD_BLANK_STATE;
static LCD_BLANK_STATE blank_state;

/********************************************************
 *  Prototype declarations of function                  *
 *******************************************************/
/*
 * LCD Driver
 */
/* Public function */
static void         mp200_lcd_remove(void);
#if defined(CONFIG_PM) || defined(CONFIG_DPM)
static int          lcd_suspend_chk_lcdout(void);

       int          IMC_reset_flg   = 0;
#endif /* CONFIG_PM || CONFIG_DPM */
static int          DPM_suspend_flg = DPM_SUSPEND_FLG_INIT;

/* Private function */
static void         mix_image(int iCallbackV4L2_tmp, int update, int type);
static void         lcd_irq_handler_callback(void);

static void         set_val_init(void);
static int          lcd_probe(void);
static void         lcd_probe_error(int iErrorMode);
static void         lcd_timer_init(void);
static void         lcd_timeout(unsigned long data);

/*
 * work queue
 */
static struct workqueue_struct *mp200_lcd_workqueue;
static struct work_struct       wk_timeout_bottom_half;
static void         lcd_timeout_bottom_half_do(struct work_struct *num);


/********************************************************
 *  Function Definitions                                *
 *******************************************************/
/******************************************************************************
* MODULE   : mp200_lcd_blank
* FUNCTION : FBIOBLANK
* RETURN   : 0       : success
*            -EINVAL : input value incorrect
* NOTE     : none
* CREATE   : 2009.9.14 Y.Utsuhara
* UPDATE   :
******************************************************************************/
int mp200_lcd_blank(int blank_mode)
{
	int backlight, output, clock;
	int lock = 0;
	unsigned long flags = 0;

	printk_dbg((_DEBUG_LCD & 0x40), "<start>\n");

	switch (blank_mode) {
	case FB_BLANK_UNBLANK:
		/* Screen: On,  HSync: On,  VSync: On */
		printk_dbg((_DEBUG_LCD & 0x40), "<FB_BLANK_UNBLANK>\n");
		backlight = 1;
		output    = 1;
		clock     = 1;
		break;
	case FB_BLANK_NORMAL:
		/* Screen: Off, HSync: On,  VSync: On */
		printk_dbg((_DEBUG_LCD & 0x40), "<FB_BLANK_NORMAL>\n");
		backlight = 0;
		output    = 1;
		clock     = 1;
		break;
	case FB_BLANK_VSYNC_SUSPEND:
		/* Screen: Off, HSync: On,  VSync: Off */
		printk_dbg((_DEBUG_LCD & 0x40), "<FB_BLANK_VSYNC_SUSPEND>\n");
		backlight = 0;
		output    = 0;
		clock     = 1;
		break;
	case FB_BLANK_HSYNC_SUSPEND:
		/* Screen: Off, HSync: Off, VSync: On */
		printk_dbg((_DEBUG_LCD & 0x40), "<FB_BLANK_HSYNC_SUSPEND>\n");
		backlight = 0;
		output    = 0;
		clock     = 1;
		break;
	case FB_BLANK_POWERDOWN:
		/* Screen: Off, HSync: Off, VSync: Off */
		printk_dbg((_DEBUG_LCD & 0x40), "<FB_BLANK_POWERDOWN>\n");
		backlight = 0;
		output    = 0;
		clock     = 0;
		break;
	default:
		printk_dbg((_DEBUG_LCD & 0x40), "<failed(%d)>\n", -EINVAL);
		return -EINVAL;
	}

	if ((backlight == 0) && (blank_state.lcd_backlight == 1)) {
		/* backlight On -> Off */
		printk_dbg((_DEBUG_LCD & 0x40), "<backlight On -> Off>\n");
		blank_state.lcd_backlight = 0;
		lcd_hw_backlight_off();
	}

	if (((output    == 0) && (blank_state.lcd_output    == 1)) ||
	    ((output    == 1) && (blank_state.lcd_output    == 0))) {
		down(&lcd_dev.sem_image_data);
		down(&lcd_dev.sem_image_refresh);
		down(&lcd_var[0].sem_image_flag);
		down(&lcd_var[1].sem_image_flag);
		spin_lock_irqsave(&lcd_dev.lcd_lock, flags);
		lock = 1;
	}

	if ((output    == 0) && (blank_state.lcd_output    == 1)) {
		/* output    On -> Off */
		printk_dbg((_DEBUG_LCD & 0x40), "<output    On -> Off>\n");
		blank_state.lcd_output    = 0;
		writel(LCD_BUSSEL_BLACK, LCDCMmioV + LCD_BUSSEL);
		lcd_suspend_chk_lcdout();
		lcd_hw_stop();
	}
	if ((clock     == 0) && (blank_state.lcd_clock     == 1)) {
		/* clock     On -> Off */
		printk_dbg((_DEBUG_LCD & 0x40), "<clock     On -> Off>\n");
		blank_state.lcd_clock     = 0;
		lcd_hw_save_reg();
		lcd_hw_reset();
	}
	if ((clock     == 1) && (blank_state.lcd_clock     == 0)) {
		/* clock     Off -> On */
		printk_dbg((_DEBUG_LCD & 0x40), "<clock     Off -> On>\n");
		blank_state.lcd_clock     = 1;
		lcd_hw_unreset();
		lcd_hw_restore_reg();	/* LCD_BUSSEL_BLACK is
					   set to LCD_BUSSEL*/
	}
	if ((output    == 1) && (blank_state.lcd_output    == 0)) {
		/* output    Off -> on */
		printk_dbg((_DEBUG_LCD & 0x40), "<output    Off -> On>\n");
		blank_state.lcd_output    = 1;
		writel(LCD_BUSSEL_LOCAL, LCDCMmioV + LCD_BUSSEL);
		lcd_hw_start();
	}

	if (lock == 1) {
		spin_unlock_irqrestore(&lcd_dev.lcd_lock, flags);
		up(&lcd_dev.sem_image_data);
		up(&lcd_dev.sem_image_refresh);
		up(&lcd_var[0].sem_image_flag);
		up(&lcd_var[1].sem_image_flag);
	}

	if ((backlight == 1) && (blank_state.lcd_backlight == 0)) {
		/* backlight Off -> On */
		printk_dbg((_DEBUG_LCD & 0x40), "<backlight Off -> On>\n");
		blank_state.lcd_backlight = 1;
		lcd_hw_backlight_on();
	}

	blank_state.current_mode = blank_mode;
	return 0;
}
EXPORT_SYMBOL(mp200_lcd_blank);


/******************************************************************************
* MODULE   : mp200_lcd_set_v4l2_image
* FUNCTION : set image data from v4l2 to LCD local
* RETURN   : 0 : success
* NOTE     : none
* CREATE   : 2006.05.25 H.Nagashima
* UPDATE   :
******************************************************************************/
int mp200_lcd_set_v4l2_image(V4L2_IMAGE_INFO *v4l2_image_info)
{
	unsigned long flags;

	int bpp_y, bpp_uv, bpp_v; /* bit par pixel  */
	int ppl_y, ppl_uv, ppl_v; /* pixel par line */

	printk_dbg((_DEBUG_LCD & 0x01), "<start>\n");

	/* Stop pause timer */
	spin_lock_irqsave(&lcd_dev.lcd_lock, flags);
	del_timer(&lcd_timer);
	iTimerFlag = LCD_TIMER_STOP;
	spin_unlock_irqrestore(&lcd_dev.lcd_lock, flags);
	iDSPPauseFlg = PAUSE_DSP_OFF;

	/* get v4l2 data semafore */
	down(&lcd_dev.sem_image_data);
	printk_dbg((_DEBUG_LCD & 0x80), "down(A) set_v4l2\n");

	if (init_is_first) {
		/* LCD not initialize */
		up(&lcd_dev.sem_image_data);
		printk_dbg((_DEBUG_LCD & 0x80), "up(A) set_v4l2\n");
		return -1;
	}

	/* DSP stop */
	if (v4l2_image_info->image_data.yrgbaddr == 0) {
		printk_dbg((_DEBUG_LCD & 0x01), "<stop DSP>\n");
		/* movie off */
		iMixDSPFlg_tmp = MIX_DSP_OFF;
		/* update Layer data */
		mix_image(CALLBACK_V4L2_OFF, UPDATE_ON, CLR_LAYER_V4L2);
		return 0;
	}

	/* select 2D(1) */
	memset(&v4l2_layer, 0, sizeof(struct l2_param));

	/* movie on */
	iMixDSPFlg_tmp   = MIX_DSP_ON;

	v4l2_layer.offset      = v4l2_image_info->image_data.size;
	v4l2_layer.position    = (v4l2_image_info->screen_data.y << IMC_Lx_POSY_SFT | v4l2_image_info->screen_data.x << IMC_Lx_POSX_SFT);  /* y, x */
	v4l2_layer.size        = (v4l2_image_info->screen_data.vsize << IMC_Lx_SIZEY_SFT | v4l2_image_info->screen_data.hsize << IMC_Lx_SIZEX_SFT); /* h, w */
	v4l2_layer.control     = IMC_Lx_CONTROL_ON;
	v4l2_layer.resize      = IMC_Lx_RESIZE_DISABLE;
	v4l2_layer.mirror      = IMC_L2x_MIRROR_OFF;

	switch (v4l2_image_info->yuvfmt) {
	case V4L2_FORMAT_YUV420Pl2: /* YUV420 Semi-Planar */
		printk_dbg((_DEBUG_LCD & 0x02), "format(YUV420Pl2)\n");
		bpp_y = 8;  bpp_uv = 8; bpp_v = 0;
		ppl_y = 1; ppl_uv = 2; ppl_v = 0;
		v4l2_layer.format      = IMC_Lx_FORMAT_YUV420Pl2;
		v4l2_layer.frameadr_yp = v4l2_image_info->image_data.yrgbaddr + v4l2_image_info->image_data.y * v4l2_layer.offset * bpp_y  / 8 / ppl_y + v4l2_image_info->image_data.x * bpp_y  / 8;
		v4l2_layer.frameadr_up = v4l2_image_info->image_data.uvaddr   + v4l2_image_info->image_data.y * v4l2_layer.offset * bpp_uv / 8 / ppl_uv + v4l2_image_info->image_data.x * bpp_uv / 8;
		v4l2_layer.frameadr_vp = 0;
		break;
	case V4L2_FORMAT_YUV422Pl2: /* YUV422 Semi-Planar */
		printk_dbg((_DEBUG_LCD & 0x02), "format(YUV422Pl2)\n");
		bpp_y = 8;  bpp_uv = 8; bpp_v = 0;
		ppl_y = 1; ppl_uv = 1; ppl_v = 0;
		v4l2_layer.format      = IMC_Lx_FORMAT_YUV422Pl2;
		v4l2_layer.frameadr_yp = v4l2_image_info->image_data.yrgbaddr + v4l2_image_info->image_data.y * v4l2_layer.offset * bpp_y  / 8 / ppl_y  + v4l2_image_info->image_data.x * bpp_y  / 8;
		v4l2_layer.frameadr_up = v4l2_image_info->image_data.uvaddr   + v4l2_image_info->image_data.y * v4l2_layer.offset * bpp_uv / 8 / ppl_uv + v4l2_image_info->image_data.x * bpp_uv / 8;
		v4l2_layer.frameadr_vp = 0;
		break;
	case V4L2_FORMAT_YUV420Pl: /* YUV420 Planar      */
		printk_dbg((_DEBUG_LCD & 0x02), "format(YUV420Pl)\n");
		bpp_y = 8;  bpp_uv = 4; bpp_v = 4;
		 ppl_y = 1; ppl_uv = 4; ppl_v = 4;
		v4l2_layer.format      = IMC_Lx_FORMAT_YUV420Pl;
		v4l2_layer.frameadr_yp = v4l2_image_info->image_data.yrgbaddr + v4l2_image_info->image_data.y * v4l2_layer.offset * bpp_y  / 8 / ppl_y  + v4l2_image_info->image_data.x * bpp_y  / 8;
		v4l2_layer.frameadr_up = v4l2_image_info->image_data.uvaddr   + v4l2_image_info->image_data.y * v4l2_layer.offset * bpp_uv / 8 / ppl_uv + v4l2_image_info->image_data.x * bpp_uv / 8;
		v4l2_layer.frameadr_vp = v4l2_image_info->image_data.vaddr    + v4l2_image_info->image_data.y * v4l2_layer.offset * bpp_v  / 8 / ppl_v  + v4l2_image_info->image_data.x * bpp_v  / 8;
		break;
	case V4L2_FORMAT_YUV422Px: /* YUV422 Interleave  */
		printk_dbg((_DEBUG_LCD & 0x02), "format(YUV422Px)\n");
		bpp_y = 16; bpp_uv = 0; bpp_v = 0;
		ppl_y = 1; ppl_uv = 0; ppl_v = 0;
		v4l2_layer.format      = IMC_Lx_FORMAT_YUV422Px;
		v4l2_layer.frameadr_yp = v4l2_image_info->image_data.yrgbaddr + v4l2_image_info->image_data.y * v4l2_layer.offset * bpp_y  / 8 / ppl_y  + v4l2_image_info->image_data.x * bpp_y  / 8;
		v4l2_layer.frameadr_up = 0;
		v4l2_layer.frameadr_vp = 0;
		break;
	default:
		printk_dbg((_DEBUG_LCD & 0x02), "format(%ld)\n",
		 v4l2_layer.format);
		up(&lcd_dev.sem_image_data); /* release data semafore */
		printk_dbg((_DEBUG_LCD & 0x80), "up(A) set_v4l2\n");
		return -1;
	}
	v4l2_layer.frameadr_yq = 0;
	v4l2_layer.frameadr_uq = 0;
	v4l2_layer.frameadr_vq = 0;

	/* set frame data from V4L2 */
	pvDevData_tmp  = v4l2_image_info->frame_data.dev;
	pvVideoBuf_tmp = v4l2_image_info->frame_data.buf;

	/* mix - image data from v4l2 and image data
	   from fb driver (DSP and 2D) */
	mix_image(CALLBACK_V4L2_ON, UPDATE_ON, SET_LAYER_V4L2);

	/* start PAUSE timer */
	spin_lock_irqsave(&lcd_dev.lcd_lock, flags);
	if (iTimerFlag == LCD_TIMER_STOP && !DPM_suspend_flg) {
		lcd_timer.expires = jiffies
		 + (MIX_DSP_TIMEOUT * 10 * HZ / 1000);
		add_timer(&lcd_timer);
		iTimerFlag = LCD_TIMER_START;
	}
	spin_unlock_irqrestore(&lcd_dev.lcd_lock, flags);

	return 0;
}
EXPORT_SYMBOL(mp200_lcd_set_v4l2_image);


/******************************************************************************
* MODULE   : mp200_lcd_set_fb_image
* FUNCTION : set image data from Frame Buffer to LCD local
* RETURN   :  0 : success
*            -1 : faile
* NOTE     : none
* CREATE   : 2006.05.25 H.Nagashima
* UPDATE   :
******************************************************************************/
int mp200_lcd_set_fb_image(FB_IMAGE_INFO *fb_image_info)
{
	printk_dbg((_DEBUG_LCD & 0x01), "<start>\n");

	/* get data semafore */
	down(&lcd_dev.sem_image_data);
	printk_dbg((_DEBUG_LCD & 0x80), "down(A) set_fb\n");

	/* Set other infomation of input data from fb */
	uiAlpha_tmp              = fb_image_info->alpha;
	uiInverseFlag_tmp        = fb_image_info->invflg;
	uiMixFrameBufferPage_tmp = fb_image_info->mix_buf_page;
	printk_dbg((_DEBUG_LCD & 0x01), "MixFrameBuffer(%d)\n",
	 uiMixFrameBufferPage_tmp);

	uiMaskColorFlag_tmp      = fb_image_info->maskcolrflg;
	uiMaskColor_tmp          = fb_image_info->maskcolr;
	uiMaskColor_tmp          = (((((uiMaskColor_tmp & 0xF800) >> 11) << 1) + ((uiMaskColor_tmp & 0x8000) >> 15)) << IMC_Lx_KEYR_SFT) | (((uiMaskColor_tmp & 0x07E0) >> 5) << IMC_Lx_KEYG_SFT) | (((((uiMaskColor_tmp & 0x001F) >>  0) << 1) + ((uiMaskColor_tmp & 0x0010) >>  4)) << IMC_Lx_KEYB_SFT);

	/* select 2D(1) */
	memset(&fb_layer, 0, sizeof(struct l01_param));

	fb_layer.format   = IMC_Lx_FORMAT_RGB565;
	fb_layer.resize   = IMC_Lx_RESIZE_DISABLE;
	fb_layer.offset   = fb_image_info->image_data.size;
	fb_layer.control  = IMC_Lx_CONTROL_ON;
	if (fb_image_info->maskcolrflg == TC_COLOR_ENABLE)
		fb_layer.keyenable = IMC_Lx_KEYEN_ENABLE;
	else
		fb_layer.keyenable = IMC_Lx_KEYEN_DISABLE;

	fb_layer.alpha    = (uiAlpha_tmp * ((1UL << 6) - 1) + ((1UL << 8) - 1) / 2) / ((1UL << 8) - 1); /* 8bit to 6bit */
	fb_layer.position = (fb_image_info->image_data.y << IMC_Lx_POSY_SFT | fb_image_info->image_data.x << IMC_Lx_POSX_SFT);  /* y, x */
	fb_layer.size     = (fb_image_info->image_data.vsize << IMC_Lx_SIZEY_SFT | fb_image_info->image_data.hsize << IMC_Lx_SIZEX_SFT); /* h, w */

	fb_layer.frameadr = fb_image_info->image_data.yrgbaddr; /* image area Y/RGB plane address */

	/* check update flag */
	switch (fb_image_info->update_flag) {
	case FB_UPDATE_ON:  /*    ioctl(MP200_FB_UPDATE_SCRN) & update = 0 */
			    /* or interval update                          */
		printk_dbg((_DEBUG_LCD & 0x02), "FB_UPDATE_ON\n");

		/* movie on */
		if (iMixDSPFlg_tmp == MIX_DSP_ON) {
			/* pause status */
			if (iDSPPauseFlg == PAUSE_DSP_ON) {
				mix_image(CALLBACK_V4L2_OFF, UPDATE_ON,
				 SET_LAYER_2D);
			} else {
				/* not pause status */
				mix_image(CALLBACK_V4L2_OFF, UPDATE_OFF,
				 SET_LAYER_2D);
			}
		} else {
			/* movie off */
			mix_image(CALLBACK_V4L2_OFF, UPDATE_ON, SET_LAYER_2D);
		}
		break;
	/* ioctl(MP200_FB_UPDATE_SCRN) & update = 1 */
	case FB_ABSOLUTERY_UPDATE:
		printk_dbg((_DEBUG_LCD & 0x02), "FB_ABSOLUTERY_UPDATE\n");
		/* movie on */
		mix_image(CALLBACK_V4L2_OFF, UPDATE_ON, SET_LAYER_2D);
		break;
	case FB_UPDATE_OFF:  /* ioctl(MP200_FB_SET_MODES) */
	default:  /* other                     */
		printk_dbg((_DEBUG_LCD & 0x02), "FB_UPDATE_OFF\n");
		mix_image(CALLBACK_V4L2_OFF, UPDATE_OFF, SET_LAYER_2D);
		break;
	}

	printk_dbg((_DEBUG_LCD & 0x02), "<end>\n");
	return 0;
}
EXPORT_SYMBOL(mp200_lcd_set_fb_image);


/******************************************************************************
* MODULE   : mix_image
* FUNCTION : By using IMG, mix 2D (and DSP) image
* RETURN   :  0 : success
*          : -1 : failed
* NOTE     : none
* CREATE   : 2006.05.25 H.Nagashima
* UPDATE   :
******************************************************************************/
static void mix_image(int iCallbackV4L2_tmp, int update, int type)
{
	unsigned long flags;
	int i = 0;

	if (update == UPDATE_ON) {
		/* get refresh semafore */
		down(&lcd_dev.sem_image_refresh);
		printk_dbg((_DEBUG_LCD & 0x80), "down(C) mix_image\n");

		spin_lock_irqsave(&lcd_dev.lcd_lock, flags);
		i = index_lcd_var_next;
		index_lcd_var_next++;
		if (index_lcd_var_next == 2)
			index_lcd_var_next = 0;
		spin_unlock_irqrestore(&lcd_dev.lcd_lock, flags);

		/* get flag semafore */
		down(&lcd_var[i].sem_image_flag);
		printk_dbg((_DEBUG_LCD & 0x80), "down(B) mix_image\n");

		lcd_var[i].lcd_refresh_flg = IMC_R_IDLE;

		/* set flag and temp data */
		lcd_var[i].iCallbackV4L2Flg      = iCallbackV4L2_tmp;
		lcd_var[i].uiMixFrameBufferPage  = uiMixFrameBufferPage_tmp;
		lcd_var[i].iMixDSPFlgToFB        = iMixDSPFlg_tmp;

		if (lcd_var[i].iCallbackV4L2Flg == CALLBACK_V4L2_ON) {
			/* set frame data from V4L2 */
			lcd_var[i].pvDevData  = pvDevData_tmp;
			lcd_var[i].pvVideoBuf = pvVideoBuf_tmp;
		}
	}

	if (iMixDSPFlg_tmp == MIX_DSP_OFF) {
		/* 8bit to 6bit */
		fb_layer.alpha  = (0xFF * ((1UL << 6) - 1)
		 + ((1UL << 8) - 1) / 2) / ((1UL << 8) - 1);

		ImcNxtKeycol.l1x_keycolor = uiMaskColor_tmp;

		/* update colorkey data */
		if (imc_hw_set_update_keycolor(&ImcNxtKeycol))
			printk_wrn("IMC update_keycolor() failed.\n");
	} else {
		fb_layer.alpha  = (uiAlpha_tmp * ((1UL << 6) - 1) + ((1UL << 8) - 1) / 2) / ((1UL << 8) - 1); /* 8bit to 6bit */

		if (v4l2_layer.format == IMC_Lx_FORMAT_YUV422Px) {
			ImcNxtBytelane.l2a_bytelane = IMC_L2x_BYTELANE_INIT;
			ImcNxtBytelane.l2b_bytelane = IMC_L2x_BYTELANE_422PX_YUYV;
		} else {
			ImcNxtBytelane.l2a_bytelane = IMC_L2x_BYTELANE_INIT;
			ImcNxtBytelane.l2b_bytelane = IMC_L2x_BYTELANE_INIT;
		}

		/* update bytelane data */
		if (imc_hw_set_update_bytelane(&ImcNxtBytelane))
			printk_wrn("IMC update_bytelane() failed.\n");
	}

	spin_lock_irqsave(&lcd_dev.lcd_lock, flags);
#if defined(CONFIG_PM) || defined(CONFIG_DPM)
	if (update == UPDATE_ON)
		lcd_var[i].ctrl_func = 2;
#endif /* CONFIG_PM || CONFIG_DPM */

	if (update == UPDATE_ON) {
		lcd_var[i].lcd_refresh_flg = IMC_R_WAIT_REFRESH;
		list_add_tail(&lcd_var[i].list, &list_lcd_var);
	}

	/* update Layer data */
	if (imc_hw_set_update_reserve(type, iMixDSPFlg_tmp, update))
		printk_dbg((_DEBUG_LCD & 0x02), "update_reserve() failed.\n");

	if (update == UPDATE_ON) {
		/* set IMC all layer data */
		ImcNxtVsync.mirror = uiInverseFlag_tmp;
		if (imc_hw_set_update_vsync(&ImcNxtVsync))
			printk_wrn("IMC update_vsync() failed.\n");
	}

	if (update == UPDATE_ON)
		change_frame();

	/* release data semafore */
	up(&lcd_dev.sem_image_data);
	printk_dbg((_DEBUG_LCD & 0x80), "up(A) mix_image\n");

	spin_unlock_irqrestore(&lcd_dev.lcd_lock, flags);

	if ((update == UPDATE_ON) && (blank_state.lcd_output == 0)) {
		printk_dbg((_DEBUG_LCD & 0x40),
		 "call lcd_callback_imc_refresh(): current blank mode = %d\n",
		 blank_state.current_mode);
		lcd_callback_imc_refresh();
	}
}


/*****************************************************************************
* MODULE   : lcd_irq_handler_callback
* FUNCTION : LCDC interrupt handler
* RETURN   :
* NOTE     : none
* CREATE   : 2007/07/09 M.Haguro
* UPDATE   :
******************************************************************************/
static void lcd_irq_handler_callback(void)
{
	FRAME_DATA frame_data;
	MP200_LCD_VAR *p_lcd_var;
	unsigned long flags;

	printk_dbg((_DEBUG_LCD & 0x01), "<start>\n");

	if (list_empty(&list_lcd_var))
		return;

	spin_lock_irqsave(&lcd_dev.lcd_lock, flags);

	p_lcd_var = list_entry(list_lcd_var.next, struct mp200_lcd_var, list);
	list_del(&p_lcd_var->list);

	p_lcd_var->lcd_refresh_flg = IMC_R_DISPLAYED;

	/* when kick mix_image by V4L2 */
	if (p_lcd_var->iCallbackV4L2Flg == CALLBACK_V4L2_ON) {
		p_lcd_var->iCallbackV4L2Flg = CALLBACK_V4L2_OFF;
		/* set frame date to V4L2 */
		frame_data.dev = p_lcd_var->pvDevData;
		frame_data.buf = p_lcd_var->pvVideoBuf;

		printk_dbg((_DEBUG_LCD & 0x02), "-> v4l2\n");
		/*call v4l2 callback function */
		mp200_v4l2_lcd_callback(frame_data);
	}

#if defined(CONFIG_PM) || defined(CONFIG_DPM)
	p_lcd_var->ctrl_func = 0;
#endif /* CONFIG_PM || CONFIG_DPM */

	/* release semafore */
	up(&p_lcd_var->sem_image_flag);
	printk_dbg((_DEBUG_LCD & 0x80), "up(B) lcd_irq\n");

	spin_unlock_irqrestore(&lcd_dev.lcd_lock, flags);

	printk_dbg((_DEBUG_LCD & 0x02), "<end>\n");
}


/*****************************************************************************
* MODULE   : lcd_irq_handler
* FUNCTION : LCDC interrupt handler
* RETURN   : IRQ_HANDLED
* NOTE     : none
* CREATE   : 2006/07/14 Y.Utsuhara
* UPDATE   : 2009.10.21
******************************************************************************/
irqreturn_t lcd_irq_handler(int irq, void *dev_id)
{
	unsigned long ulRegVal32;
	unsigned long flags;
	spin_lock_irqsave(&lcd_dev.lcd_lock, flags);

	/*  check Interuppt Status */
	ulRegVal32 = lcd_hw_chk_int_status();

	/* Interrupt Status Clear */
	lcd_hw_int_factor_clr();

	if (ulRegVal32 & LCD_UNDERRUN_BIT) {
		/* printk_err("LCDC UNDERRUN ERROR. Reset LCDC and IMC.\n");*/

		/* LCD display off */
		lcd_hw_stop();

		/* Reset LCD */
		lcd_hw_reset();
		/* UnReset LCD */
		lcd_hw_unreset();
		/* Restore LCD H/W register data */
		lcd_hw_restore_reg();

		imc_hw_reset(); /* Reset IMC */
		imc_hw_unreset(); /* UnReset IMC */

		IMC_reset_flg = 1;
		imc_hw_restore_reg();

		{
			struct list_head *list_tmp;
			MP200_LCD_VAR *p_lcd_var;
			void (*callback_func[4])(void) = {NULL, NULL, NULL, NULL};
			int i = 0;

			list_for_each(list_tmp, &list_lcd_var) {
				p_lcd_var = list_entry(list_tmp, struct mp200_lcd_var, list);
				switch (p_lcd_var->ctrl_func) {
				case 2:
					callback_func[i] = lcd_callback_imc_refresh;
					i++;
					/* FALL THROUGH */
				case 0:
				default:
					break;
				}
			}

			for (i = 0; i < 4; i++) {
				if (callback_func[i] == NULL)
					break;
				else
					callback_func[i]();
			}
			/*  Interuppt Enable Set */
			writel(LCD_UNDERRUNEN_BIT, LCDCMmioV + LCD_INTENSET);
		}
		/* LCD display on */
		lcd_hw_start();
	}

	spin_unlock_irqrestore(&lcd_dev.lcd_lock, flags);
	return IRQ_HANDLED;
}

/******************************************************************************
* MODULE   : lcd_callback_imc_refresh
* FUNCTION :
* RETURN   : none
* NOTE     : none
* CREATE   : 2007.07.09 M.Haguro
* UPDATE   :
******************************************************************************/
void lcd_callback_imc_refresh(void)
{
	FRAME_DATA    frame_data;
	unsigned long flags;
	struct list_head *list_tmp;
	MP200_LCD_VAR *p_lcd_var;
	int find = 0;

	spin_lock_irqsave(&lcd_dev.lcd_lock, flags);
	printk_dbg((_DEBUG_LCD & 0x01), "<start>\n");

	list_for_each(list_tmp, &list_lcd_var) {
		p_lcd_var = list_entry(list_tmp, struct mp200_lcd_var, list);
		if (p_lcd_var->lcd_refresh_flg == IMC_R_WAIT_REFRESH) {
			find = 1;
			break;
		}
	}

	if (find) {
		/* release refresh semafore */
		up(&lcd_dev.sem_image_refresh);
		printk_dbg((_DEBUG_LCD & 0x80),
		 "up(C) lcd_callback_imc_refresh\n");

		/* call FB Driver function "mp200_fb_callback" */
		mp200_fb_callback((int)p_lcd_var->uiMixFrameBufferPage,
		 p_lcd_var->iMixDSPFlgToFB, EM1_FB_DEVICE_LCD);

		/* when kick mix_image by V4L2 */
		if (p_lcd_var->iCallbackV4L2Flg == CALLBACK_V4L2_ON) {
			/* set frame date to V4L2 */
			frame_data.dev = p_lcd_var->pvDevData;
			frame_data.buf = p_lcd_var->pvVideoBuf;

			printk_dbg((_DEBUG_LCD & 0x02), "-> v4l2\n");
			/* call v4l2 callback function */
			mp200_v4l2_lcd_refresh_callback(frame_data);
		}

		p_lcd_var->lcd_refresh_flg = IMC_R_REFRESHED;

		lcd_irq_handler_callback();
	}

	printk_dbg((_DEBUG_LCD & 0x02), "<end>\n");
	spin_unlock_irqrestore(&lcd_dev.lcd_lock, flags);
}


/*****************************************************************************
* MODULE   : lcd_callback_imc_wb
* FUNCTION :
* RETURN   : none
* NOTE     : none
* CREATE   : 2007.07.09 M.Haguro
* UPDATE   :
******************************************************************************/
void lcd_callback_imc_wb(int status)
{
	;
}


/********************************************************
 *                                                      *
 *  Init Function Definitions                           *
 *                                                      *
 *******************************************************/

/******************************************************************************
* MODULE   : mp200_lcd_init_module
* FUNCTION : initialize LCD Driver
* RETURN   : 0     : success
*            other : fail
* NOTE     : none
* CREATE   : 2006.05.25 H.Nagashima
* UPDATE   :
******************************************************************************/
int __init mp200_lcd_init_module(void)
{
	int iRet;

	/* lcd device memset */
	memset(&lcd_dev, 0, sizeof(MP200_LCD_DEV));

	/* semafore initialize */
	sema_init(&lcd_dev.sem_image_data, 1);
	sema_init(&lcd_dev.sem_image_refresh, 1);

	/* list head initialize */
	INIT_LIST_HEAD(&list_lcd_var);

	/* spin lock initialize */
	spin_lock_init(&lcd_dev.lcd_lock);

	/* call lcd_probe */
	iRet = lcd_probe();
	if (iRet) {
		printk_err("lcd probe error\n");
		lcd_probe_error(iRet);
		return iRet;
	}

	/* variable initialized */
	set_val_init();

	/* timer initiarize */
	lcd_timer_init();

	return 0;
}

/******************************************************************************
* MODULE   : lcd_probe
* FUNCTION : LCD Driver probe
* RETURN   : 0     : success
*            other : fail
* NOTE     : none
* CREATE   : 2006.05.25 H.Nagashima
* UPDATE   :
******************************************************************************/
static int lcd_probe()
{
	/* reserve LCDC Mmio region */
#ifdef IO_ADDRESS
	LCDCMmioV = (char *)IO_ADDRESS(LCDCMmio);
#else
	if (!request_mem_region(LCDCMmio, LCDCMmioLength, DEV_NAME)) {
		printk_wrn("cannnot request LCDC mem region\n");
		return LCD_INIT_ERROR__LCDC_MMIO_NOT_RESERVE;
	}
	LCDCMmioV = ioremap_nocache(LCDCMmio, LCDCMmioLength);
	if (!LCDCMmioV) {
		printk_wrn("cannot ioremap_nocache LCDC mmio\n");
		return LCD_INIT_ERROR__LCDC_MMIO_NOT_REMAP;
	}
#endif
	printk_dbg((_DEBUG_LCD),
	 "LCDCMmio(0x%08lx)  LCDCMmioV(0x%p)  LCDCMmioLength(0x%lx)\n",
	 LCDCMmio, LCDCMmioV, LCDCMmioLength);

	/* initiarize work queue */
	mp200_lcd_workqueue = create_singlethread_workqueue(DEV_NAME);
	INIT_WORK(&wk_timeout_bottom_half, lcd_timeout_bottom_half_do);

	return 0;
}


/******************************************************************************
* MODULE   : lcd_probe_error
* FUNCTION : When fail in lcd_probe, remove lcd resource
* RETURN   : none
* NOTE     : none
* CREATE   : 2006.05.25 H.Nagashima
* UPDATE   :
******************************************************************************/
static void lcd_probe_error(int iErrorMode)
{
	switch (iErrorMode) {
	case LCD_INIT_ERROR__LCDM_START:		/* FALL THROUGH */
	case LCD_INIT_ERROR__LCD_IRQ_NOT_REQUEST:	/* FALL THROUGH */
	case LCD_INIT_ERROR__IMC_INIT:			/* FALL THROUGH */
	case LCD_INIT_ERROR__LCDM_INIT:			/* FALL THROUGH */
		exit_lcdhw();
		free_irq(INT_LCD, NULL);
		/* Reset LCD */
		lcd_hw_reset();
#ifndef IO_ADDRESS
		iounmap(LCDCMmioV);
	case LCD_INIT_ERROR__LCDC_MMIO_NOT_REMAP:	/* FALL THROUGH */
		release_mem_region(LCDCMmio, LCDCMmioLength);
	case LCD_INIT_ERROR__LCDC_MMIO_NOT_RESERVE:	/* FALL THROUGH */
#endif
	default:
		break;
	}

	return;
}


/******************************************************************************
* MODULE   : set_val_init
* FUNCTION : variable default set
* RETURN   : none
* NOTE     : none
* CREATE   : 2006.05.25 H.Nagashima
* UPDATE   :
******************************************************************************/
static void set_val_init(void)
{
	int i;

	/* other img infomation */
	uiMaskColor_tmp      = 0x0020;
	uiAlpha_tmp          = 0xFF;

	/* flags */
	iMixDSPFlg_tmp       = MIX_DSP_OFF;
	iDSPPauseFlg         = PAUSE_DSP_OFF;
	uiMaskColorFlag_tmp  = FB_MASK_COLOR_DISP_OFF;
	uiAbsolutelyUpFlag   = FB_UPDATE_OFF;

	for (i = 0; i < 2; i++) {
		sema_init(&lcd_var[i].sem_image_flag, 1);

		/* input image data from fb driver */
		lcd_var[i].uiMixFrameBufferPage = DISPLAY_FRAME_NO_A;

		/* flags */
		lcd_var[i].iMixDSPFlgToFB       = MIX_DSP_OFF;
		lcd_var[i].iCallbackV4L2Flg     = CALLBACK_V4L2_OFF;

		/* frame data */
		lcd_var[i].pvDevData            = 0x00;
		lcd_var[i].pvVideoBuf           = 0x00;

		lcd_var[i].lcd_refresh_flg = IMC_R_IDLE;
#if defined(CONFIG_PM) || defined(CONFIG_DPM)
		lcd_var[i].ctrl_func       = 0;
#endif /* CONFIG_PM || CONFIG_DPM */
	}

	blank_state.current_mode  = FB_BLANK_UNBLANK;
	blank_state.lcd_backlight = 1;
	blank_state.lcd_output    = 1;
	blank_state.lcd_clock     = 1;

	return;
}


/********************************************************
 *  Timer Function Definitions                          *
 *******************************************************/
/******************************************************************************
* MODULE   : lcd_timer_init
* FUNCTION : timer function init
* RETURN   : none
* NOTE     : private function
* CREATE   : 2006.06.07 H.Nagashima
* UPDATE   :
******************************************************************************/
static void lcd_timer_init(void)
{
	init_timer(&lcd_timer);
	lcd_timer.function = lcd_timeout;
	lcd_timer.data = 0;
	iTimerFlag = LCD_TIMER_STOP;

	return;
}


/******************************************************************************
* MODULE   : lcd_timeout
* FUNCTION : when call mp200_fb_call_timer
* RETURN   : none
* NOTE     : public function
* CREATE   : 2006.06.07 H.Nagashimia
* UPDATE   :
******************************************************************************/
static void lcd_timeout(unsigned long data)
{
	/* change LCD status to PAUSE */
	iDSPPauseFlg = PAUSE_DSP_ON;

	queue_work(mp200_lcd_workqueue, &wk_timeout_bottom_half);
	return;
}


/******************************************************************************
* MODULE   : lcd_timeout_bottom_half_do
* FUNCTION : when call mp200_fb_call_timer
* RETURN   : none
* NOTE     : public function
* CREATE   : 2006.07.21 M.Haguro
* UPDATE   :
******************************************************************************/
static void lcd_timeout_bottom_half_do(struct work_struct *num)
{
	/* get semafore */
	down(&lcd_dev.sem_image_data);
	printk_dbg((_DEBUG_LCD & 0x80), "down(A) lcd_timeout\n");

	/* mix image */
	mix_image(CALLBACK_V4L2_OFF, UPDATE_ON, SET_LAYER_2D);

	return;
}


/********************************************************
 *  Exit Function Definitions                           *
 *******************************************************/
/******************************************************************************
* MODULE   : mp200_lcd_exit_module
* FUNCTION : cleanup LCD module
* RETURN   : none
* NOTE     : none
* CREATE   : 2006.05.25 H.Nagashima
* UPDATE   :
******************************************************************************/
void mp200_lcd_exit_module()
{
	lcd_hw_stop();
	mp200_lcd_remove();
}


/******************************************************************************
* MODULE   : mp200_lcd_remove
* FUNCTION : release LCD resource
* RETURN   : none
* NOTE     : none
* CREATE   : 2006.05.29 H.Nagashima
* UPDATE   :
******************************************************************************/
static void mp200_lcd_remove()
{
	/* power off LCD */
	lcd_hw_backlight_off();
	lcd_hw_power_off();

	free_irq(INT_LCD, NULL);
	iounmap(LCDCMmioV);
	release_mem_region(LCDCMmio, LCDCMmioLength);

	/* Reset LCD */
	lcd_hw_reset();

	return;
}


#if defined(CONFIG_PM) || defined(CONFIG_DPM)
/********************************************************
 *  Suspend/Resume Function Definitions                 *
 *******************************************************/
/******************************************************************************
* MODULE   : mp200_lcd_suspend
* FUNCTION : suspend LCD driver
* RETURN   : 0 : success
* NOTE     : none
* CREATE   : 2006.08.15 H.Nagashima
* UPDATE   :
******************************************************************************/
int mp200_lcd_suspend(struct platform_device *dev, pm_message_t state)
{
	unsigned long ulRegVal32, ulBUSSEL_save;

	if ((state.event == DEV_SUSPEND_IDLE_1) &&
	    ((lcd_var[0].ctrl_func | lcd_var[1].ctrl_func) != 0)) {
		return -EBUSY;
	}

	if (blank_state.lcd_output == 1) {
		ulRegVal32 = lcd_hw_chk_bussel();
		ulBUSSEL_save = ulRegVal32;
		switch (ulRegVal32 & LCD_BUSSEL_BIT) {
		case LCD_BUSSEL_BACKCOLOR:	/* FALL THROUGH */
		case LCD_BUSSEL_BLACK:
			break;
		case LCD_BUSSEL_WB_LOCAL:	/* FALL THROUGH */
		case LCD_BUSSEL_LOCAL:		/* FALL THROUGH */
		default:
			writel(LCD_BUSSEL_BLACK, LCDCMmioV + LCD_BUSSEL);
			break;
		}
	}

	if (state.event == DEV_SUSPEND_IDLE_1) {
		if (blank_state.lcd_output == 1) {
			ulRegVal32 = lcd_hw_chk_status();
			switch (ulRegVal32 & LCD_MODSTATUS_BIT) {
			case LCD_MODSTATUS_BACKCOLOR:	/* FALL THROUGH */
			case LCD_MODSTATUS_BLACK:
				return 0;
			case LCD_MODSTATUS_WB_LOCAL:	/* FALL THROUGH */
			case LCD_MODSTATUS_LOCAL:	/* FALL THROUGH */
			default:
				return -EBUSY;
			}
		} else
			return 0;

	} else { /* state.event == PM_EVENT_SUSPEND */
		if (mp200_sleep_while_idle) {
			/* Restore IMC H/W register data after resume */
			IMC_reset_flg = 1;
			return 0;
		} else {
			DPM_suspend_flg = 1;
			/* LCD timer stop */
			del_timer(&lcd_timer);
			if (blank_state.lcd_output == 1) {
				/* LCD status check */
				lcd_suspend_chk_lcdout();
				if (blank_state.lcd_backlight == 1) {
					/* BackLight OFF  */
					lcd_hw_backlight_off();
				}
				/* LCD display off */
				lcd_hw_stop();
			}
			if (blank_state.lcd_clock == 1) {
				/* Save LCD H/W register data */
				lcd_hw_save_reg();

				/* Reset LCD */
				lcd_hw_reset();
			}
			/* Restore IMC H/W register data after resume */
			IMC_reset_flg = 1;
			return 0;
		}
	}
}


/******************************************************************************
* MODULE   : lcd_suspend_chk_lcdout
* FUNCTION : enabled suspend
* RETURN   :  0 : success
*            -1 : failed
* NOTE     : none
* CREATE   : 2008.10.09 M.Haguro
* UPDATE   :
******************************************************************************/
int lcd_suspend_chk_lcdout(void)
{
	int i = 0;
	int iRet = -1;
	int wait_max  = 32; /* ms */
	int wait_time = 2;  /* ms */
	int wait_cnt  = wait_max / wait_time;
	unsigned long ulRegVal32;

	do {
		ulRegVal32 = lcd_hw_chk_status();
		switch (ulRegVal32 & LCD_MODSTATUS_BIT) {
		case LCD_MODSTATUS_BACKCOLOR:	/* FALL THROUGH */
		case LCD_MODSTATUS_BLACK:	/* FALL THROUGH */
			i    = wait_cnt;
			iRet = 0;
			break;
		case LCD_MODSTATUS_WB_LOCAL:	/* FALL THROUGH */
		case LCD_MODSTATUS_LOCAL:
		default:
			if (i == wait_cnt) {
				printk_dbg((_DEBUG_LCD & 0x01),
				 "failed(%08lx)\n", ulRegVal32);
			} else {
				mdelay(wait_time);
			}
			break;
		}
		i++;
	} while (i <= wait_cnt);

	return iRet;
}


/******************************************************************************
* MODULE   : mp200_lcd_resume
* FUNCTION : suspend LCD driver
* RETURN   : 0 : success
* NOTE     : none
* CREATE   : 2006.08.15 H.Nagashima
* UPDATE   :
******************************************************************************/
int mp200_lcd_resume(struct platform_device *dev)
{
	struct list_head *list_tmp;
	MP200_LCD_VAR *p_lcd_var;
	void (*callback_func[4])(void) = {NULL, NULL, NULL, NULL};
	int i = 0;

	if (DPM_suspend_flg) {
		/* Restore IMC H/W register data */
		if (IMC_reset_flg)
			imc_hw_restore_reg();

		if (blank_state.lcd_clock == 1) {
			/* UnReset LCD */
			lcd_hw_unreset();

			/* Restore LCD H/W register data */
			lcd_hw_restore_reg();
		}
		if (blank_state.lcd_output == 1) {
			/* LCD display on */
			lcd_hw_start();
		}
		if (blank_state.lcd_backlight == 1) {
			/* BackLight ON */
			lcd_hw_backlight_on();
		}

		/* LCD timer start */
		if ((iTimerFlag == LCD_TIMER_START)
		 && (iMixDSPFlg_tmp == MIX_DSP_ON)) {
			lcd_timer.expires = jiffies
			 + (MIX_DSP_TIMEOUT * 10 * HZ / 1000);
			add_timer(&lcd_timer);
		}

		if (blank_state.lcd_output == 1) {
			list_for_each(list_tmp, &list_lcd_var) {
				p_lcd_var = list_entry(list_tmp,
				 struct mp200_lcd_var, list);
				switch (p_lcd_var->ctrl_func) {
				case 2:
					callback_func[i] = lcd_callback_imc_refresh;
					i++;
					/* FALL THROUGH */
				case 0:
				default:
					break;
				}
			}

			for (i = 0; i < 4; i++) {
				if (callback_func[i] == NULL)
					break;
				else
					callback_func[i]();
			}
		}

		DPM_suspend_flg = 0;
	}
	return 0;
}
#endif /* CONFIG_PM || CONFIG_DPM */


MODULE_AUTHOR("NEC Electronics Corporation.");
MODULE_DESCRIPTION("EM1 LCD control driver");
MODULE_LICENSE("GPL");


/* module alias */
#ifdef MODULE
module_init(mp200_lcd_init_module);
module_exit(mp200_lcd_exit_module);
#else
#ifndef CONFIG_FB_MP200
device_initcall(mp200_lcd_init_module);
#endif
#endif
