/*
 * File Name       : /drivers/video/mp200/nts/mp200_nts_common.h
 * Function        : NTSC Driver corresponding to V4L2
 * Release Version : Ver 1.01
 * Release Date    : 2009/10/19
 *
 * Copyright (C) NEC Electronics Corporation 2009
 *
 *
 * This program is free software;you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by Free Softwere
 * Foundation; either version 2 of License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warrnty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; If not, write to the Free Software Foundation, Inc., 59 Temple
 * Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 */


#ifndef _MP200_NTS_COMMON_H_
#define _MP200_NTS_COMMON_H_


/********************************************************
 *  Definitions                                         *
 *******************************************************/
/* for iOutmodeFlg */
#define FB_OFF			0x01
#define FB_ON			0x02
#define V4L2_OFF		0x04
#define V4L2_ON			0x08

#define FB_V4L2_OFF		0x05
#define FB_ONLY			0x06
#define V4L2_ONLY		0x09
#define FB_V4L2_ON		0x0A

#define FB_BIT			0x03
#define V4L2_BIT		0x0C

/* display frame no. for iFrameNoNow/iFrameNoNext */
#define NTS_DISP_INIT		NTS_AREASEL_BUFDISABLE
#define NTS_DISP_FRAME_A	NTS_AREASEL_BUFA
#define NTS_DISP_FRAME_B	NTS_AREASEL_BUFB
#define NTS_DISP_FRAME_C	NTS_AREASEL_BUFC


/********************************************************
 *  Structure                                           *
 *******************************************************/
/* structure video endcoder */
typedef struct encoder_reg {
	char   name[32];
	int    (*hw_init)(void *);
	void   (*hw_shutdown)(void *);
	int    (*hw_command)(unsigned int cmd, void *arg);
	void  *private_data;
} ENCODER_REG;

/* structure FBIOBLANK status */
typedef struct nts_blank_state {
	int current_mode;
	int nts_backlight;
	int nts_output;
	int nts_clock;
} NTS_BLANK_STATE;

/* structure nts device status */
struct mp200_nts_dev {
	/* semafore */
	 /* semafore v4l2/fb/img local data */
	struct semaphore         sem_img_data;
	 /* semafore v4l2/fb/img local flag */
	struct semaphore         sem_img_flag;
	/* spinlock */
	 /* spin lock for NTS               */
	spinlock_t               nts_lock;
	 /* spin lock for NTS timer         */
	spinlock_t               lock_tmr;
	 /* spin lock for callback          */
	spinlock_t               lock_callback;
	/* work queue */
	struct workqueue_struct *mp200_nts_workqueue;
	struct work_struct       wk_timeout_bottom_half;

	/* waite timer. movie is pause state */
	struct timer_list        nts_timer;
	int                      iTimerFlag;	/* timer operation status */

	/* wait queue */
	wait_queue_head_t        wait_img;
	int                      img_complete;

	/* flags */
	int                      iNtsActive;
	/* next frame buffer (2D)          */
	unsigned int             uiMixFrameBufferPage;

	/* movie & UI mix ON/OFF           */
	int                      iMixDSPFlg;
	/* movie & UI mix ON/OFF (for FB)  */
	int                      iMixDSPFlg_toFb;

	/* V4L2 pause status flag          */
	int                      iV4L2PauseFlg;
	/* need to callback to V4L2        */
	int                      iV4L2CallbackFlg;
	struct _FRAME_DATA       v4l2_callback_data;

	int                      iFrameNoNow;	/* now display frame no  */
	int                      iFrameNoNext;	/* set change frame no   */

	int                      iOutmodeFlg;	/* output i/f            */
	int                      iMixImageMode;	/*  */

	/* absolutely update flag    */
	unsigned int             uiAbsolutelyUpFlag;

	FB_IMAGE_INFO            from_fb;
	V4L2_IMAGE_INFO          from_v4l2;

	/* NTSC Encoder */
	ENCODER_REG             *encoder;

	/* FBIOBLANK */
	NTS_BLANK_STATE          blank_state;
	int                      ntsout_flg;
};

#endif /* _MP200_NTS_COMMON_H_ */
