#ifndef _ASM_X86_TRACE_CLOCK_H
#define _ASM_X86_TRACE_CLOCK_H

/*
 * linux/include/asm-x86/trace-clock.h
 *
 * Copyright (C) 2005,2006,2008
 *   Mathieu Desnoyers (mathieu.desnoyers@polymtl.ca)
 *
 * Trace clock definitions for x86.
 */

#include <linux/timex.h>
#include <asm/system.h>
#include <asm/processor.h>
#include <asm/atomic.h>

/* Minimum duration of a probe, in cycles */
#define TRACE_CLOCK_MIN_PROBE_DURATION 200

#ifdef CONFIG_HAVE_TRACE_CLOCK_32_TO_64
/* Only for testing. Never needed on x86. */
u64 trace_clock_read_synthetic_tsc(void);
#endif

#ifdef CONFIG_HAVE_UNSTABLE_TSC
extern int tsc_is_sync;

extern cycles_t trace_clock_async_tsc_read(void);

static inline u32 trace_clock_read32(void)
{
	u32 cycles;

	if (likely(tsc_is_sync)) {
		rdtsc_barrier();
		cycles = (u32)get_cycles(); /* only need the 32 LSB */
		rdtsc_barrier();
	} else
		cycles = (u32)trace_clock_async_tsc_read();
	return cycles;
}

static inline u64 trace_clock_read64(void)
{
	u64 cycles;

	if (likely(tsc_is_sync)) {
		rdtsc_barrier();
		cycles = get_cycles();
		rdtsc_barrier();
	} else
		cycles = trace_clock_async_tsc_read();
	return cycles;
}
#else /* CONFIG_HAVE_UNSTABLE_TSC */
static inline u32 trace_clock_read32(void)
{
	u32 cycles;

	rdtsc_barrier();
	cycles = (u32)get_cycles(); /* only need the 32 LSB */
	rdtsc_barrier();
	return cycles;
}

static inline u64 trace_clock_read64(void)
{
	u64 cycles;

	rdtsc_barrier();
	cycles = get_cycles();
	rdtsc_barrier();
	return cycles;
}
#endif /* CONFIG_HAVE_UNSTABLE_TSC */

/*
 * Periodic IPI to have an upper bound on TSC inaccuracy.
 * TODO: should implement this in arch/x86/kernel/trace-clock.c.
 */
static inline void trace_clock_add_timestamp(unsigned long ticks)
{ }

static inline unsigned int trace_clock_frequency(void)
{
	return cpu_khz;
}

static inline u32 trace_clock_freq_scale(void)
{
	return 1000;
}

#endif /* _ASM_X86_TRACE_CLOCK_H */
